// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#ifndef __GENOME_ANNOTATION
#define	__GENOME_ANNOTATION

#include <string>
#include <map>
#include <list>
#include <iostream>
#include <fstream>
#include <sstream>
#include "sequence.h"
#include "periodic_boundary.h"

#include <seqport.h>
#include <tofasta.h>

// OS specific defines and includes
#ifdef WIN32

// Handle the annoying carrige return problem in windows
#define ENDL	"\r" << endl
#define EOL	"\r\n"

#else

#define ENDL	endl
#define EOL	"\n"

#endif // WIN32

// A class for storing sequence search results
class SeqHit{
private:
	unsigned int match_start;
	unsigned int match_stop;
	
	// How many 3'/5' bases exactly match between primer and target?
	unsigned int clamp_1;
	unsigned int clamp_2;

	float match_tm_1; // Melting Tm for probe or first primer
	float match_tm_2; // Melting Tm for second primer

	// Id symbols for the first and second primers
	char id_1;
	char id_2;
	
	char match_type;

	bool strand_neg;
	
	// The number of Watson and Crick base pairs
	unsigned int num_match_1;
	unsigned int num_match_2;
	
	// The length of the oligos
	unsigned int len_1;
	unsigned int len_2;
	
public:
	enum {EXACT_MATCH, PROBE_MATCH, PRIMER_MATCH, MOL_PRIMER_MATCH, UNKNOWN_MATCH};

	SeqHit(const unsigned int &m_start, const unsigned int &m_stop,
		const bool &m_complement, const char &m_type) : 
		match_start(m_start), match_stop(m_stop),
		strand_neg(m_complement)
	{
		// It is possible for m_start > m_stop (i.e. circular genome)

		match_tm_1 = -1.0;
		match_tm_2 = -1.0;
		
		clamp_1 = 0;
		clamp_2 = 0;
		
		num_match_1 = 0;
		num_match_2 = 0;
	
		len_1 = 0;
		len_2 = 0;
	
		if( (m_type < EXACT_MATCH) || (m_type > UNKNOWN_MATCH) ){
			throw "Unknown SeqHit type!";
		}
		
		match_type = m_type;
	};

	SeqHit(const unsigned int &m_start, const unsigned int &m_stop, const float &m_tm,
		const bool &m_complement, const char &m_type) : 
		match_start(m_start), match_stop(m_stop), match_tm_1(m_tm),
		strand_neg(m_complement)
	{
		// It is possible for m_start > m_stop (i.e. circular genome)
		
		id_1 = 0;
		id_2 = 0;
	
		match_tm_2 = -1.0;
		
		clamp_1 = 0;
		clamp_2 = 0;
		
		num_match_1 = 0;
		num_match_2 = 0;
	
		len_1 = 0;
		len_2 = 0;
		
		if( (m_type < EXACT_MATCH) || (m_type > UNKNOWN_MATCH) ){
			throw "Unknown SeqHit type!";
		}
		
		match_type = m_type;
	};

	SeqHit(const unsigned int &m_start, const unsigned int &m_stop, const float &m_tm_1, 
		const char &m_id_1, const unsigned int &m_clamp1, const float &m_tm_2, 
		const char &m_id_2, const unsigned int &m_clamp2, const bool &m_complement, 
		const char &m_type) : 
		match_start(m_start), match_stop(m_stop),
		clamp_1(m_clamp1), clamp_2(m_clamp2),
		match_tm_1(m_tm_1), match_tm_2(m_tm_2), 
		id_1(m_id_1), id_2(m_id_2),
		strand_neg(m_complement)
	{
		num_match_1 = 0;
		num_match_2 = 0;
	
		len_1 = 0;
		len_2 = 0;
		
		// It is possible for m_start > m_stop (i.e. circular genome)
		if( (m_type < EXACT_MATCH) || (m_type > UNKNOWN_MATCH) ){
			throw "Unknown SeqHit type!";
		}
		
		match_type = m_type;
	};

	inline char type() const
	{
		return match_type;
	};

	inline unsigned int start() const
	{
		return match_start;
	};

	inline void start(const unsigned int &m_match_start)
	{
		match_start = m_match_start;
	};

	inline unsigned int stop() const
	{
		return match_stop;
	};

	inline void stop(const unsigned int &m_match_stop)
	{
		match_stop = m_match_stop;
	};

	inline float probe_tm() const
	{
		return match_tm_1;
	};

	inline float primer_tm_1() const
	{
		return match_tm_1;
	};

	inline float primer_tm_2() const
	{
		return match_tm_2;
	};
	
	inline unsigned int primer_clamp_1() const
	{
		return clamp_1;
	};

	inline unsigned int primer_clamp_2() const
	{
		return clamp_2;
	};

	inline char primer_id_1() const
	{
		return id_1;
	};

	inline char primer_id_2() const
	{
		return id_2;
	};

	inline unsigned int primer_match_1() const
	{
		return num_match_1;
	};
	
	inline void primer_match_1(const unsigned int &m_num)
	{
		num_match_1 = m_num;
	};
	
	inline unsigned int primer_match_2() const
	{
		return num_match_2;
	};
	
	inline void primer_match_2(const unsigned int &m_num)
	{
		num_match_2 = m_num;
	};
	
	inline unsigned int probe_match() const
	{
		return num_match_1;
	};
	
	inline void probe_match(const unsigned int &m_num)
	{
		num_match_1 = m_num;
	};
	
	inline unsigned int primer_len_1() const
	{
		return len_1;
	};
	
	inline void primer_len_1(const unsigned int &m_len)
	{
		len_1 = m_len;
	};
	
	inline unsigned int primer_len_2() const
	{
		return len_2;
	};
	
	inline void primer_len_2(const unsigned int &m_len)
	{
		len_2 = m_len;
	};
	
	inline unsigned int probe_len() const
	{
		return len_1;
	};
	
	inline void probe_len(const unsigned int &m_len)
	{
		len_1 = m_len;
	};
	
	inline bool complement() const
	{
		return strand_neg;
	};

	inline void complement(const bool &m_comp)
	{
		strand_neg = m_comp;
	};

	bool operator < (const SeqHit &m_copy) const
	{
		return match_start < m_copy.match_start;
	};

	bool operator > (const SeqHit &m_copy) const
	{
		return match_start > m_copy.match_start;
	};
};

class GeneAnnotation{
private:
	unsigned int gene_type;
	
	// From the SeqLoc. My current understanding is that
	// the genome is encoded in order of 5'->3' translation.
	bool complement; // Is this gene on the opposite strand?
	unsigned int gene_start, gene_stop;
	
	// Is this gene divided into multiple regions?
	std::list< std::pair<unsigned int, unsigned int> > seg_list;

	// If this is a top level record, does it overlap other top level records?
	unsigned int top_depth;
	
	// Does this record overlap other records?
	unsigned int total_depth;
	
	// Has the user selected this annotation?
	bool selected;
	
	SeqIdPtr sip;
	
	std::vector<std::string> info_map;

	// Allow the user to specify a custom color on a per-annotation
	// basis
	float *custom_color;

	// Every annotation has a rank that lets us know when it is appropriate to 
	// show or display this annotation
	unsigned char rank;

	void write_CDS_gbk(std::ofstream &m_fout, const DNA3Seq &m_seq) const;
	void write_GENE_gbk(std::ofstream &m_fout, const DNA3Seq &m_seq) const;
	void write_PSEUDO_GENE_gbk(std::ofstream &m_fout) const;
	void write_RNA_gbk(std::ofstream &m_fout) const;
	void write_tRNA_gbk(std::ofstream &m_fout) const;
	void write_IMP_gbk(std::ofstream &m_fout) const;
	void write_PRIMER_gbk(std::ofstream &m_fout) const;
	void write_USER_gbk(std::ofstream &m_fout) const;
	void write_NONE_gbk(std::ofstream &m_fout) const;
	void write_seg_GBK(std::ofstream &m_fout) const;
	
public:
	enum {MID_LEVEL_RANK = 0, TOP_LEVEL_RANK};

	enum {LOCUS = 0, LOCUS_TAG, PRODUCT, EC, NOTE, 
		PRIMER_TM, PRIMER_GC, PRIMER_HAIRPIN_TM, PRIMER_DIMER_TM, 
		COG_CODE, COG_ID,
		LAST_INFO}; // info
	enum {CDS = 0, GENE, PSEUDO_GENE, RNA, tRNA, IMP, PRIMER, USER, NONE,
		LAST_ANNOT}; // gene_type
	
	friend std::ostream& operator << (std::ostream &s, 
		const std::pair<unsigned int, GeneAnnotation> &m_annot);
	
	GeneAnnotation()
	{
		complement = false;
		selected = false;
		gene_type = NONE;
		gene_start = gene_stop = 0;
		top_depth = 0;
		total_depth = 0;
		sip = NULL;
		info_map = std::vector<std::string>(LAST_INFO);
		custom_color = NULL;
		rank = 0;
	};
	
	GeneAnnotation(const GeneAnnotation &m_copy)
	{
		complement = false;
		selected = false;
		gene_type = NONE;
		gene_start = gene_stop = 0;
		top_depth = 0;
		total_depth = 0;
		rank = 0;
		sip = NULL;
		custom_color = NULL;

		*this = m_copy;
	};
	
	~GeneAnnotation()
	{
		if(sip){
			sip = SeqIdSetFree(sip);
		}

		if(custom_color){
			delete [] custom_color;
		}
	};
	
	inline void clear()
	{
		if(sip){
			sip = SeqIdSetFree(sip);
		}
		
		if(custom_color){
			delete [] custom_color;
			custom_color = NULL;
		}

		complement = false;
		selected = false;
		gene_type = NONE;
		gene_start = gene_stop = 0;
		top_depth = 0;
		total_depth = 0;
		rank = 0;
		sip = NULL;
		info_map = std::vector<std::string>(LAST_INFO);
		seg_list.clear();
	}

	GeneAnnotation& operator=(const GeneAnnotation& m_copy)
	{
		selected = m_copy.selected;
		gene_type = m_copy.gene_type;
		complement = m_copy.complement;
		gene_start = m_copy.gene_start;
		gene_stop = m_copy.gene_stop;
		top_depth = m_copy.top_depth;
		total_depth = m_copy.total_depth;
		rank = m_copy.rank;

		if(sip){
			sip = SeqIdSetFree(sip);
		}
				
		if(m_copy.sip){
			sip = SeqIdSetDup(m_copy.sip);
		}
		
		if(custom_color){
			delete [] custom_color;
			custom_color = NULL;
		}

		if(m_copy.custom_color){
		
			custom_color = new float [4];
			memcpy(custom_color, m_copy.custom_color, 4*sizeof(float));
		}

		info_map = m_copy.info_map;
		seg_list = m_copy.seg_list;

		return (*this);
	};
	
	GeneAnnotation& operator=(const SeqFeatPtr &sfp);
	
	inline bool operator==(const SeqIdPtr &m_sip) const
	{

		// Only compare valid SeqIdPtr
		if(!sip || !m_sip){
			return false;
		}

		// An optimization for SEQID_GI
		if( (sip->choice == SEQID_GI) && (m_sip->choice == SEQID_GI) ){
			return (sip->data.intvalue == m_sip->data.intvalue);
		}

		SeqIdPtr a = sip;
		
		while(a){
			SeqIdPtr b = m_sip;
			
			while(b){
				switch(SeqIdComp(a, b)){
					case SIC_YES:
						return true;
					case SIC_NO:
						return false;
					case SIC_DIFF:
						// Couldn't compare -- keep going!
						break;
				};
				
				b = b->next;
			}
			
			a = a->next;
		}
		
		return false;
	};
	
	inline bool operator!=(const SeqIdPtr &m_sip) const
	{
		return !(*this == m_sip);
	};
	
	inline bool operator==(const GeneAnnotation &m_copy) const
	{
		// If we have valid SeqIdPtrs, compare them
		if(sip && m_copy.sip){

			// Add an optimization for SEQID_GI
			return ( (sip->choice == SEQID_GI) && (m_copy.sip->choice == SEQID_GI) ) ?
				(sip->data.intvalue == m_copy.sip->data.intvalue) : 
				(SeqIdComp(sip, m_copy.sip) == SIC_YES);
		}
		
		if(complement != m_copy.complement){
			return false;
		}
		
		if(gene_start != m_copy.gene_start){
			return false;
		}
		
		if(gene_stop != m_copy.gene_stop){
			return false;
		}
		
		return true;
	};
	
	inline bool operator<(const GeneAnnotation& m_copy) const
	{		
		const bool this_overlap = (gene_start > gene_stop);
		const bool copy_overlap = (m_copy.gene_start > m_copy.gene_stop);

		// Do either this annotation or the copy annotation (but not 
		// both) overlap the origin?
		if( (this_overlap || copy_overlap) && !(this_overlap && copy_overlap) ){

			// One of the two annotations overlaps the origin. By definition, this
			// annotation is "less than" the other.
			return this_overlap;
		}
		
		if(gene_start < m_copy.gene_start){
			return true;
		}
		
		if(gene_start > m_copy.gene_start){
			return false;
		}
		
		// start == m_copy.start
		return gene_stop < m_copy.gene_stop;
	};
	
	inline bool operator>(const GeneAnnotation& m_copy) const
	{		
		const bool this_overlap = (gene_start > gene_stop);
		const bool copy_overlap = (m_copy.gene_start > m_copy.gene_stop);

		// Do either this annotation or the copy annotation (but not 
		// both) overlap the origin?
		if( (this_overlap || copy_overlap) && !(this_overlap && copy_overlap) ){

			// One of the two annotations overlaps the origin. By definition, the
			// annotation that does NOT overlap the origin is "greater than" the other.
			return copy_overlap;
		}

		if(gene_start > m_copy.gene_start){
			return true;
		}
		
		if(gene_start < m_copy.gene_start){
			return false;
		}
		
		// start == m_copy.start
		return gene_stop > m_copy.gene_stop;
	};

	inline bool overlaps_origin() const
	{
		return (gene_start > gene_stop);
	};
	
	inline void seqid(const SeqIdPtr m_sid)
	{
		if(m_sid == NULL){
			throw "GeneAnnotation:seqid: NULL input SeqId";
		}
		
		if(sip){
			sip = SeqIdSetFree(sip);
		}
		
		sip = SeqIdSetDup(m_sid);
	};
	
	inline int gi() const
	{
		SeqIdPtr tmp_sip = sip;
	
		while(tmp_sip != NULL){
			if(tmp_sip->choice == SEQID_GI){
				return tmp_sip->data.intvalue;
			}
			
			tmp_sip = tmp_sip->next;
		}
		
		return -1;
	};
	
	// Return the genbank accesion
	inline std::string accession(const bool &m_include_version) const
	{
		SeqIdPtr tmp_sip = sip;
		char buffer[256];
		
		while(tmp_sip != NULL){
			if(tmp_sip->choice == SEQID_GENBANK){
			
				SeqIdWrite(tmp_sip, buffer, 
					(m_include_version) ? PRINTID_TEXTID_ACC_VER :
						PRINTID_TEXTID_ACC_ONLY,
					256);
				
				return buffer;
			}
			
			tmp_sip = tmp_sip->next;
		}
		
		return "";
	};
	
	inline std::string seq_id_str() const
	{
		char buffer[256];

		if(sip == NULL){
			return "";
		}
		
		SeqIdPrint(sip, buffer, '|');

		return buffer;
	};

	inline void info(unsigned int m_key, const char *m_str)
	{
		if(m_str != NULL){
			info_map[m_key] = m_str;
		}
	};
	
	inline void info(unsigned int m_key, const std::string &m_str)
	{
		info_map[m_key] = m_str;
	};
	
	inline std::string info(unsigned int m_key) const
	{	
		return info_map[m_key];
	};
	
	inline std::string name()
	{
		return (info_map[LOCUS] != "") ? info_map[LOCUS] : info_map[LOCUS_TAG];
	};

	inline void primer_id(const unsigned int &m_id)
	{
		std::stringstream sout;

		sout << "P" << m_id;

		info_map[LOCUS] = sout.str();
	};

	inline void primer_tm(const float &m_tm)
	{
		std::stringstream sout;

		sout << "Tm = " << m_tm;

		info_map[PRIMER_TM] = sout.str();
	};

	inline void primer_dimer_tm(const float &m_tm)
	{
		std::stringstream sout;

		sout << "Dimer Tm = " << m_tm;

		info_map[PRIMER_DIMER_TM] = sout.str();
	};

	inline void primer_hairpin_tm(const float &m_tm)
	{
		std::stringstream sout;

		sout << "Hairpin Tm = " << m_tm;

		info_map[PRIMER_HAIRPIN_TM] = sout.str();
	};

	inline void primer_gc(const float &m_gc)
	{
		std::stringstream sout;

		sout << "%G+C = " << 100.0f*m_gc;

		info_map[PRIMER_GC] = sout.str();
	};

	// Does this annotation encode a protein?
	inline bool is_protein() const
	{
		return (gene_type < RNA);
	};

	// What is the rank of this annotation?
	inline unsigned char annot_rank () const
	{
		return rank;
	};

	inline void annot_rank (const unsigned char &m_rank)
	{
		rank = m_rank;
	};

	inline bool is_gene() const
	{
		// This function assumes that all types greater than tRNA are
		// NOT genes!
		return (gene_type <= tRNA);
	};

	inline bool is_intergenic() const
	{
		return (gene_type == NONE);
	};
	
	inline bool has_id_str() const
	{
		return !((gene_type == IMP) || (gene_type == NONE));
	};

	inline bool is_complement() const
	{
		return complement;
	};

	inline void is_complement(const bool &m_comp)
	{
		complement = m_comp;
	};

	inline void start(const unsigned int &m_start)
	{
		gene_start = m_start;
	};
	
	inline  unsigned int start() const
	{
		return gene_start;
	};
	
	inline void stop(const unsigned int &m_stop)
	{
		gene_stop = m_stop;
	};
	
	inline unsigned int stop() const
	{
		return gene_stop;
	};

	inline unsigned int length(const unsigned int &m_seq_len) const
	{
		if(gene_start > gene_stop){
			
			if(m_seq_len < gene_start){
				throw __FILE__ ":length: gene_start is out of bounds!";
			}
			
			return (m_seq_len - gene_start) + gene_stop + 1;
		}

		return gene_stop - gene_start + 1;
	};

	inline unsigned int depth() const
	{
		return top_depth;
	};
	
	inline unsigned int annot_depth() const
	{
		return total_depth;
	};
	
	inline void depth(const unsigned int &m_depth)
	{
		top_depth = m_depth;
	};
	
	inline void annot_depth(const unsigned int &m_depth)
	{
		total_depth = m_depth;
	};
	
	inline void strand(const unsigned int &m_strand)
	{
		complement = (m_strand == Seq_strand_minus);
	};
	
	inline unsigned int strand() const
	{
		if(complement){
			return Seq_strand_minus;
		}
		
		return Seq_strand_plus;
	};
	
	inline void type(const unsigned int &m_type)
	{
		if((m_type < 0) || (m_type > NONE)){
			throw "GeneAnnotation: Unknown gene type";
		}
		
		gene_type = m_type;
	};
	
	inline unsigned int type() const
	{
		return gene_type;
	};
	
	inline void select(const bool &m_flag)
	{
		selected = m_flag;
	};

	inline bool select() const
	{
		return selected;
	};

	inline void toggle_select()
	{
		selected = !selected;
	};

	inline void segments(const std::list< std::pair<unsigned int, unsigned int> > &m_seg_list)
	{
		seg_list = m_seg_list;
	};

	inline bool is_segmented() const
	{
		return (seg_list.empty() == false);
	};

	inline const std::list< std::pair<unsigned int, unsigned int> >& segments() const
	{
		return seg_list;
	};
	
	// Turn a range or segment string into a start/stop and segment list
	void range(const std::string &m_range);
	
	// Using the start/stop and segment list, return a valid range string
	std::string range() const;
	
	DNA3Seq subseq(const DNA3Seq &m_seq, const unsigned int &m_start, const unsigned int &m_stop) const;

	// Handle the special case of a gene that overlaps the genome start location
	inline bool handle_gene_start_overlap(const unsigned int &m_genome_len)
	{
		if( (gene_start == 0) && (seg_list.empty() == false) ){

			// Look for a segment stop that is equal to m_genome_len - 1
			const unsigned int tmp_len = m_genome_len - 1;
			std::list< std::pair<unsigned int, unsigned int> >::const_iterator iter;
			
			unsigned int tmp_start = 0;
			unsigned int tmp_stop = 0;

			for(iter = seg_list.begin();iter != seg_list.end();iter++){
				
				if(iter->first == 0){
					tmp_stop = iter->second;
				}

				if(iter->second == tmp_len){
					tmp_start = iter->first;
				}
			}

			if((tmp_start != 0) && (tmp_stop != 0)){
				// We found a match! Set the new start and stop files
				// and clear the seq_list.
				seg_list.clear();
				gene_start = tmp_start;
				gene_stop = tmp_stop;

				return true;
			}

		}
		
		if(gene_stop >= m_genome_len){
			gene_stop = gene_stop - m_genome_len;
			
			return true;
		}
		
		if(gene_start > gene_stop){
			return true;
		}
		
		return false;
	};

	inline bool is_custom_color() const
	{
		return (custom_color != NULL);
	};

	inline float* color() const
	{
		return custom_color;
	};

	inline void color(const float m_color[])
	{
		if(custom_color){
			delete [] custom_color;
		}

		custom_color = new float [4];

		memcpy(custom_color, m_color, 3*sizeof(float));

		custom_color[3] = 0.0f;
	};

	inline void color(const float &m_r, const float &m_g, const float &m_b)
	{
		if(custom_color){
			delete [] custom_color;
		}
		
		custom_color = new float [4];
		
		custom_color[0] = m_r;
		custom_color[1] = m_g;
		custom_color[2] = m_b;
		custom_color[3] = 0.0f;		
	};

	inline void color_alpha(const float &m_alpha)
	{
		if(custom_color != NULL){
			custom_color[3] = m_alpha;
		}
	};

	void invert_seg_list(const unsigned int &m_num_base);
	
	void write_gbk(std::ofstream &m_fout, const DNA3Seq &m_seq) const;
};

// Classes to manage an entire DNA molecule
class DNAMol{
private:
	SeqIdPtr sip;
	
	// The sequence data
	DNA3Seq seq;
	
	// From the SeqLoc:	
	bool complement;
	unsigned int start, stop;
	
	unsigned int num_top_level, num_mid_level;
	bool is_circular; // Is this genome circular or linear?
	float gc_content;
	
	float a_content;
	float t_content;
	float c_content;
	float g_content;
	float other_content;

	unsigned int scale_mode;

	std::vector<std::string> info_map;
	
	std::list<GeneAnnotation> gene_list;
		
	// Load from a file [this is a wrapper for
	// loadASN(SeqEntryPtr sep)].
	bool loadASN(const std::string &m_filename, std::streampos &m_pos);

	// Load from the NCBI web site [this is a wrapper for
	// loadASN(SeqEntryPtr sep].
	bool loadNetASN(const unsigned int &m_gi);

	// Load from the NCBI web site and save to disk
	bool loadNetASN(const unsigned int &m_gi, const std::string &m_filename);

	// Load from a pointer
	bool loadASN(SeqEntryPtr sep, SeqIdPtr sid = NULL);

	// Load from a FASTA file
	bool loadFASTA(const std::string &m_filename);

	// Load from a PTT file
	bool loadPTT(const std::string &m_filename);
	
	// Load from a GBK file
	bool loadGBK(const std::string &m_filename, std::streampos &m_pos);
	void loadGBKFeatures(std::ifstream &m_fin);

	// Load from a EMBL file
	bool loadEMBL(const std::string &m_filename, std::streampos &m_pos);
	void loadEMBLFeatures(std::ifstream &m_fin);

	void taxaInfo(ValNodePtr m_vnp);
	
public:
	// How to we rank genes and other annotations?
	enum {GENE_CENTRIC = 0, SIZE_CENTRIC};

	enum {SOURCE, TAXA_NAME, LINEAGE, 
		GENUS, SPECIES, SUBSPECIES, LAST_INFO}; // info_map
	
	// Data input types types:
	// ASN_1 = NCBI's ASN.1 annotation format (either binary or ASCII)
	// FASTA = ASCII sequence data in FASTA format
	// GBK = ASCII Genbank annontation flat file (similar to EMBL)
	// EMBL = ASCII EMBL annotation file (similar to GBK)
	// PTT = Protein translation table -- no sequence!
	enum {ASN_1, FASTA, GBK, EMBL, PTT, LAST_ANNOT_FORMAT};
	
	friend std::ostream& operator << (std::ostream &s, 
		const DNAMol &m_mol);
		
	DNAMol()
	{
		start = stop = 0;
		gc_content = 0.0f;
		complement = false;
		sip = NULL;
		num_top_level = num_mid_level = 0;
		is_circular = true; // Assume that we're working with bacteria
		info_map = std::vector<std::string>(LAST_INFO);
		//scale_mode = GENE_CENTRIC;
		scale_mode = SIZE_CENTRIC;
	};
	
	~DNAMol()
	{
		if(sip){
			sip = SeqIdSetFree(sip);
		}
	};
	
	// A helper function for processing gene annotation lists
	// (i.e. sorting, adding intergenic space, counting the numner of 
	// annotations, etc.).
	void processGeneList(const bool &m_loading = false);

	inline std::string seq_id_str() const
	{
		char buffer[256];

		if(sip == NULL){
			return "";
		}
		
		SeqIdPrint(sip, buffer, '|');

		return buffer;
	};
	
	// Return the genbank accesion
	inline std::string accession(const bool &m_include_version) const
	{
		SeqIdPtr tmp_sip = sip;
		char buffer[256];
		
		while(tmp_sip != NULL){
			if(tmp_sip->choice == SEQID_GENBANK){
			
				SeqIdWrite(tmp_sip, buffer, 
					(m_include_version) ? PRINTID_TEXTID_ACC_VER :
						PRINTID_TEXTID_ACC_ONLY,
					256);
				
				return buffer;
			}
			
			tmp_sip = tmp_sip->next;
		}
		
		return "";
	};
	
	inline int gi() const
	{
		SeqIdPtr tmp_sip = sip;
	
		while(tmp_sip != NULL){
			if(tmp_sip->choice == SEQID_GI){
				return tmp_sip->data.intvalue;
			}
			
			tmp_sip = tmp_sip->next;
		}
		
		return -1;
	};
	
	inline bool operator==(const SeqIdPtr &m_sip)
	{
		// Only compare valid SeqIdPtr
		if(!sip || !m_sip){
			return false;
		}

		// An optimization for SEQID_GI
		if((sip->choice == SEQID_GI) && (m_sip->choice == SEQID_GI)){
			return (sip->data.intvalue == m_sip->data.intvalue);
		}

		SeqIdPtr a = sip;
		
		while(a){
			SeqIdPtr b = m_sip;
			
			while(b){
				switch(SeqIdComp(a, b)){
					case SIC_YES:
						return true;
					case SIC_NO:
						return false;
					case SIC_DIFF:
						// Couldn't compare -- keep going!
						break;
				};
				
				b = b->next;
			}
			
			a = a->next;
		}
		
		return false;
	};

	inline std::list<GeneAnnotation>::iterator begin()
	{
		
		return gene_list.begin();
	};
	
	inline std::list<GeneAnnotation>::iterator end()
	{
		
		return gene_list.end();
	};

	inline std::list<GeneAnnotation>::const_iterator begin() const
	{
		
		return gene_list.begin();
	};
	
	inline std::list<GeneAnnotation>::const_iterator end() const
	{
		
		return gene_list.end();
	};

	inline bool empty() const
	{
		return (seq.size() == 0);
	};

	inline bool has_sequence_data() const
	{
		DNA3Seq::iterator iter;
		
		for(iter = seq.begin();iter != seq.end();iter++){
			
			if(iter.base() < DNA3Seq::UNKNOWN){
				return true;
			}
		}
		
		return false;
	};
	
	unsigned int num_select() const
	{
		std::list<GeneAnnotation>::const_iterator iter;
		
		unsigned int count = 0;
		
		for(iter = gene_list.begin();iter != gene_list.end();iter++){
		
			if( iter->select() ){
				count++;
			}
		}
		
		return count;
	};
	
	// Return the first selected annotation
	inline std::list<GeneAnnotation>::const_iterator selected_annotation() const
	{
		std::list<GeneAnnotation>::const_iterator iter;
		
		for(iter = gene_list.begin();iter != gene_list.end();iter++){
		
			if( iter->select() ){
				return iter;
			}
		}
		
		// If we get here, there is no selected annotation
		return gene_list.end();
	};
	
	void update_annotation(const GeneAnnotation &m_annot, 
		const std::list<GeneAnnotation>::const_iterator &m_iter);

	inline DNA3Seq& sequence()
	{
		return seq;
	};
	
	inline float gc() const 
	{
		return gc_content;
	};

	inline float a() const 
	{
		return a_content;
	};

	inline float t() const 
	{
		return t_content;
	};

	inline float g() const 
	{
		return g_content;
	};

	inline float c() const 
	{
		return c_content;
	};

	inline float other() const 
	{
		return other_content;
	};

	inline bool circular() const
	{
		return is_circular;
	};

	inline unsigned int strand() const
	{
		if(complement){
			return Seq_strand_minus;
		}
		
		return Seq_strand_plus;
	};

	// Load and append the annotation file
	bool load(const std::string &m_filename, const unsigned int &m_type, std::streampos &m_pos);
	bool load(const int &m_gi);

	// Load and append the specified annotation as well as saving it to disk
	// with the given filename
	bool load_and_save(const int &m_gi, const std::string &m_filename);

	bool import(const DNAMol &m_mol, const unsigned int &m_type);
	
	void clear_select();

	// Remove any previously computed primers
	void clear_primers();

	// Flush all data
	void clear()
	{
		if(sip){
			sip = SeqIdSetFree(sip);
		}
		
		start = stop = 0;
		info_map.clear();

		seq = "";
		
		gene_list.clear();
	};
	
	inline std::string info(unsigned int m_key) const
	{	
		return info_map[m_key];
	};
	
	inline unsigned int top_level_count() const
	{
		return num_top_level;
	};

	inline unsigned int mid_level_count() const
	{
		return num_mid_level;
	};

	inline void mode(const unsigned int &m_mode)
	{
		if(scale_mode == m_mode){
			// Already in the correct mode
			return;
		}

		if( (m_mode < GENE_CENTRIC) || (m_mode > SIZE_CENTRIC) ){
			throw "Illegal scale mode";
		}

		scale_mode = m_mode;

		if(!empty()){
			// Recompute depths and ranks of all annotations
			processGeneList();
		}
	};

	// Allow the user to search for all annotations that
	// contain or partially contain the given DNA sequence.
	std::list< std::list<GeneAnnotation>::iterator > find_seq(const std::list<DNA3Seq> &m_seq, 
		std::list<SeqHit> &hit_list);
	std::list< std::list<GeneAnnotation>::iterator > find_name(const std::string &m_name);
	std::list< std::list<GeneAnnotation>::iterator > find_gi(const unsigned int &m_gi);
	std::list< std::list<GeneAnnotation>::iterator > find_base(const unsigned int &m_base);
	std::list< std::list<GeneAnnotation>::iterator > find_text(const std::list<std::string> &m_text);
	
	// Search using DNA hybridization
	std::list< std::list<GeneAnnotation>::iterator > find_probe(const DNA3Seq &m_probe, 
		const float &m_salt, const float &m_strand, const float &m_tm, 
		const float &m_identity, std::list<SeqHit> &hit_list);

	std::list< std::list<GeneAnnotation>::iterator > find_primers(const DNA3Seq &m_primer_1, 
		const DNA3Seq &m_primer_2, const float &m_salt, const float &m_strand, 
		const float &m_tm, const unsigned int &m_exact_match, 
		const unsigned int &m_min_amplicon_len, 
		const unsigned int &m_max_amplicon_len,
		const float &m_identity,
		std::list<SeqHit> &hit_list);
	
	std::list< std::list<GeneAnnotation>::iterator > find_MOL_primers(const DNA3Seq &m_primer_1, 
		const DNA3Seq &m_primer_2, const float &m_salt, const float &m_strand, 
		const float &m_tm, const unsigned int &m_exact_match, 
		const float &m_identity,
		std::list<SeqHit> &hit_list);

	// Change from 5'->	3' to 3'->5'
	void invert();

	// Compute primer annotations that pass the specified criteria
	void compute_primers(const unsigned int &m_options, const unsigned int &m_len,
						 const unsigned int &m_run,
						 const float &m_salt, const float &m_strand,
						 const float &m_min_gc, const float &m_max_gc, 
						 const float &m_tm, const float &m_mis_prime_tm);
	
	// Turn all currently selected text into annotation
	void annotate_selection(const std::list<SeqHit> &m_hit_list);

	// Delete all currently selected annotations
	bool delete_annotation();
	
	void export_fasta_genome(std::ofstream &m_fout);
	void export_fasta_dna(std::ofstream &m_fout, 
		const std::list<unsigned int> &m_annot);
	void export_fasta_protein(std::ofstream &m_fout, 
		const std::list<unsigned int> &m_annot);
	void export_gbk(std::ofstream &m_fout, 
		const std::list<unsigned int> &m_annot);
	void export_ptt(std::ofstream &m_fout, 
		const std::list<unsigned int> &m_annot);
};

// Win32 STL implementation requires std::greater for std::list.sort
// I suspect that other implementations use std::less

namespace std{
	template <> 
	struct greater< list<GeneAnnotation>::iterator >
	{
		bool operator()(const list<GeneAnnotation>::iterator &m_a, 
			const list<GeneAnnotation>::iterator &m_b)
		{
			return greater<GeneAnnotation> () (*m_a, *m_b); 
		};
	};
}


#ifdef NOT_USED
namespace std{
	template <> 
	struct std::greater< list<GeneAnnotation>::iterator >
	{
		bool operator()(const list<GeneAnnotation>::iterator &m_a, 
			const list<GeneAnnotation>::iterator &m_b)
		{
			return std::greater<GeneAnnotation> () (*m_a, *m_b); 
		};
	};
}
#endif // NOT_USED

// In annotation.cpp
std::ostream& operator << (std::ostream &s, const GeneAnnotation &m_annot);
int file_type(const std::string &m_filename);
std::string wrap_string(const std::string &m_str, const unsigned int &m_len);
bool integer(const std::string &m_int);
std::string trim_space(const std::string &m_str);
std::string pack_string(const std::string &m_str);

// In annotation_util.cpp
bool read_range(std::ifstream &m_fin, std::pair<unsigned int, unsigned int> &m_range,
	std::list< std::pair<unsigned int, unsigned int> > &m_seg_list);
unsigned int list_to_int(std::list<char> &m_number);
char *error_msg(const char *m_error);

// In genome_translate.cpp
std::string dna_to_protein(DNA3Seq &m_seq, 
	const unsigned int &m_offset, const bool &m_trim_stop);

// In annotation_export.cpp
std::string itoa(const int &m_int);

#endif // __GENOME_ANNOTATION
