// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"
#include "GV.h"
#include "MainFrm.h"
#include "GVDoc.h"
#include "GVView.h"

#include <fstream>

using namespace std;

// Local funtions
void write_char(ofstream &fout, const unsigned char &c, const GLfloat &m_dx);
void write_char_triangle(ofstream &fout, const unsigned char &c, const GLfloat &m_dx);

void CGVView::write_font(const string &m_filename, GLYPHMETRICSFLOAT agmf[], 
		const unsigned char &m_first, const unsigned char &m_last)
{
	try{
		// Assume that no screen dimensions have been set yet
		glViewport(0, 0, 1, 1); // Render each character to the unit square

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		gluOrtho2D(-1.0f, 1.0f, -1.0f, 1.0f);

		// Open the file for writings
		ofstream fout(m_filename.c_str());

		if(!fout){
			throw "Unable to open font file for writing";
		}
		
		// Write the C/C++ source code file 
		fout << "#include \"font_file.h\"" << endl;
		
		fout << endl;

		fout << "void LoadFont(FontMetricGL &m_font)" << endl;
		fout << "{" << endl;

		// Save the font sizes
		unsigned char c;

		for(c = m_first;c <= m_last;c++){
			fout << "\tm_font[" << (int)c << "] = " << agmf[c - m_first].gmfCellIncX << "f;" << endl;
		};

		fout << endl;
		fout << "\tint offset = glGenLists(" << m_last - m_first + 1 << ");" << endl;
		fout << endl;

		for(c = m_first;c <= m_last;c++){
			fout << "\t// '" << c << "'" << endl;

			fout << "\tglNewList(" << c -  m_first << " + offset, GL_COMPILE);" << endl;
			
			// Assume that all character gylphs are composed of triangles.
			// This will allow much simpler code generation that should
			// hopefully speed up compile times
			//write_char(fout, c, agmf[c - m_first].gmfBlackBoxX);
			write_char_triangle(fout, c, agmf[c - m_first].gmfBlackBoxX);

			// Add the offset for the next character (if any)
			fout << "\tglTranslatef(" << agmf[c - m_first].gmfCellIncX << "f, 0.0f, 0.0f);" << endl;
			fout << "\tglEndList();" << endl;

			fout << endl;
		}
		
		fout << "\tglListBase(offset - " << (int)m_first << ");" << endl;

		fout << "}" << endl;

		if(fout){
			fout.close();
		}
	}
	catch(const char *error){
		AfxMessageBox(error);
	}
	catch(...){
		AfxMessageBox("Unhandled exception writing font file");
	}
}

void write_char(ofstream &fout, const unsigned char &c, const GLfloat &m_dx)
{
	const unsigned int feedback_size = 1048576;
	GLfloat *feedback = new GLfloat [feedback_size];

	// Start up the feed back mechanism
	glFeedbackBuffer(feedback_size, GL_2D, feedback);
	glRenderMode(GL_FEEDBACK);

	// Render the character
	glPushMatrix();
		glTranslatef(0.0f, 0.0f, 0.0f);
		glCallLists(1, GL_UNSIGNED_BYTE, &c);
	glPopMatrix();

	int num_obj = glRenderMode(GL_RENDER);
	
	// Was the buffer big enough to render all of the OpenGL objects?
	if(num_obj < 0){
		throw "Feedback buffer to small for writing font file";
	}
	
	const GLfloat *end = feedback + num_obj;
	GLfloat *loc = feedback;

	const GLfloat offset_x = 0.5f;

	// Initialize min and max to values safely out of bounds
	GLfloat min_x = 99999.0f, 
			max_x = -99999.0f;

	// Get the font dimensions
	while (loc < end) {
		int token = (int)*loc;
		loc++;
		if(token != GL_POLYGON_TOKEN){
			throw "Encountered a non GL_POLYGON_TOKEN in feedback buffer!";
		}

		// GL_POLYGON_TOKEN:
		// Since we're using GL_2D for the feedback, there are only nvertices*2 values to 
		// store
		int nvertices = (int)*loc;
		loc++;

		for(unsigned int i = 0;i < nvertices;i++){
			GLfloat tmp_x = loc[0] - offset_x;

			if(min_x > tmp_x){
				min_x = tmp_x;
			}

			if(max_x < tmp_x){
				max_x = tmp_x;
			}

			loc += 2;
		}
	}

	GLfloat font_scale = m_dx/(max_x - min_x);

	loc = feedback;

	// Write the font polygons as a call list
	while (loc < end) {
		int token = (int)*loc;
		loc++;
		if(token != GL_POLYGON_TOKEN){
			throw "Encountered a non GL_POLYGON_TOKEN in feedback buffer!";
		}

		// GL_POLYGON_TOKEN:
		// Since we're using GL_2D for the feedback, there are only nvertices*2 values to 
		// store
		int nvertices = (int)*loc;
		loc++;

		switch(nvertices){
			case 3:
				fout << "\tglBegin(GL_TRIANGLES);" << endl;
				break;
			case 4:
				fout << "\tglBegin(GL_QUADS);" << endl;
				break;
			default:
				fout << "\tglBegin(GL_POLYGON);" << endl;
				break;
		};

		for(unsigned int i = 0;i < nvertices;i++){
			fout << "\t\tglVertex2f(" << (loc[0] - offset_x )*font_scale
				 << ", " << (loc[1] - offset_x )*font_scale
				 << ");" << endl;

			loc += 2;
		}

		fout << "\tglEnd();" << endl;
	}

	if(feedback){
		delete [] feedback;
	}
}

void write_char_triangle(ofstream &fout, const unsigned char &c, const GLfloat &m_dx)
{
	const unsigned int feedback_size = 1048576;
	GLfloat *feedback = new GLfloat [feedback_size];

	// Start up the feed back mechanism
	glFeedbackBuffer(feedback_size, GL_2D, feedback);
	glRenderMode(GL_FEEDBACK);

	// Render the character
	glPushMatrix();
		glTranslatef(0.0f, 0.0f, 0.0f);
		glCallLists(1, GL_UNSIGNED_BYTE, &c);
	glPopMatrix();

	int num_obj = glRenderMode(GL_RENDER);
	
	// Was the buffer big enough to render all of the OpenGL objects?
	if(num_obj < 0){
		throw "Feedback buffer to small for writing font file";
	}
	
	const GLfloat *end = feedback + num_obj;
	GLfloat *loc = feedback;

	const GLfloat offset_x = 0.5f;

	// Initialize min and max to values safely out of bounds
	GLfloat min_x = 99999.0f, 
			max_x = -99999.0f;

	// Get the font dimensions
	while (loc < end) {
		int token = (int)*loc;
		loc++;
		if(token != GL_POLYGON_TOKEN){
			throw "Encountered a non GL_POLYGON_TOKEN in feedback buffer!";
		}

		// GL_POLYGON_TOKEN:
		// Since we're using GL_2D for the feedback, there are only nvertices*2 values to 
		// store
		int nvertices = (int)*loc;
		loc++;

		for(unsigned int i = 0;i < nvertices;i++){
			GLfloat tmp_x = loc[0] - offset_x;

			if(min_x > tmp_x){
				min_x = tmp_x;
			}

			if(max_x < tmp_x){
				max_x = tmp_x;
			}

			loc += 2;
		}
	}

	GLfloat font_scale = m_dx/(max_x - min_x);

	loc = feedback;

	// All tokens are assumed to be triangles
	fout << "\tglBegin(GL_TRIANGLES);" << endl;

	// Write the font polygons as a call list
	while (loc < end) {
		int token = (int)*loc;
		loc++;
		if(token != GL_POLYGON_TOKEN){
			throw "Encountered a non GL_POLYGON_TOKEN in feedback buffer!";
		}

		// GL_POLYGON_TOKEN:
		// Since we're using GL_2D for the feedback, there are only nvertices*2 values to 
		// store
		int nvertices = (int)*loc;
		loc++;

		// Assume that all gylphs are rendered with triangles. If this
		// assumption is not true, throw an error
		if(nvertices != 3){
			throw __FILE__ ":write_char_triangle: Non-triangle polygon encountered";
		}

		for(unsigned int i = 0;i < nvertices;i++){
			fout << "\t\tglVertex2f(" << (loc[0] - offset_x )*font_scale
				 << ", " << (loc[1] - offset_x )*font_scale
				 << ");" << endl;

			loc += 2;
		}
	}

	fout << "\tglEnd();" << endl;

	if(feedback){
		delete [] feedback;
	}
}
