#ifndef __NUCLEIC_CRUCIBLE
#define __NUCLEIC_CRUCIBLE


// A class interface for computing nucleic acid melting temperatures.
// The functional form for Tm is based on the work of:
//
// Owczary et. al., Biopoly 44: 217-239, 1997
// Allawi and SantaLucia, Biochemistry 36: 10581-10594, 1997
// SantaLucia, PNAS 95: 1460-1465
//
// The Tm currently depends on the following:
// 1) Sequence composition
// 2) [Na+] (salt concentration)
// 3) Sequence length (via [Na+] correction)
// 4) [C_T] (total concentration of DNA single strands)
//
// Tm = (dH_d + dH_n + dH_s)/(Rln{[C_T]/alpha} + 
//	dS_d + dS_n + dS_s + dS_salt)
//
// dH_d == Di-nucleotide Enthalpy
// dH_n == Nucleation Enthalpy
// dH_s == Symmetry Enthalpy
// dS_d == Di-nucleotide Entropy
// dS_n == Nucleation Entropy
// dS_s == Symmetry Entropy
// R ==  The Gas constant
// alpha == self-complementarity correction
//

// Physical constants
#define	NC_ZERO_C	(273.15f) 		// K
#define	NC_R		(1.9872e-3f) 		// Kcal/(Mol . K) -- McQuarrie Stat. Mech.

// Default parameters
#define	NC_CT		(1.0e-7f)		// Mol <-- Total strand concentration
#define	NC_Na		(0.2f)		// Mol <-- Monovalent cation

#include <list>
#include <string>

#include "circle_buffer.h"

#define	NUM_BASE_PAIR		25

#define	CBUFFER_SIZE		128

class NucCruc{
	public:
		// The GAP symbol must be the first non-ATGC base
		// (for tests that use >= GAP for valid bases)
		// Note!!! Don't change the order of symbols in this enum ---
		// the complement function depends on this particular order!
		enum{A = 0, C, G, T, GAP, UNKNOWN};
		
		enum{SANTA_LUCIA = 0, PRIMER3};
	private:
		
		static float param_H[NUM_BASE_PAIR][NUM_BASE_PAIR];
		static float param_S[NUM_BASE_PAIR][NUM_BASE_PAIR];
		
		static float param_nucleation_H[NUM_BASE_PAIR];
		static float param_nucleation_S[NUM_BASE_PAIR];
		
		static bool param_init;
		static unsigned int parameter_set;
		
		float na_concentration;
		float strand_concentration;
		
		// Below what identity should we NOT bother computing
		// the hybridization Tm?
		float identity_cutoff;

		// What are the minimum and maximum oligo length?
		unsigned int min_len;
		unsigned int max_len;
		unsigned int hairpin_loop_size;
		
		// An internal duplex buffer for efficient computation of Melting
		std::list< std::pair<unsigned char, unsigned char> > duplex_buffer;
		
		// Internal buffers for reference and target sequences
		CircleBuffer<unsigned char, CBUFFER_SIZE> reference_buffer;
		CircleBuffer<unsigned char, CBUFFER_SIZE> target_buffer;
		
		// Initialize the free energy parameters
		void init_param_AS();
		void init_param_Primer3();
		
		inline unsigned char CharToBase(const char &m_base) const
		{
			// What kind of base is this?
			switch(m_base){
				case 'A':
				case 'a':
					return A;
				case 'C':
				case 'c':
					return C;
				case 'G':
				case 'g':
					return G;
				case 'T':
				case 't':
					return T;
				case '-':
					return GAP;
				default:
					return UNKNOWN;
			};

		};
		
		inline unsigned char CharToBaseComplement(const char &m_base) const
		{
			// What kind of base is this?
			switch(m_base){
				case 'A':
				case 'a':
					return T;
				case 'C':
				case 'c':
					return G;
				case 'G':
				case 'g':
					return C;
				case 'T':
				case 't':
					return A;
				case '-':
					return GAP;
				default:
					return UNKNOWN;
			};

		};
		
		inline unsigned char complement(const unsigned char &m_base) const
		{
			return (m_base > T) ? m_base : T - m_base;
		};

		// Use the internal duplex buffer for the dimer/hairpin 
		// calculation. Note that this function will overwrite the 
		// reference and target buffers!!!
		// m_configuration controls the dangling base
		// topology:
		// m_configuration == true ... compute hairpin
		// m_configuration == false ... compute dimer
		float MaxDimerTm_duplex(const bool &m_configuration);

	public:
	
	#ifdef _USE_PRIMER3
	NucCruc(const unsigned int &m_param_set = PRIMER3);
	#else
	NucCruc(const unsigned int &m_param_set = SANTA_LUCIA);
	#endif //_USE_PRIMER3
	
	inline float Salt() const
	{
		return na_concentration;
	};
	
	inline void Salt(const float &m_na_concentration)
	{
		na_concentration = m_na_concentration;
	};
	
	inline float Strand() const
	{
		return strand_concentration;
	};
	
	inline void Strand(const float &m_strand_concentration)
	{
		strand_concentration = m_strand_concentration;
	};
	
	inline float IdentityCutoff() const
	{
		return identity_cutoff;
	};

	inline void IdentityCutoff(const float &m_cut_off)
	{
		if(m_cut_off > 1.0f){
			throw "Please use a Tm identity < 1.0";
		}

		identity_cutoff = m_cut_off;
	};

	inline bool is_exact_match(const unsigned int &m_base_pair) const;

	bool is_exact_match(const unsigned char &m_base_1, const unsigned char &m_base_2) const
	{
		return ( (m_base_1 > T) || (m_base_2 > T) ) ? 
			(m_base_1 != m_base_2) : // There is no GAP-GAP interaction defined
			complement(m_base_1) == m_base_2;
	};

	// Here is where all the work is done!
	float Tm(const std::list< std::pair<unsigned char, unsigned char> > &m_duplex,
		std::pair<unsigned int, unsigned int> *m_align,
		const float &m_DH_bias, const float &m_DS_bias) const;
			
	// Wrappers to process user input formats into vectors of base-pairs
	// that can be passed to the private Tm() function
	float Tm(const std::string &m_seq,
		std::pair<unsigned int, unsigned int> *m_align = NULL,
		const float &m_DH_bias = 0.0f, const float &m_DS_bias = 0.0f) const;
	
	float Tm(const std::string &m_seq_a, const std::string &m_seq_b, 
		const bool &m_complement_strand_b, 
		std::pair<unsigned int, unsigned int> *m_align = NULL,
		const float &m_DH_bias = 0.0f, const float &m_DS_bias = 0.0f) const;
	
	// How many 3' bases of m_seq_a, exactly match m_seq_b?
	unsigned int anchor(const std::string &m_seq_a, const std::string &m_seq_b, 
		const bool &m_complement_strand_b) const;
	
	// Use the internal duplex buffer
	inline float Tm(std::pair<unsigned int, unsigned int> *m_align = NULL,
		const float &m_DH_bias = 0.0f, const float &m_DS_bias = 0.0f) const
	{
		unsigned int len = duplex_buffer.size();
	
		if(len < min_len){
			throw "NucCruc::Tm: Duplex buffer is less than minimum length";
		}

		if(len > max_len){
			throw "NucCruc::Tm: Duplex buffer is greater than maximum length";
		}

		return Tm(duplex_buffer, m_align, m_DH_bias, m_DS_bias);
	};
	
	// Use the internal reference/target buffer
	float Tm_reference(const float &m_DH_bias = 0.0f, const float &m_DS_bias = 0.0f) const;
		
	float MaxDimerTm(const std::string &m_seq) const;
	float MaxDimerTm(const std::string &m_seq_a, const std::string &m_seq_b, 
		const bool &m_reverse_strand_b) const;
		
	float MaxHairPinTm(const std::string &m_seq) const;

	// Use the first strand of the internal duplex buffer. These functions
	// overwrite the interenal reference and target buffers
	float MaxHairPinTm();
	float MaxDimerTm();

	// Manipulate the internal duplex buffer
	inline void clear()
	{
		duplex_buffer.clear();
	};
	
	inline void pop_front()
	{
		duplex_buffer.pop_front();
	};
	
	inline void pop_back()
	{
		duplex_buffer.pop_back();
	};
	
	inline unsigned int size() const
	{
		return duplex_buffer.size();
	};
	
	inline bool empty() const
	{
		return duplex_buffer.empty();
	};
	
	inline void push_back(const char &m_a, const char &m_b, const bool &m_complement_b)
	{
		unsigned char base_a, base_b;
		
		if(m_complement_b){
			base_a = CharToBase(m_a);
			base_b = CharToBaseComplement(m_b);
			
			if((base_a > GAP) || (base_b > GAP)){
				throw "NucCruc::Tm: Illegal base!";
			}
			
			duplex_buffer.push_back(std::make_pair(base_a, base_b));
		}
		else{
			base_a = CharToBase(m_a);
			base_b = CharToBase(m_b);
			
			if((base_a > GAP) || (base_b > GAP)){
				throw "NucCruc::Tm: Illegal base!";
			}
			
			duplex_buffer.push_back(std::make_pair(base_a, base_b));
		}
	};
	
	inline void push_back(const char &m_base)
	{	
		const unsigned char base_a = CharToBase(m_base);

		if(base_a > GAP){
			throw "NucCruc::Tm: Illegal base!";
		}

		const unsigned char base_b = complement(base_a);
			
		duplex_buffer.push_back(std::make_pair(base_a, base_b));
	};

	inline void push_back_native(const unsigned char &m_base)
	{	
		if(m_base > GAP){
			throw "NucCruc::Tm: Illegal base!";
		}
	
		duplex_buffer.push_back(std::make_pair(m_base, complement(m_base)));
	};

	// Use a copy of m_base_b in case we need to take the complement
	inline void push_back_native(const unsigned char &m_base_a, unsigned char m_base_b, 
		const bool &m_complement_b)
	{
		if(m_complement_b){
			m_base_b = complement(m_base_b);
		}

		if((m_base_a > GAP) || (m_base_b > GAP)){
			throw "NucCruc::Tm: Illegal base!";
		}
			
		duplex_buffer.push_back(std::make_pair(m_base_a, m_base_b));
	};

	inline void push_front(const char &m_a, const char &m_b, const bool &m_complement_b)
	{
		unsigned char base_a, base_b;
		
		if(m_complement_b){
			base_a = CharToBase(m_a);
			base_b = CharToBaseComplement(m_b);
			
			if((base_a > GAP) || (base_b > GAP)){
				throw "NucCruc::Tm: Illegal base!";
			}
			
			duplex_buffer.push_front(std::make_pair(base_a, base_b));
		}
		else{
			base_a = CharToBase(m_a);
			base_b = CharToBase(m_b);
			
			if((base_a > GAP) || (base_b > GAP)){
				throw "NucCruc::Tm: Illegal base!";
			}
			
			duplex_buffer.push_front(std::make_pair(base_a, base_b));
		}
	};
	
	inline void push_front(const char &m_base)
	{		
		const unsigned char base_a = CharToBase(m_base);

		if(base_a > GAP){
			throw "NucCruc::Tm: Illegal base!";
		}

		const unsigned char base_b = complement(base_a);
			
		duplex_buffer.push_front(std::make_pair(base_a, base_b));
	};

	inline void push_front_native(const unsigned char &m_base)
	{		
		if(m_base > GAP){
			throw "NucCruc::Tm: Illegal base!";
		}
			
		duplex_buffer.push_front(std::make_pair(m_base, complement(m_base)));
	};

	// Use a copy of m_base_b in case we need to take the complement
	inline void push_front_native(const unsigned char &m_base_a, unsigned char m_base_b, 
		const bool &m_complement_b)
	{		
		if(m_complement_b){
			m_base_b = complement(m_base_b);
		}
	
		if((m_base_a > GAP) || (m_base_b > GAP)){
			throw "NucCruc::Tm: Illegal base!";
		}
			
		duplex_buffer.push_front(std::make_pair(m_base_a, m_base_b));
	};

	// Manipulate the internal reference and target buffers. These functions are
	// here to make it easy to "convolve" a reference sequence against a target
	// sequence
	inline void clear_target()
	{
		target_buffer.clear();
	};
	
	inline void reference(const std::string &m_seq)
	{
		unsigned int len = m_seq.size();
		
		// Set the reference sequence
		reference_buffer.clear();
		
		for(unsigned int i = 0;i < len;i++){
			unsigned char base = CharToBase(m_seq[i]);

			if(base >= GAP){
				throw "NucCruc::Tm: Illegal base!";
			}

			reference_buffer.push_back(base);
		}
	};
	
	inline void pop_front_target()
	{
		target_buffer.pop_front();
	};
	
	inline void pop_back_target()
	{
		target_buffer.pop_back();
	};
	
	inline unsigned int size_reference() const
	{
		return reference_buffer.size();
	};
	
	inline unsigned int size_target() const
	{
		return target_buffer.size();
	};
	
	inline bool empty_target() const
	{
		return target_buffer.empty();
	};
	
	inline void push_back_target(const char &m_a, const bool &m_complement_b)
	{
		if(m_complement_b){
			target_buffer.push_back(CharToBaseComplement(m_a));
		}
		else{
			target_buffer.push_back(CharToBase(m_a));
		}
	};
	
	inline void push_back_target_native(const unsigned char &m_a, const bool &m_complement_b)
	{
		if(m_complement_b){
			target_buffer.push_back(complement(m_a));
		}
		else{
			target_buffer.push_back(m_a);
		}
	};

	inline void push_front_target(const char &m_a, const bool &m_complement_b)
	{
		if(m_complement_b){
			target_buffer.push_front(CharToBaseComplement(m_a));
		}
		else{
			target_buffer.push_front(CharToBase(m_a));
		}
	};
	
	inline void push_front_target_native(const unsigned char &m_a, const bool &m_complement_b)
	{
		if(m_complement_b){
			target_buffer.push_front(complement(m_a));
		}
		else{
			target_buffer.push_front(m_a);
		}
	};

	// How many 3' bases of the reference, exactly match the target?
	unsigned int anchor_reference() const;
	unsigned int anchor3_reference() const;
	
	// How many 5' bases of the reference, exactly match the target?
	unsigned int anchor5_reference() const;
	
	// Do at LEAST m_anchor bases of the reference match the target
	// (this should be more efficient than anchor_reference() is only a
	// yes/no answer is required).
	bool anchor_reference(const unsigned int &m_anchor) const;

	friend std::ostream& operator << (std::ostream &s, const NucCruc &m_melt);
	
	// Compute the fractional GC content of an external sequence
	inline float gc_content(const std::string &m_seq) const
	{
		unsigned int len = m_seq.size();
		float gc = 0.0f;

		for(unsigned int i = 0;i < len;i++){
			switch(m_seq[i]){
				case 'G':
				case 'g':
				case 'C':
				case 'c':
					gc ++;
					break;
			};
		}

		return (gc/len);
	};
	
	// Compute the fractional GC content of the internal duplex buffer
	float gc_content() const;
	
	// Compute the fractional GC content of the internal reference buffer
	float gc_content_reference() const;
	
	// The fractional identity of the sequences stored in the reference
	// buffers
	float identity_reference() const;
	
	// The number of Watson and Crick matches between the sequences stored in the
	// reference buffers
	unsigned int num_identity_reference() const;
};

// A helper function
std::string complement(const std::string &m_seq);

#endif // __NUCLEIC_CRUCIBLE

