// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include <string>
#include <sstream>
#include "primer.h"

using namespace std;

int PCRPrimer::operator() (const CircleBuffer<unsigned char, MAX_PCR_LENGTH> &m_primer) const
{
	int status = PCR_VALID;
	
	if( m_primer.empty() ){
		throw "PCRPrimer::operator(): Empty primer buffer";
	}

	// Is this the last base?
	if( (mask & NO_3_T) && (m_primer.back() == T) ){
		if(report_verbose){
			status |= NO_3_T;
		}
		else{
			return NO_3_T;
		}
	}
	
	// Is this the first base?
	if( (mask & NO_5_G) && (m_primer.front() == G) ){
		if(report_verbose){
			status |= NO_5_G;
		}
		else{
			return NO_5_G;
		}
	}
	
	const unsigned int len = m_primer.size();

	CircleBuffer<unsigned char, MAX_PCR_LENGTH>::const_iterator iter;
	
	unsigned int TC_run = 0;
	unsigned int AG_run = 0;
	unsigned int G_run = 0;
	unsigned int C_run = 0;
	unsigned int max_TC_run = 0;
	unsigned int max_AG_run = 0;
	unsigned int num_GC = 0;

	unsigned int index = 0;

	// The list must be packed from 5' -> 3'
	for(iter = m_primer.begin();iter != m_primer.end();iter++, index++){
		switch(*iter){
			case A:
				AG_run ++;
				TC_run = 0;
				G_run = 0;
				C_run = 0;
				break;
			case T:
				TC_run++;
				AG_run = 0;
				G_run = 0;
				C_run = 0;
				break;
			case G:
				AG_run ++;
				G_run ++;
				TC_run = 0;
				C_run = 0;
				num_GC ++;
				break;
			case C:
				TC_run ++;
				C_run ++;
				AG_run = 0;
				G_run = 0;
				num_GC ++;
				break;
		};

		if(max_TC_run < TC_run){
			max_TC_run = TC_run;
		}

		if(max_AG_run < AG_run){
			max_AG_run = AG_run;
		}

		if( (mask & MULTI_5_GC) && (index == 4) ){
			if(max_TC_run >= 2){
				if(num_GC > 2){
					if(report_verbose){
						status |= MULTI_5_GC;
					}
					else{
						return MULTI_5_GC;
					}
				}
			}
			else{
				if(num_GC > 3){
					if(report_verbose){
						status |= MULTI_5_GC;
					}
					else{
						return MULTI_5_GC;
					}
				}
			}
		}
	}

	if(mask & NO_POLY_RUNS){
		if(max_TC_run >= run_len){
			if(report_verbose){
				status |= NO_POLY_RUNS;
			}
			else{
				return NO_POLY_RUNS;
			}
		}

		if(max_AG_run >= run_len){
			if(report_verbose){
				status |= NO_POLY_RUNS;
			}
			else{
				return NO_POLY_RUNS;
			}
		}
	}

	if(mask & POLY_3_GC){
		if(C_run >= 3){
			if(report_verbose){
				status |= POLY_3_GC;
			}
			else{
				return POLY_3_GC;
			}
		}

		if(G_run >= 3){
			if(report_verbose){
				status |= POLY_3_GC;
			}
			else{
				return POLY_3_GC;
			}
		}
	}

	if(mask & GC_CONTENT){
		const float gc_content = (float)(num_GC)/len;

		if(gc_content < gc_min){
			if(report_verbose){
				status |= GC_CONTENT;
			}
			else{
				return GC_CONTENT;
			}
		}

		if(gc_content > gc_max){
			if(report_verbose){
				status |= GC_CONTENT;
			}
			else{
				return GC_CONTENT;
			}
		}
	}

	// If we get here, this primer is valid!
	return status;
}

int PCRPrimer::operator() (const string &m_primer) const
{
	CircleBuffer<unsigned char, MAX_PCR_LENGTH> primer_list;

	const unsigned int len = m_primer.size();

	for(unsigned int i = 0;i < len;i++){
		switch(m_primer[i]){
			case 'A':
			case 'a':
				primer_list.push_back(A);
				break;
			case 'T':
			case 't':
				primer_list.push_back(T);
				break;
			case 'G':
			case 'g':
				primer_list.push_back(G);
				break;
			case 'C':
			case 'c':
				primer_list.push_back(C);
				break;
			default:
				return BAD_BASE;
		};
	}

	// If we get here, this primer is valid!
	return (*this)(primer_list);
}

int PCRPrimer::operator() () const
{
	return (*this)(primer_buffer);
}


string PCRPrimer::str() const
{
	string primer_str;

	primer_str.resize( primer_buffer.size() );

	CircleBuffer<unsigned char, MAX_PCR_LENGTH>::const_iterator iter;

	unsigned int index = 0;

	// The list must be packed from 5' -> 3'
	for(iter = primer_buffer.begin();iter != primer_buffer.end();iter++, index++){
		switch(*iter){
			case A:
				primer_str[index] = 'A';
				break;
			case T:
				primer_str[index] = 'T';
				break;
			case G:
				primer_str[index] = 'G';
				break;
			case C:
				primer_str[index] = 'C';
				break;
		};
	}
	
	return primer_str;
}


string  PCRPrimer::error(const int &m_code) const
{
	stringstream sout;
	bool delim = false;
	
	if(m_code == PCR_VALID){
		return "PCR_VALID";
	}
	
	if(m_code == BAD_BASE){
		return "BAD_BASE";
	}
	
	if(m_code & POLY_3_GC){
		sout << "POLY_3_GC";
		delim = true;
	}
		
	if(m_code & MULTI_5_GC){
		if(delim){
			sout << ", ";
		}
		
		sout << "MULTI_5_GC";
		delim = true;
	}
	
	if(m_code & NO_POLY_RUNS){
		if(delim){
			sout << ", ";
		}
		
		sout << "NO_POLY_RUNS";
		delim = true;
	}
	
	if(m_code & NO_3_T){
		if(delim){
			sout << ", ";
		}
		
		sout << "NO_3_T";
		delim = true;
	}
	
	if(m_code & GC_CONTENT){
		if(delim){
			sout << ", ";
		}
		
		sout << "GC_CONTENT";
		delim = true;
	}
	
	return sout.str();
}
