// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "genome.h"
#include "nuc_cruc.h"
#include "univout.h"

#include <algorithm>
#include <sstream>

using namespace std;

// Local functions

// The NCBIErrHookProc is required to handle an NCBI errors (the default
// error handler terminates the program!).
static int LIBCALLBACK
NCBIErrHookProc(const ErrDesc *err)
{
	// Note that throwing an error here is not portable -- it works for
	// WIN32 but not OS X. Simply return a non-zero value.

	//throw "NCBIErrHookProc: An NCBI library error has occured";

	return 1;
}

Genome::Genome(const unsigned int &m_mode)
{
	mol.mode(m_mode);

	is_shown = true;
	is_annotation_shown = true;
	is_selected = false;
	is_blending = false;

	display_gc_graph = false;
	display_at_graph = false;
	display_gc_skew_graph = false;
	display_at_skew_graph = false;
	display_keto_skew_graph = false;
	display_purine_skew_graph = false;

	display_tm_graph = false;
	display_genome_plot = false;
	
	// Start off with the default font scale
	text_font_scale_factor = 1.0f;

	genome_plot_bounds_y = pair<float, float>(0.0f, 0.0f);
	sort_rank = 0.0f;
	base_center = 0;
	num_base = 0;
	
	shift_buffer = 0.0f;

	// Total window size is 2*content_window + 1
	content_window = 50;

	// Total window size is 2*tm_window + 1
	tm_window = 8;

	salt = NC_Na;	// M
	strand = NC_CT;	// M

	// There are no search results yet
	cur_search_iter = search_results.end();
	cur_search_hit_iter = hit_list.end();
	
	// Start with auto-picking disabled
	clear_auto_pick();
	
	// Set initial color choices:
	select_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_SELECT_R);
	select_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_SELECT_G);
	select_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_SELECT_B);
	select_color[3] = 1.0f;

	// The background color must be kept in sync with
	// GenomePage class.
	background_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_BACKGROUND_R);
	background_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_BACKGROUND_G);
	background_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_BACKGROUND_B);
	background_color[3] = 1.0f;

	gene_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_GENE_R);
	gene_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_GENE_G);
	gene_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_GENE_B);
	gene_color[3] = 1.0f;

	pseudo_gene_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_PSEUDO_GENE_R);
	pseudo_gene_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_PSEUDO_GENE_G);
	pseudo_gene_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_PSEUDO_GENE_B);
	pseudo_gene_color[3] = 1.0f;

	cds_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_CDS_R);
	cds_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_CDS_G);
	cds_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_CDS_B);
	cds_color[3] = 1.0f;

	annot_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_ANNOT_R);
	annot_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_ANNOT_G);
	annot_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_ANNOT_B);
	annot_color[3] = 1.0f;

	primer_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_PRIMER_R);
	primer_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_PRIMER_G);
	primer_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_PRIMER_B);
	primer_color[3] = 1.0f;

	user_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_USER_R);
	user_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_USER_G);
	user_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_USER_B);
	user_color[3] = 1.0f;

	rna_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_RNA_R);
	rna_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_RNA_G);
	rna_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_RNA_B);
	rna_color[3] = 1.0f;

	intergenic_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_INTERGENIC_R);
	intergenic_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_INTERGENIC_G);
	intergenic_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_INTERGENIC_B);
	intergenic_color[3] = 1.0f;

	text_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_TEXT_R);
	text_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_TEXT_G);
	text_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_TEXT_B);
	text_color[3] = 1.0f;

	select_neg_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_SELECT_NEG_R);
	select_neg_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_SELECT_NEG_G);
	select_neg_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_SELECT_NEG_B);
	select_neg_color[3] = 1.0f;

	select_plus_color[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_SELECT_PLUS_R);
	select_plus_color[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_SELECT_PLUS_G);
	select_plus_color[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_SELECT_PLUS_B);
	select_plus_color[3] = 1.0f;
}

bool Genome::load(const std::string &m_filename, const unsigned int m_type, streampos &m_pos)
{
	bool ret = mol.load(m_filename, m_type, m_pos);

	// Save the total number of bases
	num_base = mol.sequence().size();

	// Compute the maximum depths
	list<GeneAnnotation>::iterator iter;
	
	max_gene_depth = max_annot_depth = 0;

	for(iter = mol.begin();iter != mol.end();iter++){
		if(max_gene_depth < iter->depth()){
			max_gene_depth = iter->depth();
		}

		if(max_annot_depth < iter->annot_depth()){
			max_annot_depth = iter->annot_depth();
		}
	}
	
	clear_auto_pick();
	
	return ret;
}

void Genome::load(const int &m_gi)
{
	mol.load(m_gi);

	// Save the total number of bases
	num_base = mol.sequence().size();

	// Compute the maximum depths
	list<GeneAnnotation>::iterator iter;
	
	max_gene_depth = max_annot_depth = 0;

	for(iter = mol.begin();iter != mol.end();iter++){
		if(max_gene_depth < iter->depth()){
			max_gene_depth = iter->depth();
		}

		if(max_annot_depth < iter->annot_depth()){
			max_annot_depth = iter->annot_depth();
		}
	}
	
	clear_auto_pick();
}

void Genome::load_and_save(const int &m_gi, const string &m_filename)
{
	mol.load_and_save(m_gi, m_filename);

	// Save the total number of bases
	num_base = mol.sequence().size();
	
	// Compute the maximum depths
	list<GeneAnnotation>::iterator iter;
	
	max_gene_depth = max_annot_depth = 0;

	for(iter = mol.begin();iter != mol.end();iter++){
	
		if(max_gene_depth < iter->depth()){
			max_gene_depth = iter->depth();
		}

		if(max_annot_depth < iter->annot_depth()){
			max_annot_depth = iter->annot_depth();
		}
	}
	
	clear_auto_pick();
}

bool Genome::import(const Genome &m_genome, const unsigned int &m_type)
{

	if( mol.import(m_genome.mol, m_type) == false){
		return false;
	}
	
	// Re-compute the maximum depths
	list<GeneAnnotation>::iterator iter;
	
	max_gene_depth = max_annot_depth = 0;

	for(iter = mol.begin();iter != mol.end();iter++){
		if( max_gene_depth < iter->depth() ){
			max_gene_depth = iter->depth();
		}

		if( max_annot_depth < iter->annot_depth() ){
			max_annot_depth = iter->annot_depth();
		}
	}
	
	clear_auto_pick();
	
	return true;
}

string Genome::select_info()
{
	list<GeneAnnotation>::iterator iter;
	stringstream s;

	for(iter = mol.begin();iter != mol.end();iter++){
	
		if( iter->select() ){

			s << make_pair(seq_size(), *iter) << ENDL;

			// Compute the base composition statistics as well
			// as the user plot statisitics
			s << base_composition_info(*iter);
			
			if(display_genome_plot == true){
			
				s << genome_plot_info(*iter);
			}
			
			return s.str();
		}
	}

	// If we get here, there we no selected annotations, so display
	// info about the enitre genome
	s << mol;
	
	if(mol.circular()){
		s << "Circular molecule with " << seq_size() << " bases" << ENDL;
	}
	else{
		s << "Linear molecule with " << seq_size() << " bases" << ENDL;
	}
	
	s << "%GC = " << 100.0f*mol.gc() << ENDL;
	s << "%A = " << 100.0f*mol.a() << ENDL;
	s << "%T = " << 100.0f*mol.t() << ENDL;
	s << "%G = " << 100.0f*mol.g() << ENDL;
	s << "%C = " << 100.0f*mol.c() << ENDL;
	s << "%other = " << 100.0f*mol.other() << ENDL;

	return s.str();
}

string Genome::base_composition_info(GeneAnnotation &m_annot)
{
	stringstream s;
	
	float annot_gc = 0.0f;
	float annot_a = 0.0f;
	float annot_t = 0.0f;
	float annot_g = 0.0f;
	float annot_c = 0.0f;
	float annot_other = 0.0f;
	
	const unsigned int annot_length = m_annot.length( seq_size() );
	  
	DNA3Seq::iterator iter = mol.sequence().begin() + m_annot.start();
	
	for(unsigned int base_index = 0;base_index < annot_length;base_index++){
	
		switch( iter.base() ){
			case DNA3Seq::A:
				annot_a ++;
				break;
			case DNA3Seq::T:
				annot_t ++;
				break;
			case DNA3Seq::G:
				annot_g ++;
				annot_gc ++;
				break;
			case DNA3Seq::C:
				annot_c ++;
				annot_gc ++;
				break;
			case DNA3Seq::UNKNOWN:
				annot_other ++;
				break;
		};

		iter++;

		if( iter == mol.sequence().end() ){
			iter = mol.sequence().begin();
		}
	}
	
	annot_a /= annot_length;
	annot_t /= annot_length;
	annot_g /= annot_length;
	annot_c /= annot_length;
	annot_other /= annot_length;
	annot_gc /= annot_length;
	
	s << "%GC = " << 100.0f*annot_gc << ENDL;
	s << "%A = " << 100.0f*annot_a << ENDL;
	s << "%T = " << 100.0f*annot_t << ENDL;
	s << "%G = " << 100.0f*annot_g << ENDL;
	s << "%C = " << 100.0f*annot_c << ENDL;
	s << "%other = " << 100.0f*annot_other << ENDL;
	
	return s.str();
}

// Compute genome plot statistics for a linearly interpolated genome plot
string Genome::genome_plot_info(GeneAnnotation &m_annot)
{
	stringstream s;
	
	float ave_plot = 0.0f;
	float max_plot = 0.0f;
	float min_plot = 0.0f;
	
	const unsigned int plot_start = m_annot.start();
	const unsigned int plot_stop = m_annot.stop();
	unsigned int count = 0;
	
	vector< pair<unsigned int, float> >::const_iterator last_iter = genome_plot.begin();
	vector< pair<unsigned int, float> >::const_iterator curr_iter = last_iter;
	
	curr_iter ++;
	
	if(plot_start > plot_stop){
	
		while( curr_iter != genome_plot.end() ){
			
			// Only include those points that fall within the annotation
			if( ( (last_iter->first > plot_stop) && (last_iter->first < plot_start) ) &&
			   ( (curr_iter->first > plot_stop) && (curr_iter->first < plot_start) ) ){
				
				last_iter = curr_iter;
				curr_iter ++;

				continue;
			}
			
			const float dx = curr_iter->first - last_iter->first;
			const float inv_dx = (dx == 0.0f) ? 0.0f : 1.0f/dx;
			const float delta = curr_iter->second - last_iter->second;
			
			for(unsigned int i = last_iter->first;i < curr_iter->first;i++){
			
				if( (i > plot_stop) && (i < plot_start) ){
					continue;
				}
				
				const float value = last_iter->second + delta*inv_dx*(i - last_iter->first);
				
				ave_plot += value;
			
				if(count == 0){
					max_plot = value;
					min_plot = value;
				}
				else{
					max_plot = max(max_plot, value);
					min_plot = min(min_plot, value);
				}
				
				count ++;
			}
			
			last_iter = curr_iter;
			curr_iter ++;
		}
	}
	else{ // plot_stop > plot_start
	
		while( curr_iter != genome_plot.end() ){
			
			// Only include those points that fall within the annotation
			if( (last_iter->first < plot_start) && (curr_iter->first < plot_start) ){
	
				last_iter = curr_iter;
				curr_iter ++;
				continue;
			}
			
			if( (last_iter->first > plot_stop) && (curr_iter->first > plot_stop) ){
				break;
			}
			
			const float dx = curr_iter->first - last_iter->first;
			const float inv_dx = (dx == 0.0f) ? 0.0f : 1.0f/dx;
			const float delta = curr_iter->second - last_iter->second;
			
			for(unsigned int i = last_iter->first;i < curr_iter->first;i++){
			
				if( (i > plot_stop) && (i < plot_start) ){
					continue;
				}
				
				const float value = last_iter->second + delta*inv_dx*(i - last_iter->first);
				
				ave_plot += value;
			
				if(count == 0){
					max_plot = value;
					min_plot = value;
				}
				else{
					max_plot = max(max_plot, value);
					min_plot = min(min_plot, value);
				}
				
				count ++;
			}
			
			last_iter = curr_iter;
			curr_iter ++;
		}
	}
	
	if(count > 0){
		ave_plot /= count;
	}
	
	s << "<plot> = " << ave_plot << ENDL;
	s << "max(plot) = " << max_plot << ENDL;
	s << "min(plot) = " << min_plot << ENDL;
	
	return s.str();
}

// Change from 5'->3' to 3'->5'
void Genome::complement()
{
	// Any data structure that contains base coordinates
	// needs to be updated!
	mol.invert();

	// Transform the view to the inverted coordinates.
	base_center = num_base - 1 - base_center;

	// Invert the genome plot (if present)
	if(genome_plot.empty() == false){
		vector< std::pair<unsigned int, float> >::iterator plot_iter;

		for(plot_iter = genome_plot.begin();plot_iter != genome_plot.end();plot_iter++){
			plot_iter->first = num_base - 1 - plot_iter->first;
		}

		// Re-sort the data
		sort( genome_plot.begin(), genome_plot.end() );
	}

	// Are there any selected sequence fragments?
	if(hit_list.empty()){
		return;
	}

	// Invert hit list
	list<SeqHit>::iterator iter;

	for(iter = hit_list.begin();iter != hit_list.end();iter++){
		// From the NCBI sdk.doc file:
		// "Locations of regions on Bioseqs are always given as 
		//  integer offsets, also described in the Biological 
		//  Sequences chapter. So the first residue is always 0 and 
		//  the last residue is always (length - 1)." 
		
		//     0  .........................  (N-1)
		//     5' -----> ---> -->   ------>  3'
		//
		//     3' -----> ---> -->   ------>  5'
		// (N-1)  .........................  0
		
		// start' = N - 1 - stop
		// stop' = N - 1 - start
		unsigned int tmp_start = iter->start();
		
		iter->start(num_base - 1 - iter->stop());
		iter->stop(num_base - 1 - tmp_start);

		// Switch the strand
		iter->complement( !iter->complement() );
	}

	hit_list.sort();

	// OSX apears to invalidate the iterator after sorting (bug).
	// So set this iterator after sorting the hit list.
	cur_search_hit_iter = hit_list.end();

}

bool Genome::color(const unsigned char &m_r, const unsigned char &m_g, const unsigned char &m_b,
		const unsigned int &m_feature)
{
	if(m_feature == CUSTOM_COLOR){
		list<GeneAnnotation>::iterator iter;
		bool status = false;

		for(iter = mol.begin();iter != mol.end();iter++){
			if(iter->select()){
				
				iter->color((GLfloat)(m_r)/255.0f, 
						    (GLfloat)(m_g)/255.0f, 
							(GLfloat)(m_b)/255.0f);
				status = true;
			}
		}

		return status;
	}

	switch(m_feature){
		case SELECT_COLOR:
			select_color[0] = (GLfloat)(m_r)/255.0f;
			select_color[1] = (GLfloat)(m_g)/255.0f;
			select_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case BACKGROUND_COLOR:
			background_color[0] = (GLfloat)(m_r)/255.0f;
			background_color[1] = (GLfloat)(m_g)/255.0f;
			background_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case TEXT_COLOR:
			text_color[0] = (GLfloat)(m_r)/255.0f;
			text_color[1] = (GLfloat)(m_g)/255.0f;
			text_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case GENE_COLOR:
			gene_color[0] = (GLfloat)(m_r)/255.0f;
			gene_color[1] = (GLfloat)(m_g)/255.0f;
			gene_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case PSEUDO_GENE_COLOR:
			pseudo_gene_color[0] = (GLfloat)(m_r)/255.0f;
			pseudo_gene_color[1] = (GLfloat)(m_g)/255.0f;
			pseudo_gene_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case CDS_COLOR:
			cds_color[0] = (GLfloat)(m_r)/255.0f;
			cds_color[1] = (GLfloat)(m_g)/255.0f;
			cds_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case ANNOT_COLOR:
			annot_color[0] = (GLfloat)(m_r)/255.0f;
			annot_color[1] = (GLfloat)(m_g)/255.0f;
			annot_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case PRIMER_COLOR:
			primer_color[0] = (GLfloat)(m_r)/255.0f;
			primer_color[1] = (GLfloat)(m_g)/255.0f;
			primer_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case USER_COLOR:
			user_color[0] = (GLfloat)(m_r)/255.0f;
			user_color[1] = (GLfloat)(m_g)/255.0f;
			user_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case RNA_COLOR:
			rna_color[0] = (GLfloat)(m_r)/255.0f;
			rna_color[1] = (GLfloat)(m_g)/255.0f;
			rna_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case INTERGENIC_COLOR:
			intergenic_color[0] = (GLfloat)(m_r)/255.0f;
			intergenic_color[1] = (GLfloat)(m_g)/255.0f;
			intergenic_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case SELECT_NEG_COLOR:
			select_neg_color[0] = (GLfloat)(m_r)/255.0f;
			select_neg_color[1] = (GLfloat)(m_g)/255.0f;
			select_neg_color[2] = (GLfloat)(m_b)/255.0f;
			break;
		case SELECT_PLUS_COLOR:
			select_plus_color[0] = (GLfloat)(m_r)/255.0f;
			select_plus_color[1] = (GLfloat)(m_g)/255.0f;
			select_plus_color[2] = (GLfloat)(m_b)/255.0f;
			break;
	};

	return true;
}

bool Genome::color(const GLfloat m_color[], const unsigned int &m_feature)
{
	if(m_feature == CUSTOM_COLOR){
		list<GeneAnnotation>::iterator iter;
		bool status = false;

		for(iter = mol.begin();iter != mol.end();iter++){
			if(iter->select()){
				
				iter->color(m_color);
				status = true;
			}
		}

		return status;
	}

	switch(m_feature){
		case SELECT_COLOR:
			memcpy(select_color, m_color, 3*sizeof(GLfloat));
			break;
		case BACKGROUND_COLOR:
			memcpy(background_color, m_color, 3*sizeof(GLfloat));
			break;
		case TEXT_COLOR:
			memcpy(text_color, m_color, 3*sizeof(GLfloat));
			break;
		case GENE_COLOR:
			memcpy(gene_color, m_color, 3*sizeof(GLfloat));
			break;
		case PSEUDO_GENE_COLOR:
			memcpy(pseudo_gene_color, m_color, 3*sizeof(GLfloat));
			break;
		case CDS_COLOR:
			memcpy(cds_color, m_color, 3*sizeof(GLfloat));
			break;
		case ANNOT_COLOR:
			memcpy(annot_color, m_color, 3*sizeof(GLfloat));
			break;
		case PRIMER_COLOR:
			memcpy(primer_color, m_color, 3*sizeof(GLfloat));
			break;
		case USER_COLOR:
			memcpy(user_color, m_color, 3*sizeof(GLfloat));
			break;
		case RNA_COLOR:
			memcpy(rna_color, m_color, 3*sizeof(GLfloat));
			break;
		case INTERGENIC_COLOR:
			memcpy(intergenic_color, m_color, 3*sizeof(GLfloat));
			break;
		case SELECT_NEG_COLOR:
			memcpy(select_neg_color, m_color, 3*sizeof(GLfloat));
			break;
		case SELECT_PLUS_COLOR:
			memcpy(select_plus_color, m_color, 3*sizeof(GLfloat));
			break;
	};

	return true;
}

string Genome::selected_seq_as_protein()
{
	bool status = false;
	stringstream seq;

	list<GeneAnnotation>::iterator iter;

	for(iter = mol.begin();iter != mol.end();iter++){
		if(iter->select()){
			status = true;

			// Output the fast header
			// seq << "> " << iter->seq_id_str() << " " << iter->info(GeneAnnotation::PRODUCT) << ENDL;

			// Output the sequence
			seq << translate_sequence(iter->start(), iter->stop()) << ENDL;

		}
	}
	
	if(status == false){
		seq << "Please select one or more genes" << ENDL;
	}

	return seq.str();
}

string Genome::selected_seq_as_dna()
{
	bool status = false;
	stringstream seq;

	list<GeneAnnotation>::iterator iter;

	for(iter = mol.begin();iter != mol.end();iter++){
		if(iter->select()){
			status = true;

			DNA3Seq local_seq;

			if(iter->is_complement()){
				local_seq = mol.sequence().subseq(iter->start(), iter->stop() - iter->start() + 1).complement();
			}
			else{
				local_seq = mol.sequence().subseq(iter->start(), iter->stop() - iter->start() + 1);
			}
			
			float gc_content = 100.0f*local_seq.gc_content();

			// Output the fasta header
			seq << "> ";

			switch(iter->type()){
				case GeneAnnotation::CDS:
				case GeneAnnotation::GENE:
				case GeneAnnotation::PSEUDO_GENE:
				case GeneAnnotation::RNA:
				case GeneAnnotation::tRNA:
					seq << iter->seq_id_str() << " " 
						<< iter->name() << " " 
						<< iter->info(GeneAnnotation::PRODUCT) << " "
						<< "%G+C = " << gc_content << ENDL;

					break;
				case GeneAnnotation::USER:
				case GeneAnnotation::IMP:
					seq << iter->name() << " " 
						<< iter->info(GeneAnnotation::PRODUCT) << " "
						<< "%G+C = " << gc_content << ENDL;

					break;
				case GeneAnnotation::PRIMER:
					seq << "Primer "
						<< iter->info(GeneAnnotation::LOCUS)
						<< ", "
						<< iter->info(GeneAnnotation::PRIMER_TM)
						<< ", " << iter->info(GeneAnnotation::PRIMER_GC) << ENDL;

					break;
				case GeneAnnotation::NONE:
					seq << "intergenic space"  << " "
						<< "%G+C = " << gc_content << ENDL;
					break;
			};

			// Output the sequence
			seq << local_seq << ENDL;
		}
	}

	if(status == false){
		seq << "Please select one or more annotations" << ENDL;
	}

	return seq.str();
}

string Genome::sequence(const long &m_start, const long &m_stop)
{
	stringstream output;
	unsigned int start, stop;


	if( wrap(m_stop - m_start, num_base) > wrap(m_start - m_stop, num_base)){
		stop = (unsigned int)m_start;
		start = (unsigned int)m_stop;
	}
	else{
		start = (unsigned int)m_start;
		stop = (unsigned int)m_stop;
	}

	unsigned int seq_len = 1;
	
	// Compute the subsequence length (i.e. the distance between start and stop
	// in periodic coordinates).
	for(long i = (long)start;i != (long)stop;i++){
		seq_len ++;

		// Wrap around
		if(i == (long)num_base - 1){
			i = -1;
		}
	}

	// Get the subseq (force the DNA molecule to be circular even if this
	// is not actually the case).
	DNA3Seq local_seq = mol.sequence().subseq( (unsigned int)start, seq_len, true );

	float gc_content = 100.0f*local_seq.gc_content();

	output << "> " << seq_len << " bases (+)[" << start << ".." << stop << "] " 
		   << "%G+C = " << gc_content << ENDL
		   << local_seq << ENDL
		   << "> " << seq_len << " bases (-)[" << start << ".." << stop << "] " 
		   << "%G+C = " << gc_content << ENDL
		   << local_seq.complement() << ENDL;

	return output.str();
}

float Genome::cur_tm(const GLfloat &m_x)
{
	unsigned int cur_base = cur_base_loc(m_x);
	const unsigned int total_window = 2*tm_window + 1;

	DNA3Seq::iterator iter;
	NucCruc melt;

	melt.Salt(salt);
	melt.Strand(strand);

	if(cur_base >= tm_window){
		iter = mol.sequence().begin() + (cur_base - tm_window);
	}
	else{
		iter = mol.sequence().begin() + ( seq_size() - (tm_window - cur_base) );
	}

	for(cur_base = 0;cur_base < total_window;cur_base++){
		
		switch(iter.base()){
			case DNA3Seq::A:
				melt.push_back_native(NucCruc::A);
				break;
			case DNA3Seq::T:
				melt.push_back_native(NucCruc::T);
				break;
			case DNA3Seq::G:
				melt.push_back_native(NucCruc::G);
				break;
			case DNA3Seq::C:
				melt.push_back_native(NucCruc::C);
				break;
			default:
				melt.push_back_native(NucCruc::UNKNOWN);
				break;
		};

		iter++;

		if(iter == mol.sequence().end()){
			iter = mol.sequence().begin();
		}
	}

	return melt.Tm();
}

float Genome::cur_gc(const GLfloat &m_x)
{
	float gc = 0.0f;
	unsigned int cur_base = cur_base_loc(m_x);
	const unsigned int total_window = 2*content_window + 1;

	DNA3Seq::iterator iter;
	
	if(cur_base >= content_window){
		iter = mol.sequence().begin() + (cur_base - content_window);
	}
	else{
		iter = mol.sequence().begin() + ( seq_size() - (content_window - cur_base) );
	}
		
	for(cur_base = 0;cur_base < total_window;cur_base++){
		unsigned char c = iter.base();
		
		gc += ( (c == DNA3Seq::G) || (c == DNA3Seq::C));

		iter++;

		if(iter == mol.sequence().end()){
			iter = mol.sequence().begin();
		}
	}

	return (100.0f*gc/total_window);
}

float Genome::cur_at(const GLfloat &m_x)
{
	float at = 0.0f;
	unsigned int cur_base = cur_base_loc(m_x);
	const unsigned int total_window = 2*content_window + 1;

	DNA3Seq::iterator iter;
	
	if(cur_base >= content_window){
		iter = mol.sequence().begin() + (cur_base - content_window);
	}
	else{
		iter = mol.sequence().begin() + ( seq_size() - (content_window - cur_base) );
	}
		
	for(cur_base = 0;cur_base < total_window;cur_base++){
		unsigned char c = iter.base();
		
		at += ( (c == DNA3Seq::A) || (c == DNA3Seq::T) );

		iter++;

		if(iter == mol.sequence().end()){
			iter = mol.sequence().begin();
		}
	}

	return (100.0f*at/total_window);
}

float Genome::cur_gc_skew(const GLfloat &m_x)
{
	int gPc = 0;
	int gMc = 0;
	unsigned int cur_base = cur_base_loc(m_x);
	const unsigned int total_window = 2*content_window + 1;

	DNA3Seq::iterator iter;
	
	if(cur_base >= content_window){
		iter = mol.sequence().begin() + (cur_base - content_window);
	}
	else{
		iter = mol.sequence().begin() + ( seq_size() - (content_window - cur_base) );
	}
		
	for(cur_base = 0;cur_base < total_window;cur_base++){
		unsigned char c = iter.base();
		
		switch(c){
			case DNA3Seq::G:
				gPc ++;
				gMc ++;
				break;
			case DNA3Seq::C:
				gPc ++;
				gMc --;
				break;
		};

		iter++;

		if(iter == mol.sequence().end()){
			iter = mol.sequence().begin();
		}
	}
	
	return (gPc == 0) ? 0.0f : ((float)gMc)/gPc;
}

float Genome::cur_at_skew(const GLfloat &m_x)
{
	int aPt = 0;
	int aMt = 0;
	unsigned int cur_base = cur_base_loc(m_x);
	const unsigned int total_window = 2*content_window + 1;

	DNA3Seq::iterator iter;
	
	if(cur_base >= content_window){
		iter = mol.sequence().begin() + (cur_base - content_window);
	}
	else{
		iter = mol.sequence().begin() + ( seq_size() - (content_window - cur_base) );
	}
		
	for(cur_base = 0;cur_base < total_window;cur_base++){
		unsigned char c = iter.base();
		
		switch(c){
			case DNA3Seq::A:
				aPt ++;
				aMt ++;
				break;
			case DNA3Seq::T:
				aPt ++;
				aMt --;
				break;
		};

		iter++;

		if(iter == mol.sequence().end()){
			iter = mol.sequence().begin();
		}
	}
	
	return (aPt == 0) ? 0.0f : ((float)aMt)/aPt;
}

float Genome::cur_keto_skew(const GLfloat &m_x)
{
	int gPt = 0;
	int aPc = 0;
	unsigned int cur_base = cur_base_loc(m_x);
	const unsigned int total_window = 2*content_window + 1;

	DNA3Seq::iterator iter;
	
	if(cur_base >= content_window){
		iter = mol.sequence().begin() + (cur_base - content_window);
	}
	else{
		iter = mol.sequence().begin() + ( seq_size() - (content_window - cur_base) );
	}
		
	for(cur_base = 0;cur_base < total_window;cur_base++){
		unsigned char c = iter.base();
		
		switch(c){
			case DNA3Seq::A:
				aPc++;
				break;
			case DNA3Seq::T:
				gPt++;
				break;
			case DNA3Seq::G:
				gPt++;
				break;
			case DNA3Seq::C:
				aPc++;
				break;
		};

		iter++;

		if(iter == mol.sequence().end()){
			iter = mol.sequence().begin();
		}
	}
	
	return ((float)(gPt - aPc))/total_window;
}

float Genome::cur_purine_skew(const GLfloat &m_x)
{
	int gPa = 0;
	int tPc = 0;
	unsigned int cur_base = cur_base_loc(m_x);
	const unsigned int total_window = 2*content_window + 1;

	DNA3Seq::iterator iter;
	
	if(cur_base >= content_window){
		iter = mol.sequence().begin() + (cur_base - content_window);
	}
	else{
		iter = mol.sequence().begin() + ( seq_size() - (content_window - cur_base) );
	}
		
	for(cur_base = 0;cur_base < total_window;cur_base++){
		unsigned char c = iter.base();
		
		switch(c){
			case DNA3Seq::A:
				gPa++;
				break;
			case DNA3Seq::T:
				tPc++;
				break;
			case DNA3Seq::G:
				gPa++;
				break;
			case DNA3Seq::C:
				tPc++;
				break;
		};

		iter++;

		if(iter == mol.sequence().end()){
			iter = mol.sequence().begin();
		}
	}
	
	return ((float)(gPa - tPc))/total_window;
}

float Genome::cur_genome_plot(const GLfloat &m_x)
{
	unsigned int cur_base = cur_base_loc(m_x);
	const unsigned int plot_len = genome_plot.size();
	
	for(unsigned int i = 0;i < plot_len;i++){

		if(genome_plot[i].first >= cur_base){
			return genome_plot[i].second;
		}
	}

	return 0.0f;
}

void Genome::set_genome_plot(const list< pair<unsigned int, float> > &m_plot)
{
	list< pair<unsigned int, float> >::const_iterator iter = m_plot.begin();

	// Remove any existing data
	genome_plot.clear();

	if( m_plot.empty() ){
		genome_plot_bounds_y = pair<float, float>(0.0f, 0.0f);
		display_genome_plot = false;
		return;
	}

	// Turn off the genome plot display until we have copied all of the data
	display_genome_plot = false;

	if(seq_size() < iter->first){
		throw "Genome::set_genome_plot: base overflow";
	}

	// Make insertion efficient
	genome_plot.reserve( m_plot.size() );

	genome_plot_bounds_y.first = iter->second;
	genome_plot_bounds_y.second = iter->second;

	genome_plot.push_back(*iter);

	iter++;

	while( iter != m_plot.end() ){
		if(seq_size() < iter->first){
			// throw "Genome::set_genome_plot: base overflow";
			break; // Truncate the plot data to match the amount of sequence
		}

		genome_plot_bounds_y.first = min(genome_plot_bounds_y.first, iter->second);
		genome_plot_bounds_y.second = max(genome_plot_bounds_y.second, iter->second);
		
		genome_plot.push_back(*iter);
		iter++;
	}
	
	// Turn on the genome plot display
	display_genome_plot = true;	
}

void Genome::compute_primers(const unsigned int &m_options, const unsigned int &m_len,
							 const unsigned int &m_run,
							 const float &m_salt, const float &m_strand,
							 const float &m_min_gc, const float &m_max_gc, 
							 const float &m_tm, const float &m_mis_prime_tm)
{
	uout << "Computing primers for " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl;

	// Clear any previously computed primers
	mol.clear_primers();

	mol.compute_primers(m_options, m_len, m_run, m_salt, m_strand,
		m_min_gc, m_max_gc, m_tm, m_mis_prime_tm);

	// Reprocess the annotation list
	mol.processGeneList();

	// Compute the maximum depths
	list<GeneAnnotation>::iterator iter;
	
	max_gene_depth = max_annot_depth = 0;

	// Recompute the maximum depth
	for(iter = mol.begin();iter != mol.end();iter++){
		if(max_gene_depth < iter->depth()){
			max_gene_depth = iter->depth();
		}

		if(max_annot_depth < iter->annot_depth()){
			max_annot_depth = iter->annot_depth();
		}
	}
	
	clear_auto_pick();
}

void Genome::clear_primers()
{

	// Keep the user informed
	uout << "Clearing primers for " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl;

	mol.clear_primers();

	// Reprocess the annotation list
	mol.processGeneList();

	// Compute the maximum depths
	list<GeneAnnotation>::iterator iter;
	
	max_gene_depth = max_annot_depth = 0;

	// Recompute the maximum depth
	for(iter = mol.begin();iter != mol.end();iter++){
		if(max_gene_depth < iter->depth()){
			max_gene_depth = iter->depth();
		}

		if(max_annot_depth < iter->annot_depth()){
			max_annot_depth = iter->annot_depth();
		}
	}
	
	clear_auto_pick();
}

void Genome::mode(const unsigned int &m_mode)
{
	mol.mode(m_mode);

	// Compute the maximum depths
	list<GeneAnnotation>::iterator iter;
	
	max_gene_depth = max_annot_depth = 0;

	for(iter = mol.begin();iter != mol.end();iter++){
		if(max_gene_depth < iter->depth()){
			max_gene_depth = iter->depth();
		}

		if(max_annot_depth < iter->annot_depth()){
			max_annot_depth = iter->annot_depth();
		}
	}
}

bool Genome::annotate_selection()
{	
	if( hit_list.empty() ){
		return false;
	}

	mol.annotate_selection(hit_list);

	// Reprocess the annotation list
	mol.processGeneList();

	// Compute the maximum depths
	list<GeneAnnotation>::iterator iter;
	
	max_gene_depth = max_annot_depth = 0;

	for(iter = mol.begin();iter != mol.end();iter++){
		if(max_gene_depth < iter->depth()){
			max_gene_depth = iter->depth();
		}

		if(max_annot_depth < iter->annot_depth()){
			max_annot_depth = iter->annot_depth();
		}
	}

	clear_auto_pick();
	
	return true;
}

bool Genome::delete_annotation()
{
	if(mol.delete_annotation() == false){
		return false;
	}

	// Clear any existing search results that may reference one of the
	// deleted annotations
	search_results.clear();
	cur_search_iter = search_results.end();

	// Reprocess the annotation list
	mol.processGeneList();

	// Compute the maximum depths
	list<GeneAnnotation>::iterator iter;
	
	max_gene_depth = max_annot_depth = 0;

	for(iter = mol.begin();iter != mol.end();iter++){
		if(max_gene_depth < iter->depth()){
			max_gene_depth = iter->depth();
		}

		if(max_annot_depth < iter->annot_depth()){
			max_annot_depth = iter->annot_depth();
		}
	}
	
	clear_auto_pick();
	
	return true;
}

///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////

GenomePage::GenomePage()
{
	// Set default behaviour here
	text_font_size = 12;
	zoom_level = ZOOM_LEVEL_TOP;
	scale_mode = DNAMol::SIZE_CENTRIC;
	mouse_status = GP_MOUSE_CLEAR;
	
	text_font_scale_factor = 1.0f;

	salt = NC_Na;	// M
	strand = NC_CT;	// M

	content_window_size = 25;
	tm_window_size = 17;

	max_genome_display = 5;
	
	// The buffer has not been initialized
	jump_to_buffer = -1;

	base_window = DEFAULT_BASE_WINDOW;
	y_bottom = 0.0f;
	y_top = 1.0f;

	mouse_loc[X] = 0;
	mouse_loc[Y] = 0;

	scroll_loc = -1;
	zoom_loc = -1;
	
	// Turn off momentum for now
	momentum = false;

	// We're not animating until the user gives us a spin
	spinning = false;
	
	// Don't enable auto-picking unless the user tells us to
	pointer_auto_pick = false;
	
	// We're not plotting multiple sequence conservation until the user
	// tells us to.
	conservation_plot = false;

	// Set an error handler for the NCBI library
	old_error_hook = Nlm_ErrSetHandler(NCBIErrHookProc);
	
	genome_export_format = EXPORT_INVALID;
}

bool GenomePage::load(const std::string &m_filename, const int &m_type /* = ASN_1*/)
{	
	bool has_sequence_data = true;
	
	if(m_type < 0){
		throw "Could not open file";
	}
	
	if(m_type == DNAMol::PTT){
		throw "PTT files have no associated sequence. Please use Genome | import annotation ...";
	}
	
	try{
		// Provide the user with some timing info
		time_t profile = time(NULL);

		// Are any genomes currently loaded? If not adjust the amount of zoom
		// as needed
		const bool zoom_to_genome = genome_list.empty();

		list<Genome> tmp_list;
		bool keep_reading = true;
		streampos file_loc = 0;
		
		unsigned int max_base_window = 0;
		
		// Since there may be multiple genomes in a file, keep reading
		// until we can't load another
		while(keep_reading){

			tmp_list.push_back( Genome(scale_mode) );

			// Load data into a temp file name to make error
			// recovery more robust (windows message can sneek between a 
			// throw and catch --> trying to render a malformed Genome).
			keep_reading = tmp_list.back().load(m_filename, m_type, file_loc);

			// If there is no sequence associated with this genome, don't save it!
			if( tmp_list.back().empty() ){
				tmp_list.pop_back();
				break;
			}
			
			// Did we read real bases?
			has_sequence_data = ( has_sequence_data && tmp_list.back().has_sequence_data() );
			
			max_base_window = max(max_base_window, tmp_list.back().seq_size()/2);
			
			// Set GenomePage level quantities:
			tmp_list.back().zoom(zoom_level, base_window);
			tmp_list.back().color(background, Genome::BACKGROUND_COLOR);

			// Set the font scaling
			tmp_list.back().font_scale_factor(text_font_scale_factor);
		}

		// To avoid pay the cost of copying a potentially large genome,
		// use splice to transfer the elements of tmp_list into genome_list.
		genome_list.splice(genome_list.end(), tmp_list);
		
		if( zoom_to_genome && (max_base_window < DEFAULT_BASE_WINDOW) ){
			zoom(max_base_window);
		}
		
		// Re-compute the view size
		view_size();
		
		profile = time(NULL) - profile;
	}
	catch(const char *error){
		throw error;
	}
	catch(...){
		throw "Unknown datafile loading error";
	}
	
	return has_sequence_data;
}

bool GenomePage::load(const unsigned int &m_gi)
{	
	bool has_sequence_data = true;
	
	try{
		// Provide the user with some timing info
		time_t profile = time(NULL);

		// Are any genomes currently loaded? If not adjust the amount of zoom
		// as needed
		const bool zoom_to_genome = genome_list.empty();
		
		list<Genome> tmp_list;

		tmp_list.push_back( Genome(scale_mode) );

		// Load data into a temp file name to make error
		// recovery more robust (windows message can sneek between a 
		// throw and catch --> trying to render a malformed Genome).
		tmp_list.back().load(m_gi);
		
		// Did we read real bases?
		has_sequence_data = ( has_sequence_data && tmp_list.back().has_sequence_data() );
			
		unsigned int max_base_window = tmp_list.back().seq_size()/2;
		
		// Set GenomePage level quantities:
		tmp_list.back().zoom(zoom_level, base_window);
		tmp_list.back().color(background, Genome::BACKGROUND_COLOR);

		// Set the font scaling
		tmp_list.back().font_scale_factor(text_font_scale_factor);

		// To avoid pay the cost of copying a potentially large genome,
		// use splice to transfer the elements of tmp_list into genome_list.
		genome_list.splice(genome_list.end(), tmp_list);
		
		if( zoom_to_genome && (max_base_window < DEFAULT_BASE_WINDOW) ){
			zoom(max_base_window);
		}
		
		// Re-compute the view size
		view_size();

		profile = time(NULL) - profile;

		uout << "Download record in " << profile << " sec" << uout_endl;
	}
	catch(const char *error){
		throw error;
	}
	catch(...){
		throw "Unknown ASN.1 loading error";
	}
	
	return has_sequence_data;
}

bool GenomePage::load_and_save(const unsigned int &m_gi, const string &m_filename)
{	
	bool has_sequence_data = true;
	
	try{
		// Are any genomes currently loaded? If not adjust the amount of zoom
		// as needed
		const bool zoom_to_genome = genome_list.empty();
		
		list<Genome> tmp_list;

		tmp_list.push_back( Genome(scale_mode) );

		// Load data into a temp file name to make error
		// recovery more robust (windows message can sneek between a 
		// throw and catch --> trying to render a malformed Genome).
		tmp_list.back().load_and_save(m_gi, m_filename);

		// Does this genome have sequence data?
		if(tmp_list.back().empty() == true){
			return has_sequence_data;
		}
		
		// Did we read real bases?
		has_sequence_data = ( has_sequence_data && tmp_list.back().has_sequence_data() );
			
		unsigned int max_base_window = tmp_list.back().seq_size()/2;
		
		// Set GenomePage level quantities:
		tmp_list.back().zoom(zoom_level, base_window);
		tmp_list.back().color(background, Genome::BACKGROUND_COLOR);

		// Set the font scaling
		tmp_list.back().font_scale_factor(text_font_scale_factor);

		// To avoid pay the cost of copying a potentially large genome,
		// use splice to transfer the elements of tmp_list into genome_list.
		genome_list.splice(genome_list.end(), tmp_list);
		
		if( zoom_to_genome && (max_base_window < DEFAULT_BASE_WINDOW) ){
			zoom(max_base_window);
		}
		
		// Re-compute the view size
		view_size();
	}
	catch(const char *error){
		throw error;
	}
	catch(...){
		throw "Unknown ASN.1 loading error";
	}
	
	return has_sequence_data;
}

bool GenomePage::import_annotation(const std::string &m_filename, const int &m_type /* = -1*/)
{	
	// The user must have at least one genome selected!
	if( (num_display() > 1) &&  (num_select() < 1) ){
		throw "Please select at least one destination genome";
	}
	
	int tmp_type = m_type;
	
	if(tmp_type < 0){
		// Autodetect the file type
		tmp_type = file_type(m_filename);
	}
	
	if(tmp_type < 0){
		throw "Could not open file";
	}

	if(tmp_type == DNAMol::FASTA){
		throw "FASTA file do not contain annotation!";
	}
	
	try{
		// Provide the user with some timing info
		time_t profile = time(NULL);
		
		list<Genome>::iterator genome_iter;
		Genome tmp_genome(scale_mode);
		streampos file_loc = 0;
		
		// Read in the genome
		tmp_genome.load(m_filename, tmp_type, file_loc);

		// Is there only a single displayed genome?
		const bool single_genome = (num_display() == 1);
		
		for(genome_iter = genome_list.begin();genome_iter != genome_list.end();
			genome_iter++){

			if( single_genome || genome_iter->select() ){
				
				// Add the annotations in tmp_genome to the
				// annotations contained in *genome_iter
				genome_iter->import(tmp_genome, tmp_type);
			}
		}
		profile = time(NULL) - profile;
		
		uout << "Loaded file in " << profile << " sec" << uout_endl;
	}
	catch(const char *error){
		throw error;
	}
	catch(...){
		throw "Unknown datafile loading error";
	}
	
	return true;
}

void GenomePage::zoom(const int &m_base_window)
{
	// Make sure there is not an over- or under-flow in the
	// base_window calculation
	if(m_base_window <= 0){
		return;
	}

	if(m_base_window > BASE_TOP_THRESHOLD){
		zoom_level = ZOOM_LEVEL_TOP;
		base_window = m_base_window;
	}
	else{
		if(m_base_window > BASE_MID_THRESHOLD){
			zoom_level = ZOOM_LEVEL_MED;
			base_window = m_base_window;
		}
		else{
			zoom_level = ZOOM_LEVEL_SEQ;
				
			// Make sure that we don't drop below the minimum
			if(base_window < MIN_BASE_WINDOW){
				base_window = MIN_BASE_WINDOW;
			}
			else{
				base_window = m_base_window;
			}
		}
	}

	list<Genome>::iterator iter;
	
	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		iter->zoom(zoom_level, base_window);
	}

	// Set up a perspective for 2D viewing. Note that the
	// X axis has units of Nucleotide base.
	glLoadIdentity();
	gluOrtho2D(0.0, 2*base_window, y_bottom, y_top);
}

void GenomePage::fit_to_window()
{
	// Restore the default values
	base_window = 0;
	zoom_level = ZOOM_LEVEL_TOP;

	list<Genome>::iterator iter;
	
	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		base_window = max(base_window, iter->seq_size()/2);
	}

	zoom(base_window);
	jump_to(0);
}

// Jump to the given base for the selected genomes
// If no base is provided (i.e. m_base_num < 0) use the internal buffer
void GenomePage::jump_to(const long &m_base_num /* = -1*/)
{
	long offset = (m_base_num >= 0) ? m_base_num : jump_to_buffer;

	if(offset < 0){
		return;
	}

	list<Genome>::iterator iter;
	
	// If no genomes are selected, move ALL genomes!
	if(num_select() == 0){
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			iter->jump_to(offset);
		}
	}
	else{
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if(iter->select()){
				iter->jump_to(offset);
			}
		}
	}
}

void GenomePage::jump_buffer_push(const unsigned int &m_i)
{
	if(jump_to_buffer < 0){
		jump_to_buffer = m_i;
	}
	else{
		jump_to_buffer = 10*jump_to_buffer + m_i;
	}
}


void GenomePage::background_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	background[0] = (GLfloat)(m_r)/255.0f;
	background[1] = (GLfloat)(m_g)/255.0f;
	background[2] = (GLfloat)(m_b)/255.0f;

	glClearColor(background[0], 
				 background[1], 
				 background[2], 
				 background[3]);

	list<Genome>::iterator iter;
	
	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		iter->color(m_r, m_g, m_b, Genome::BACKGROUND_COLOR);
	}
}

void GenomePage::gene_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::GENE_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the gene color" << uout_endl;
	}
}

void GenomePage::pseudo_gene_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::PSEUDO_GENE_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the pseudo-gene color" << uout_endl;
	}
}

void GenomePage::cds_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::CDS_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the CDS color" << uout_endl;
	}
}

void GenomePage::rna_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::RNA_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the RNA color" << uout_endl;
	}
}

void GenomePage::annot_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::ANNOT_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the annotation color" << uout_endl;
	}
}

void GenomePage::primer_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::PRIMER_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the primer color" << uout_endl;
	}
}

void GenomePage::user_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::USER_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the user color" << uout_endl;
	}
}

void GenomePage::text_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::TEXT_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the text color" << uout_endl;
	}
}

void GenomePage::intergenic_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::INTERGENIC_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the intergenic space color" << uout_endl;
	}
}

void GenomePage::seq_plus_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::SELECT_PLUS_COLOR);
			
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the plus strand selection color" << uout_endl;
	}
}

void GenomePage::seq_minus_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::SELECT_NEG_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the minus strand selection color" << uout_endl;
	}
}

void GenomePage::select_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			iter->color(m_r, m_g, m_b, Genome::SELECT_COLOR);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to set the selection color" << uout_endl;
	}
}

void GenomePage::custom_color(const unsigned char &m_r, const unsigned char &m_g,
	const unsigned char &m_b)
{
	list<Genome>::iterator iter;
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			if(iter->color(m_r, m_g, m_b, Genome::CUSTOM_COLOR) == true){
				status = true;
			}
		}
	}

	if(status == false){
		uout << "Please select at least one annotation to apply custom color" << uout_endl;
	}
}

void GenomePage::scroll(const GLfloat &m_dy)
{
	// Count the number of displayed genomes
	list<Genome>::iterator iter;
	unsigned int render_count = 0;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->show()){
			render_count++;
		}
	}

	// Is the user allowed to scroll yet?
	if(render_count <= max_genome_display){
		return;
	}

	// Make sure that we don't scoll off the top or bottom of the page
	if(m_dy > 0.0f){
		// Check the top of the page
		GLfloat delta = min(m_dy, render_count - y_top);
		
		y_top += delta;
		y_bottom += delta;
	}
	else{ // m_dy < 0.0f
		// Check the bottom of the page
		GLfloat delta = min(-m_dy, y_bottom);
		
		y_top -= delta;
		y_bottom -= delta;
	}

	glLoadIdentity();
	gluOrtho2D(0.0, 2*base_window, y_bottom, y_top);
}

bool GenomePage::initialize_OGL()
{
	// mat_specular and mat_shininess are not used!
	mat_specular[0] = 0.3f;
	mat_specular[1] = 0.3f;
	mat_specular[2] = 0.3f;
	mat_specular[3] = 0.3f;

	mat_shininess[0] = 15.0f;

	background[0] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_BACKGROUND_R);
	background[1] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_BACKGROUND_G);
	background[2] = NORMALIZE_COLOR_CHANNEL(GV_DEFAULT_BACKGROUND_B);
	background[3] = 0.0f;

	light_color[0] = 1.0f;
	light_color[1] = 1.0f;
	light_color[2] = 1.0f;
	light_color[3] = 1.0f;

	light_position[0] = 0.0f;
	light_position[1] = 0.0f;
	light_position[2] = 1.0f;

	//glEnable(GL_DEPTH_TEST);
	glDisable(GL_DEPTH_TEST);
  
  	glClearColor(background[0], 
				 background[1], 
				 background[2], 
				 background[3]);

  	//glClearDepth(1.0);
  	//glEnable(GL_NORMALIZE);

	// Anti-aliasted lines
	// Only enable anti-aliasing on machines with fast
	// graphics cards -- how can we test for this?
#ifdef WIN32
	glEnable(GL_LINE_SMOOTH);
#endif // WIN32

	// How should dotted lines look?
	glLineStipple(1, 0x00FF); // Dashed

	// Cull back faces of polygons
	//glCullFace(GL_BACK);
	//glEnable(GL_CULL_FACE);

	// Need to turn on GL_DITHER for 3D objects and turn it off for
	// lines and points.
	glDisable(GL_DITHER);
	
	// Need to set glShadeModel to GL_SMOOTH for 3D objects 
	// and set it to GL_FLAT for lines and points.
	//glShadeModel(GL_SMOOTH);
	//glShadeModel(GL_FLAT);

	// Set the current lighting parameters

	// We oly need a simple light model for now
	//glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  	//glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);

	// For now, create a single diffuse light source
	//glLightfv(GL_LIGHT0, GL_AMBIENT, light_color);

	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_color);
	//glLightfv(GL_LIGHT0, GL_SPECULAR, light_color);
	
	// Don't set the light position for 2D rendering
	//glLightfv(GL_LIGHT0, GL_POSITION, light_position);
	
	// A spot light
	glLightfv(GL_LIGHT1, GL_SPECULAR, light_color);
	glLightfv(GL_LIGHT1, GL_POSITION, light_position);
	
	glEnable(GL_LIGHT0);

  	glEnable(GL_LIGHTING);

	// DEBUG
	//glEnable(GL_POLYGON_SMOOTH);

	//GLubyte stipple_mask[] = {
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55,
	//	0xAA, 0xAA, 0xAA, 0xAA, 0x55, 0x55, 0x55, 0x55
	//};
	//glPolygonStipple(stipple_mask);
	
	// We don't currently need transparency. Actually we do,
	// but enable and disable the blending in the
	// enable_blend() and disable_blend() member functions
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// This is for smoothing polygons
	//glBlendFunc(GL_SRC_ALPHA_SATURATE, GL_ONE)

	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	return true;
}

bool GenomePage::setup_view_OGL(const unsigned int &m_width, const unsigned int &m_height)
{
	// Save the screen dimensions
	screen_size(m_width, m_height);

	glViewport(0, 0, m_width, m_height);
	
	glMatrixMode(GL_PROJECTION);
    	glLoadIdentity();

	// Set up a perspective for 2D viewing. Note that the
	// X axis has units of Nucleotide base.
	gluOrtho2D(0.0, 2*base_window, y_bottom, y_top);
			 
	return true;
}

#ifdef _DEBUG
void GenomePage::get_error_OGL(char *m_file, int m_line)
{
	char *error_header;
	static char error_msg[1024];
	GLenum error_code;

	while((error_code = glGetError()) != GL_NO_ERROR){
		switch(error_code){
			case GL_INVALID_ENUM:
				error_header = "GL INVALID_ENUM";
				break;
			case GL_INVALID_VALUE:
				error_header = "GL INVALID_VALUE";
				break;
			case GL_INVALID_OPERATION:
				error_header = "GL INVALID_OPERATION";
				break;
			case GL_STACK_OVERFLOW:
				error_header = "GL STACK_OVERFLOW";
				break;
			case GL_STACK_UNDERFLOW:
				error_header = "GL STACK_UNDERFLOW";
				break;
			case GL_OUT_OF_MEMORY:
				error_header = "GL OUT_OF_MEMORY";
				break;
		};

		sprintf(error_msg, "%s in %f @ %d", error_header, m_file, m_line);

		throw error_msg;
	}
}
#endif // _DEBUG

// Hide any selected genomes
bool GenomePage::hide_selection()
{
	list<Genome>::iterator genome_iter;
	bool status = false;
	
	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++){

		if(genome_iter->select()){
			status = true;
			
			// deselect the genome
			genome_iter->select(false);

			// Hide the genome
			genome_iter->show(false);
		}
	}

	if(status){
		// If genomes have been hidden, recompute the
		// y_top and y_bottom view boundaries
		max_mol( max_mol() );
	}
	else{
		uout << "Please select at least one genome to hide" << uout_endl;
	}

	return status;
}

// Hide any selected genome annotation
bool GenomePage::hide_selection_annotation()
{
	list<Genome>::iterator genome_iter;
	bool status = false;
	
	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++){

		if(genome_iter->select()){
			status = true;

			// Hide the genome
			genome_iter->show_annotation(false);
		}
	}

	if(!status){
		uout << "Please select at least one genome to hide annotation" << uout_endl;
	}

	return status;
}

// Show all selected genome annotation
bool GenomePage::show_annotation()
{
	list<Genome>::iterator genome_iter;
	bool status = false;
	
	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++){

		if(genome_iter->select()){
			status = true;

			// Hide the genome
			genome_iter->show_annotation(true);
		}
	}

	if(!status){
		uout << "Please select at least one genome to show annotation" << uout_endl;
	}

	return status;
}

// Show all genomes
bool GenomePage::show_all()
{
	list<Genome>::iterator genome_iter;
	bool ret_value = false;
	
	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++){

		if(genome_iter->show() == false){
			ret_value = true;

			genome_iter->show(true);
		}
	}

	if(ret_value){
		// If genomes have been hidden, recompute the
		// y_top and y_bottom view boundaries
		max_mol( max_mol() );
	}

	return ret_value;
}

bool GenomePage::increase_genome_display()
{
	if( max_mol() < num_display() ){
		max_mol(max_mol() + 1);
		return true;
	}

	return false;
}

bool GenomePage::decrease_genome_display()
{
	if(max_mol() > 1){
		if( max_mol() > num_display() ){
			max_mol(num_display() - 1);
		}
		else{
			max_mol(max_mol() - 1);
		}
		return true;
	}

	return false;
}


// Deselect any selected genomes
bool GenomePage::clear_selection(const bool m_clear_translation /* = true*/)
{
	list<Genome>::iterator genome_iter;
	
	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++){

		genome_iter->select(false);
	}
	
	if(m_clear_translation){
		// Clear the translation buffer
		translation_buffer_base.clear();
		translation_buffer_genome.clear();
	}
	
	uout << "Ready" << uout_endl;

	// Always force a redraw
	return true;
}

// Select all non-hidden genomes
bool GenomePage::select_all()
{	
	list<Genome>::iterator genome_iter;
	
	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++){

		if(genome_iter->show()){
			genome_iter->select(true);
		}
	}

	return true;
}

// Delete any selected genomes
bool GenomePage::delete_selection()
{
	// Make a list of elements to delete
	list< list<Genome>::iterator > reaper;

	list<Genome>::iterator genome_iter;
	
	bool status = false;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++){

		if(genome_iter->select()){
			// We found an element that needs to be deleted
			reaper.push_back(genome_iter);

			// Delselect this genome even though we're about to delete it.
			genome_iter->select(false);

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome for deletion" << uout_endl;
		return false;
	}

	// Delete all elements referenced in the repear list
	list< list<Genome>::iterator >::iterator reaper_iter;

	for(reaper_iter = reaper.begin();reaper_iter != reaper.end();reaper_iter++){
		genome_list.erase(*reaper_iter);
	}

	// Adjust the view as needed
	view_size();

	return true;
}

bool GenomePage::delete_annotation()
{
	list<Genome>::iterator iter;
	
	bool status = false;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){

		if(iter->select() && iter->delete_annotation()){
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one annotation for deletion" << uout_endl;
		return false;
	}

	return true;
}

// Shift selected genomes down one place in the genome list
bool GenomePage::shift_selection_down()
{
	// Set the rank order for all genomes
	list<Genome>::iterator genome_iter;
	int rank;
	bool status = false;

	for(rank = 0, genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++, rank++){
		
		if(genome_iter->select()){
			if(rank == 0){
				// Put the top element on the bottom
				genome_iter->rank(genome_list.size());
			}
			else{
				// Move this element up one
				genome_iter->rank(rank - 1.5f);
			}

			status = true;
		}
		else{
			genome_iter->rank(rank);
		}
	}
	
	if(status == false){
		uout << "Please select at least one genome to shift down" << uout_endl;
		return false;
	}

	genome_list.sort();

	return true;
}

// Shift selected genomes to the bottom of the view
// Preserve the relative order of selected and non-selected genomes .
bool GenomePage::shift_selection_to_bottom()
{
	// Set the rank order for all genomes
	list<Genome>::iterator genome_iter;
	int rank;
	bool status = false;

	int rank_offset = 2*genome_list.size();

	for(rank = 0, genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++, rank++){
		
		if(genome_iter->select()){
			// Move this element to the bottom
			genome_iter->rank(rank - rank_offset);

			status = true;
		}
		else{
			genome_iter->rank(rank);
		}
	}
	
	if(status == false){
		uout << "Please select at least one genome to shift to the bottom" << uout_endl;
		return false;
	}

	// Sort genomes by rank (from high to low)
	genome_list.sort();

	return true;
}

// Shift selected genomes up one place in the genome list
bool GenomePage::shift_selection_up()
{
	// Set the rank order for all genomes
	list<Genome>::iterator genome_iter;
	int rank;
	bool status = false;
	unsigned int num_genome = genome_list.size();

	for(rank = 0, genome_iter = genome_list.begin();genome_iter != genome_list.end();
		genome_iter++, rank++){
		
		if(genome_iter->select()){
			if(rank == (int)(num_genome - 1)){
				// Put the bottom element on top
				genome_iter->rank(-0.5f);
			}
			else{
				// Move this element up one
				genome_iter->rank(rank + 1.5f);
			}

			status = true;
		}
		else{
			genome_iter->rank(rank);
		}
	}

	if(status == false){
		uout << "Please select at least one genome to shift up" << uout_endl;
		return false;
	}

	genome_list.sort();

	return true;
}

// Take the complement of all selected genomes
bool GenomePage::complement_selection()
{
	bool status = false;
	list<Genome>::iterator genome_iter;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
		if(genome_iter->select()){
			genome_iter->complement();
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome to take the complement of" << uout_endl;
	}

	return status;
}

// Toggle the display of the %G+C graph of any selected genomes
bool GenomePage::toggle_gc_selection()
{
	bool status = false;
	list<Genome>::iterator genome_iter;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
		if(genome_iter->select()){
			genome_iter->toggle_gc_display();
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome for plotting %G+C" << uout_endl;
	}

	return status;
}

bool GenomePage::toggle_at_selection()
{
	bool status = false;
	list<Genome>::iterator genome_iter;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
		if(genome_iter->select()){
			genome_iter->toggle_at_display();
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome for plotting %A+T" << uout_endl;
	}

	return status;
}

bool GenomePage::toggle_at_skew_selection()
{
	bool status = false;
	list<Genome>::iterator genome_iter;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
		if(genome_iter->select()){
			genome_iter->toggle_at_skew_display();
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome for plotting AT Skew" << uout_endl;
	}

	return status;
}

bool GenomePage::toggle_gc_skew_selection()
{
	bool status = false;
	list<Genome>::iterator genome_iter;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
		if(genome_iter->select()){
			genome_iter->toggle_gc_skew_display();
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome for plotting GC Skew" << uout_endl;
	}

	return status;
}

bool GenomePage::toggle_keto_skew_selection()
{
	bool status = false;
	list<Genome>::iterator genome_iter;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
		if(genome_iter->select()){
			genome_iter->toggle_keto_skew_display();
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome for plotting Keto Skew" << uout_endl;
	}

	return status;
}

bool GenomePage::toggle_purine_skew_selection()
{
	bool status = false;
	list<Genome>::iterator genome_iter;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
		if(genome_iter->select()){
			genome_iter->toggle_purine_skew_display();
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome for plotting purine Skew" << uout_endl;
	}

	return status;
}

// Toggle the display of the genome plot of any selected genomes
bool GenomePage::toggle_genome_plot_selection()
{
	bool status = false;
	list<Genome>::iterator genome_iter;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
		if(genome_iter->select()){
			genome_iter->toggle_genome_plot_display();
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome for plotting user data" << uout_endl;
	}

	return status;
}

// Toggle the display of the Tm graph of any selected genomes
bool GenomePage::toggle_tm_selection()
{
	bool status = false;
	list<Genome>::iterator genome_iter;

	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
		if(genome_iter->select()){
			genome_iter->toggle_tm_display();
			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one genome for plotting Tm" << uout_endl;
	}

	return status;
}

// Note that the conservation plot is a property of the GenomePage, not individual genomes
bool GenomePage::toggle_conservation_plot()
{
	conservation_plot = !conservation_plot;

	// Always returns true since there is no way to fail!
	return true;
}

bool GenomePage::key_left_arrow()
{
	// The key board overrides any mouse-induced spinning
	spinning = false;

	// Shift by 10%
	const float dx = -0.1f;
	list<Genome>::iterator iter;

	// If no genomes are selected, move everything
	if(num_select() == 0){
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			iter->shift(dx);
		}
	}
	else{

		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if(iter->select()){
				iter->shift(dx);
			}
		}
	}

	return true;
}

bool GenomePage::key_right_arrow()
{
	// The key board overrides any mouse induced spinning
	spinning = false;

	// Shift by 10%
	const float dx = 0.1f;
	list<Genome>::iterator iter;

	// If no genomes are selected, move everything
	if(num_select() == 0){
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			iter->shift(dx);
		}
	}
	else{

		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if(iter->select()){
				iter->shift(dx);
			}
		}
	}

	return true;
}

bool GenomePage::key_up_arrow()
{
	const float zoom_fract = 0.10f;

	zoom((int)(base_window*(1.0 - zoom_fract)));

	return true;
}

bool GenomePage::key_down_arrow()
{
	const float zoom_fract = 0.10f;

	zoom((int)(base_window*(1.0 + zoom_fract)));

	return true;
}

bool GenomePage::key_page_up()
{
	const float scroll_fract = 0.20f;

	scroll(scroll_fract);

	return true;
}

bool GenomePage::key_page_down()
{
	const float scroll_fract = 0.20f;

	scroll(-scroll_fract);

	return true;
}

bool GenomePage::key_plus()
{
	return increase_genome_display();
}

bool GenomePage::key_minus()
{
	return decrease_genome_display();
}

unsigned int GenomePage::cur_base_loc()
{
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_base_loc( (2.0f*base_window*mouse_loc[X])/width );
		}

		yl = yh;
		yh += dy;
	}

	return 0;
}

bool GenomePage::base_loc_change(const int &m_x)
{
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return (iter->cur_base_loc((2.0f*base_window*mouse_loc[X])/width) !=
					iter->cur_base_loc((2.0f*base_window*m_x)/width) );
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

bool GenomePage::cur_show_tm()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_show_tm();
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

float GenomePage::cur_tm()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_tm((2.0f*base_window*mouse_loc[X])/width);
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

bool GenomePage::cur_show_gc()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_show_gc();
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

float GenomePage::cur_gc()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_gc((2.0f*base_window*mouse_loc[X])/width);
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

bool GenomePage::cur_show_at()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_show_at();
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

float GenomePage::cur_at()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_at((2.0f*base_window*mouse_loc[X])/width);
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

bool GenomePage::cur_show_gc_skew()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_show_gc_skew();
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

float GenomePage::cur_gc_skew()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_gc_skew((2.0f*base_window*mouse_loc[X])/width);
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

bool GenomePage::cur_show_at_skew()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_show_at_skew();
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

float GenomePage::cur_at_skew()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_at_skew((2.0f*base_window*mouse_loc[X])/width);
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

bool GenomePage::cur_show_keto_skew()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_show_keto_skew();
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

float GenomePage::cur_keto_skew()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_keto_skew((2.0f*base_window*mouse_loc[X])/width);
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

bool GenomePage::cur_show_purine_skew()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_show_purine_skew();
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

float GenomePage::cur_purine_skew()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_purine_skew((2.0f*base_window*mouse_loc[X])/width);
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

bool GenomePage::cur_show_genome_plot()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_show_genome_plot();
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

float GenomePage::cur_genome_plot()
{
	
	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;

	GLfloat y = mouse_loc[Y]*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!

			// Return the selected genome
			return iter->cur_genome_plot((2.0f*base_window*mouse_loc[X])/width);
		}

		yl = yh;
		yh += dy;
	}

	return false;
}

string GenomePage::selected_seq_as_protein()
{
	bool status = false;
	stringstream seq;

	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			seq <<  iter->selected_seq_as_protein();

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one annotation to display protein sequence" << uout_endl;
	}

	return seq.str();
}

string GenomePage::selected_seq_as_dna()
{
	bool status = false;
	stringstream seq;

	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			seq << iter->selected_seq_as_dna();

			status = true;
		}
	}

	if(status == false){
		uout << "Please select at least one annotation to display DNA sequence" << uout_endl;
	}

	return seq.str();
}

void GenomePage::font_scale_index(const int &m_index)
{
	// Map this index into a font scale factor
	int index = max(m_index, min_font_scale_index());

	index = min(index, max_font_scale_index());

	const float scale = (float)((index - min_font_scale_index()))/
			(max_font_scale_index() - min_font_scale_index());

	text_font_scale_factor = min_font_scale() + scale*(max_font_scale() - min_font_scale());

	// Update all genome font scales to reflect the new value
	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		iter->font_scale_factor(text_font_scale_factor);
	}
}

void GenomePage::content_window(const unsigned int &m_index)
{
	// Map this index into a %GC window size
	unsigned int index = max(m_index, min_content_window());

	index = min(index, max_content_window());

	// The GC window size must allways be odd (i.e. centered on a base)
	if(index & 1 == 1){
		// odd
		content_window_size  = index;
	}
	else{
		// even
		content_window_size  = index + 1;
	}

	// Pass the half-width of the GC window to the genomes
	// (i.e. (content_window_size - 1)/2

	index = (content_window_size - 1)/2;

	// Update all genome GC windows to reflect the new value
	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		iter->content_window_size(index);
	}
}

void GenomePage::tm_window(const unsigned int &m_index)
{
	// Map this index into a Tm window size
	unsigned int index = max(m_index, min_tm_window());

	index = min(index, max_tm_window());

	// The Tm window size must allways be odd (i.e. centered on a base)
	if(index & 1 == 1){
		// odd
		tm_window_size  = index;
	}
	else{
		// even
		tm_window_size  = index + 1;
	}

	// Pass the half-width of the GC window to the genomes
	// (i.e. (content_window_size - 1)/2

	index = (tm_window_size - 1)/2;

	// Update all genome Tm windows to reflect the new value
	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		iter->tm_window_size(index);
	}
}

void GenomePage::tm_salt_index(const int &m_index)
{
	// Map an integer index into a [Na+]
	// Map this index into a Tm window size
	int index = max(m_index, min_tm_salt_index());

	index = min(index, max_tm_salt_index());

	salt = min_tm_ln_salt() + (max_tm_ln_salt() - min_tm_ln_salt())*
			(index - min_tm_salt_index())/(max_tm_salt_index() - min_tm_salt_index());

	salt = exp(salt);

	// Update all genome [Na+] values to reflect the new value
	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		iter->tm_salt(salt);
	}
}

void GenomePage::tm_strand_index(const int &m_index)
{
	// Map an integer index into a [Strand]
	// Map this index into a Tm window size
	int index = max(m_index, min_tm_strand_index());

	index = min(index, max_tm_strand_index());

	strand = min_tm_ln_strand() + (max_tm_ln_strand() - min_tm_ln_strand())*
			(index - min_tm_strand_index())/(max_tm_strand_index() - min_tm_strand_index());

	strand = exp(strand);

	// Update all genome [Strand] values to reflect the new value
	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		iter->tm_strand(strand);
	}
}


bool GenomePage::compute_primers(const unsigned int &m_options, const unsigned int &m_len,
								 const unsigned int &m_run,
								 const float &m_salt, const float &m_strand,
								 const float &m_min_gc, const float &m_max_gc, 
								 const float &m_tm, const float &m_mis_prime_tm)
{
	bool status = false;

	// Validate the inputs
	if(m_len > PCR_PRIMER_MAX_LEN){
		throw "Primer length is too long";
	}
	
	if(m_len < PCR_PRIMER_MIN_LEN){
		throw "Primer length is too small";
	}

	if(m_salt > PCR_PRIMER_MAX_SALT){
		throw "[Na+] is too large";
	}

	if(m_salt < PCR_PRIMER_MIN_SALT){
		throw "[Na+] is too small";
	}

	if(m_strand > PCR_PRIMER_MAX_STRAND){
		throw "[Strand] is too large";
	}

	if(m_strand < PCR_PRIMER_MIN_STRAND){
		throw "[Strand] is too small";
	}

	if(m_min_gc > m_max_gc){
		throw "min %G+C is greater than max %G+C";
	}

	if(m_min_gc < 0.0f){
		throw "min %G+C is less than zero (i.e. too small)";
	}
	
	if(m_max_gc > 100.0f){
		throw "max %G+C is greater than 100% (i.e. too large)";
	}

	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			
			iter->compute_primers(m_options, m_len, m_run, 
				m_salt, m_strand,
				// Convert %G+C into fractional G+C
				m_min_gc*0.01f, m_max_gc*0.01f, 
				m_tm, m_mis_prime_tm);

			status = true;
		}
	}
	
	if(status == false){
		throw "Please select at least one genome to pick primers for";
	}

	return status;
}


bool GenomePage::clear_primers()
{
	bool status = false;

	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(iter->select()){
			
			iter->clear_primers();

			status = true;
		}
	}

	if(status == false){
			uout << "Please select at least one genome to clear primers for"
				 << uout_endl;
	}

	return status;
}

void GenomePage::mode(const unsigned int &m_mode)
{
	// Are we already in the requested mode?
	if(scale_mode == m_mode){
		return;
	}

	scale_mode = m_mode;

	if(m_mode == DNAMol::GENE_CENTRIC){
		uout << "Changing to gene centric view" << uout_endl;
	}
	else if(m_mode == DNAMol::SIZE_CENTRIC){
		uout << "Changing to annotation size centric view" << uout_endl;
	}

	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){

		iter->mode(m_mode);
	}

	if(scale_mode == DNAMol::GENE_CENTRIC){
		uout << "View is gene centric" << uout_endl;
	}
	else if(scale_mode == DNAMol::SIZE_CENTRIC){
		uout << "View is annotation size centric" << uout_endl;
	}
}


bool GenomePage::annotate_selection()
{
	bool status = false;

	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
			
		if( iter->annotate_selection() ){
			status = true;
		}
	}
	
	return status;
}

void GenomePage::export_genomes(const string &m_filename)
{
	if(num_annotations_to_export() == 0){
		throw "No annotations selected for export";
	}

	uout << "Exporting genome(s) to " << m_filename << uout_endl;
	
	switch(genome_export_format){
		case EXPORT_FASTA_GENOME:
			export_fasta_genome(m_filename);
			break;
		case EXPORT_FASTA_DNA:
			export_fasta_dna(m_filename);
			break;
		case EXPORT_FASTA_PROTEIN:
			export_fasta_protein(m_filename);
			break;
		case EXPORT_GBK:
			export_gbk(m_filename);
			break;
		case EXPORT_PTT:
			export_ptt(m_filename);
			break;
		default:
			throw "Invalid export format";
			break;
	};
	
	uout << "Finished exporting genome(s)" << uout_endl;
}

selected_annotation_info GenomePage::selected_annotation()
{
	unsigned int num_select_genome = 0;
	
	list<Genome>::const_iterator genome_iter;
	list<Genome>::const_iterator selected_genome_iter;
	
	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){
			
		if( genome_iter->select() ){
		
			selected_genome_iter = genome_iter;
			num_select_genome ++;
		}
	}
	
	if(num_select_genome == 0){
	
		return selected_annotation_info();
	}
	
	if( (num_select_genome != 1) || (selected_genome_iter->num_select() != 1) ){
		return selected_annotation_info(selected_genome_iter);
	}
	
	list<GeneAnnotation>::const_iterator annot_iter = 
		selected_genome_iter->selected_annotation();
		
	selected_annotation_info ret(selected_genome_iter, annot_iter);
	
	// Save the iterator *and* the contents of the iterator as well!
	// This will allow us to handle cases where the genome or annotation
	// have been deleted *before* the user tries to edit the annotation
	ret = *annot_iter;
	
	return ret;
}

void GenomePage::update_annotation(const selected_annotation_info &m_annot)
{
	if(m_annot.is_genome_valid() == false){
		throw "Invalid genome for annotation";
	}

	// Look up of the genome annotation
	list<Genome>::iterator genome_iter;
	
	for(genome_iter = genome_list.begin();genome_iter != genome_list.end();genome_iter++){

		if( (list<Genome>::const_iterator)(genome_iter) == m_annot.genome() ){
			break;
		}
	}
	
	if( genome_iter == genome_list.end() ){
		throw "Could not find destination genome for annotation";
	}
	
	if(m_annot.is_genome_valid() == false){
		genome_iter->update_annotation(m_annot);
	}
	else{
		genome_iter->update_annotation( m_annot, m_annot.annotation() );
	}
}

