// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "genome.h"
#include <sstream>
#include <algorithm>
#include "GLtoPS.h"
#include "nuc_cruc.h"

// Include the gif generation header from NCBI
#include "gifgen.h"

#ifdef WIN32

// for MM timers (you'll need WINMM.LIB)
#include <mmsystem.h>

# else

// Works for linux and OSX
#include <sys/time.h>

#endif // WIN32

using namespace std;

#ifdef USE_BITMAP_FONT
#define	TranslateFontBegin(X, Y, S)		glRasterPos2f(X, Y)
#define	TranslateFontEnd()				
#define	
#else
#define	TranslateFontBegin(X, Y, S)		glPushMatrix(); \
										glTranslatef(X, Y, 0.0f); \
										glScalef(font_dx*S*GP_POLYGON_FONT_SCALE, \
											S*GP_POLYGON_FONT_SCALE*font_dy, 1.0f);
										
#define	TranslateFontEnd()				glPopMatrix();
#endif // USE_BITMAP_FONT


// Fractional size (compared to window width) to display annotation labels
#define		SHOW_LABEL_THRESHOLD	(0.0175f)
#define		SHOW_LABEL_SCALE		(25.0f)

// The smallest allowed alpha value for text
#define		MIN_TEXT_ALPHA			(0.1f)

// Fractional sizes for annotation arrow components
#define		SEGMENT_DY				(0.15f)
#define		SHAFT_DX				(0.8f)

/////////////////////////// File scope global variables ///////////////////////////
// These variables are a necessary evil for transmitting global size information
// to local rendering routines. This is required because font rendering does not
// (in general) transform like a geometric feature.
float font_dy;
float font_dx;

// Note that since PostScript can not handle alpha blending, we use
// the following kludge:
bool postscript_output = false;

void Genome::render(const unsigned int &m_width, 
					const GLfloat &m_lower_y, const GLfloat &m_upper_y,
					const unsigned int &m_clip_edge, 
					const FontMetricGL &m_font_metric,
					const bool &m_is_printing)
{
	
	const int unsigned window_edge = 2*base_window;

	// Do we need to "clip" ? Note that we don't use a clipping plane
	// because of the transformations used to render text (which 
	// throws off the clipping).
	const bool clip_genome = (window_edge <= seq_size() );

	// Only show the selection effect if we're not printing
	// or copying the view.
	if( (select() == true) && !m_is_printing ){
		render_select(m_lower_y, m_upper_y);
	}
	
	switch(zoom_level){
		case ZOOM_LEVEL_TOP:
		
			render_top(m_width, m_lower_y, m_upper_y, m_font_metric, false, 
					clip_genome, m_clip_edge);
			
			break;
		case ZOOM_LEVEL_MED:
		
			render_mid(m_width, m_lower_y, m_upper_y, m_font_metric, false, 
				clip_genome, m_clip_edge);
			break;
		case ZOOM_LEVEL_SEQ:
		
			// Blend based on the zoom level
			const GLfloat blend_alpha = (GLfloat)base_window/BASE_MID_THRESHOLD;

			enable_blend(blend_alpha);

			render_mid(m_width, m_lower_y, m_upper_y, m_font_metric, false, 
				clip_genome, m_clip_edge);

			disable_blend();

			render_seq(m_width, m_lower_y, m_upper_y, false);

			break;
	}

	if(display_gc_graph){
		render_GC_graph(m_width, m_lower_y, m_upper_y);
	}

	if(display_at_graph){
		render_AT_graph(m_width, m_lower_y, m_upper_y);
	}

	if(display_gc_skew_graph){
		render_GC_skew_graph(m_width, m_lower_y, m_upper_y);
	}

	if(display_at_skew_graph){
		render_AT_skew_graph(m_width, m_lower_y, m_upper_y);
	}

	if(display_keto_skew_graph){
		render_keto_skew_graph(m_width, m_lower_y, m_upper_y);
	}

	if(display_purine_skew_graph){
		render_purine_skew_graph(m_width, m_lower_y, m_upper_y);
	}

	if(display_tm_graph){
		render_Tm_graph(m_width, m_lower_y, m_upper_y);
	}

	if(display_genome_plot){
		render_genome_plot(m_width, m_lower_y, m_upper_y);
	}

	if( (zoom_level != ZOOM_LEVEL_SEQ) && ( auto_pick_iter != mol.end() ) ){
		
		render_auto_pick(auto_pick_iter, m_lower_y, m_upper_y, m_font_metric);
	}
	
	// Render genome info
	render_info(m_lower_y, m_upper_y);
}

void Genome::render_top(const unsigned int &m_width, 
						 const GLfloat &m_lower_y, const GLfloat &m_upper_y,
						 const FontMetricGL &m_font_metric,
						 const bool &m_select, const bool &m_clip,
						 const unsigned int &m_clip_edge)
{
	if(!is_annotation_shown){
		return;
	}

	GLfloat dy = m_upper_y - m_lower_y;

	GLfloat upper_y = 0.9f*dy + m_lower_y;
	GLfloat lower_y = 0.1f*dy + m_lower_y;

	GLfloat delta = (upper_y - lower_y)/(max_gene_depth + 1.0f);

	// Render the genome as a set of non-overlapping arrows
	// (pointing in 5'->3' direction of the translated strand).
	// 
	//                 -------
	//  --           ------
	// ---- ----- -------       -----
	unsigned int base_index; // The location of this gene
	unsigned int base_start;
	unsigned int base_stop ;

	base_range(base_start, base_stop);

	// The size of the window in bases
	const int window_size_x = 2*(int)(base_window);

	const GLfloat label_alpha_offset = SHOW_LABEL_THRESHOLD*window_size_x;
	const GLfloat label_alpha_norm =(window_size_x > 0) ? SHOW_LABEL_SCALE/window_size_x : 1.0f;
	
	long x1, x2, x1p, x2p;
	GLfloat y1, y2;
	list<GeneAnnotation>::iterator iter;

	// Do we need to render any selection info (this rendering should
	// be done at the end of the for loop to insure that all text appears
	// on top of any graphics and is easy to read).
	list<GeneAnnotation>::iterator select_iter = mol.end();

	for(base_index = 0, iter = mol.begin();iter != mol.end();iter++, base_index++){
		if(iter->annot_rank() != GeneAnnotation::TOP_LEVEL_RANK){
			continue;
		}

		if(iter->select()){
			select_iter = iter;
		}

			// Pre-clip the view
		if(overlap(base_start, base_stop, iter->start(), iter->stop(), seq_size() ) == false){
			continue;
		}

		if(m_select){ // selection
			glLoadName(base_index);
		}
		else{	// screen rendering
			set_color(iter);
		}
		
		y1 = lower_y + delta*iter->depth();
		y2 = y1 + delta;

		// Map the annotation element onto the display
		// while being respectful of the periodic boundary conditions.
		// All features are rendered relative to base_start,
		// which is positioned at the left most edge of the screen
		x1 = (long)(iter->start()) - base_start;
		x2 = x1 + iter->stop() - iter->start() + 1;

		if(base_start < base_stop){
			if(iter->start() > iter->stop()){
				x2 += seq_size();
			}
		}
		else{ // base_start > base_stop
			// Use <= to prevent 1 base annotations from falling
			// throught the cracks (so to speak).
			if( iter->start() <= iter->stop() ){
			
				if(iter->stop() <= base_start){
				
					x1 += seq_size();
					x2 += seq_size();
				}
			}
			else{ // iter->start() > iter->stop()
				x2 += seq_size();
			}
		}

		const GLfloat label_alpha = (m_select == true) ? -1.0f : 
			min(1.0f, (iter->length( seq_size() ) - label_alpha_offset)*label_alpha_norm);

		for( long i = -(long)seq_size();i <= (long)seq_size();i += (long)seq_size() ){
			x1p = x1 + i;
			x2p = x2 + i;

			if((x1p > window_size_x) || (x2p < 0)){
				continue;
			}

			if(m_select){ // selection
				glLoadName(base_index);
			}
			else{	// screen rendering
				set_color(iter);
			}

			if(iter->type() == GeneAnnotation::NONE){
				// intergenic space gets a box, not an arrow
				render_box(x1p, y1, x2p, y2);
			}
			else{
				if( iter->is_complement() ){
				
					if( iter->is_segmented() ){
						render_left_segmented_arrow(x1p, y1, x2p, y2, iter, m_font_metric, label_alpha);
					}
					else{
						render_left_arrow(x1p, y1, x2p, y2, iter, m_font_metric, label_alpha);
					}
				}
				else{
				
					if( iter->is_segmented() ){
						render_right_segmented_arrow(x1p, y1, x2p, y2, iter, m_font_metric, label_alpha);
					}
					else{
						render_right_arrow(x1p, y1, x2p, y2, iter, m_font_metric, label_alpha);
					}
				}
			}
		}
	}

	// Apply the clipping rectangle *before* writing any annotation text to the screen
	if(m_clip == true){
		
		glColor4fv(background_color);

		// In order to match the glClearColor when drawing a polygon, we must
		// disable GL_LIGHTING
		glDisable(GL_LIGHTING);

		glBegin(GL_POLYGON);
	
			// Render a rectange to cover up the periodic wrapping of annotations
			glVertex2f(window_size_x, lower_y);
			glVertex2f(window_size_x, upper_y);
			glVertex2f(m_clip_edge, upper_y);
			glVertex2f(m_clip_edge, lower_y);
			
		glEnd();
		
		glEnable(GL_LIGHTING);
	}

	if(!m_select && ( select_iter != mol.end() ) ){
		render_annot_info(m_lower_y, m_upper_y, *select_iter);
	}
}

void Genome::render_mid(const unsigned int &m_width, 
						  const GLfloat &m_lower_y, const GLfloat &m_upper_y,
						  const FontMetricGL &m_font_metric,
						  const bool &m_select, const bool &m_clip,
						  const unsigned int &m_clip_edge)
{
	if(!is_annotation_shown){
		return;
	}

	GLfloat dy = m_upper_y - m_lower_y;

	GLfloat upper_y = 0.9f*dy + m_lower_y;
	GLfloat lower_y = 0.1f*dy + m_lower_y;

	GLfloat delta = (upper_y - lower_y)/(max_annot_depth + 1.0f);

	// Render the genome as a set of non-overlapping arrows
	// (pointing in 5'->3' direction of the translated strand).
	// 
	//                 -------
	//  --           ------
	// ---- ----- -------       -----
	unsigned int base_index; // The location of this gene
	unsigned int base_start;
	unsigned int base_stop ;

	base_range(base_start, base_stop);

	// The size of the window in bases
	const int window_size_x = 2*(int)(base_window);

	const GLfloat label_alpha_offset = SHOW_LABEL_THRESHOLD*window_size_x;
	const GLfloat label_alpha_norm =(window_size_x > 0) ? SHOW_LABEL_SCALE/window_size_x : 1.0f;

	long x1, x2, x1p, x2p;
	GLfloat y1, y2;

	list<GeneAnnotation>::iterator iter;
	
	// Do we need to render any selection info (this rendering should
	// be done at the end of the for loop to insure that all text appears
	// on top of any graphics and is easy to read).
	list<GeneAnnotation>::iterator select_iter = mol.end();

	for(base_index = 0, iter = mol.begin();iter != mol.end();iter++, base_index++){

		if(iter->select()){
			select_iter = iter;
		}

		// Pre-clip the view
		if(overlap(base_start, base_stop, iter->start(), iter->stop(), seq_size() ) == false){
			continue;
		}

		y1 = lower_y + delta*iter->annot_depth();
		y2 = y1 + delta;

		// Map the annotation element onto the display
		// while being respectful of the periodic boundary conditions.
		// All features are rendered relative to base_start,
		// which is positioned at the left most edge of the screen
		x1 = (long)(iter->start()) - base_start;
		x2 = x1 + iter->stop() - iter->start() + 1;

		if(base_start < base_stop){
			if(iter->start() > iter->stop()){
				x2 += seq_size();
			}
		}
		else{ // base_start > base_stop
			// Use <= to prevent 1 base annotations from falling
			// throught the cracks (so to speak).
			if(iter->start() <= iter->stop()){
				if(iter->stop() <= base_start){
					x1 += seq_size();
					x2 += seq_size();
				}
			}
			else{ // iter->start() > iter->stop()
				x2 += seq_size();
			}
		}
		
		GLfloat label_alpha = (m_select == true) ? -1.0f : 
			min(1.0f, (iter->length( seq_size() ) - label_alpha_offset)*label_alpha_norm);

		// Clamp the text alpha to the overall blending alpha -- if we don't
		// clamp the text alpha, then the annotation lable can overwrite the
		// sequence text.
		label_alpha = min(label_alpha, select_color[3]);

		for( long i = -(long)seq_size();i <= (long)seq_size();i += (long)seq_size() ){
			x1p = x1 + i;
			x2p = x2 + i;

			if((x1p > window_size_x) || (x2p < 0)){
				continue;
			}

			if(m_select){ // selection
				glLoadName(base_index);
			}
			else{	// screen rendering
				set_color(iter);
			}
			
			if(iter->type() == GeneAnnotation::NONE){
				// intergenic space gets a box, not an arrow
				render_box(x1p, y1, x2p, y2);
			}
			else{
				if(iter->is_complement()){
					if(iter->is_segmented()){
						render_left_segmented_arrow(x1p, y1, x2p, y2, iter, m_font_metric, label_alpha);
					}
					else{
						render_left_arrow(x1p, y1, x2p, y2, iter, m_font_metric, label_alpha);
					}
				}
				else{
					if(iter->is_segmented()){
						render_right_segmented_arrow(x1p, y1, x2p, y2, iter, m_font_metric, label_alpha);
					}
					else{
						render_right_arrow(x1p, y1, x2p, y2, iter, m_font_metric, label_alpha);
					}
				}
			}
		}
	}
	
	// Apply the clipping rectangle *before* writing any annotation text to the screen
	if(m_clip == true){
		
		glColor4fv(background_color);

		// In order to match the glClearColor when drawing a polygon, we must
		// disable GL_LIGHTING
		glDisable(GL_LIGHTING);

		glBegin(GL_POLYGON);
	
			// Render a rectange to cover up the periodic wrapping of annotations
			glVertex2f(window_size_x, lower_y);
			glVertex2f(window_size_x, upper_y);
			glVertex2f(m_clip_edge, upper_y);
			glVertex2f(m_clip_edge, lower_y);
			
		glEnd();
		
		glEnable(GL_LIGHTING);
	}

	if(!m_select && ( select_iter != mol.end() ) ){
		render_annot_info(m_lower_y, m_upper_y, *select_iter);
	}
}

void Genome::render_seq(const unsigned int &m_width, 
						const GLfloat &m_lower_y, const GLfloat &m_upper_y,
						const bool &m_select)
{
	// Do render sequence! This allow Genomorama to function (in a very limited way) as
	// a sequence alignment tool.
	//if(!is_annotation_shown){
	//	return;
	//}

	unsigned int base_start;
	unsigned int base_stop ;
	unsigned int base_index;

	base_range(base_start, base_stop);

	GLfloat y = 0.5f*(m_upper_y + m_lower_y);
	
	const unsigned int window_size_x = 2*base_window;

	// Render the genome as a string of characters representing
	// nucleotide bases (from 5'->3')

	DNA3Seq::iterator iter;

	if( hit_list.empty() ){
		// Set the text color
		glColor4fv(text_color);

		iter = mol.sequence().begin() + base_start;

		for(base_index = 0;base_index <= window_size_x;base_index++){
			unsigned char base = iter.c_base();

			TranslateFontBegin(base_index, y, text_font_scale_factor);
			glCallLists(1, GL_UNSIGNED_BYTE, &base);
			TranslateFontEnd();

			iter++;

			if( iter == mol.sequence().end() ){
				iter = mol.sequence().begin();
			}
		}
	}
	else{ // Their are entries in the hit list

		iter = mol.sequence().begin() + base_start;

		for(base_index = 0;base_index <= window_size_x;base_index++){

			// Does this base overlap with a sequence match?
			list<SeqHit>::iterator hit_iter;
			for(hit_iter = hit_list.begin();hit_iter != hit_list.end();hit_iter++){
				if(overlap( wrap(base_start + base_index, seq_size() ), 
						   hit_iter->start(), 
						   hit_iter->stop(), 
						   seq_size() ) == true){
					break;
				}
			}

			if( hit_iter != hit_list.end() ){
				if( hit_iter->complement() ){
					glColor4fv(select_neg_color);
				}
				else{
					glColor4fv(select_plus_color);
				}
			}
			else{
				// This base is not selected, use the default text color
				glColor4fv(text_color);
			}

			unsigned char base = iter.c_base();

			TranslateFontBegin(base_index, y, text_font_scale_factor);
			glCallLists(1, GL_UNSIGNED_BYTE, &base);
			TranslateFontEnd();

			iter++;

			if( iter == mol.sequence().end() ){
				iter = mol.sequence().begin();
			}
		}
	}
}

void Genome::render_GC_graph(const unsigned int &m_width, const GLfloat &m_lower_y, 
							 const GLfloat &m_upper_y)
{
	const GLfloat dy = m_upper_y - m_lower_y;
	unsigned int plot_start;
	unsigned int plot_stop;
	unsigned int plot_buffer;
	unsigned int plot_buffer_max;
#ifdef _WINDOWS
	const unsigned int plot_window = min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = max(1, base_window/GC_PLOT_POINTS);
#else
	const unsigned int plot_window = std::min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = std::max(1u, base_window/GC_PLOT_POINTS);
#endif // _WINDOWS
	const unsigned int gc_window_size = 2*plot_window + 1;

	unsigned int gc_count = 0;

	DNA3Seq::iterator iter_head, iter_tail;

	plot_buffer_max = gc_window_size;

	base_range(plot_start, plot_stop);

	// The plot_start and plot_stop are computed like base_range()
	plot_start = (plot_start >= plot_window) ? 
			(plot_start - plot_window) % seq_size() :
			seq_size() - (plot_window - plot_start);

	plot_stop = (plot_stop + plot_window) % seq_size();

	iter_head = iter_tail = mol.sequence().begin() + plot_start;

	for(plot_buffer = 0;plot_buffer < plot_buffer_max;plot_buffer++){
		switch(iter_tail.base()){
			case DNA3Seq::G:
			case DNA3Seq::C:
				gc_count++;
				break;
		};

		iter_tail++;

		if( iter_tail == mol.sequence().end() ){
			iter_tail = mol.sequence().begin();
		}
	}
	
	// Draw the graph in the inverse background color
	glColor4f(1.0f - background_color[0],
			  1.0f - background_color[1],
			  1.0f - background_color[2],
			  1.0f);

	glBegin(GL_LINE_STRIP);

		plot_buffer = 0;
		plot_buffer_max = 2*base_window;

		while(plot_buffer != plot_buffer_max){
			
			if(plot_buffer % plot_stride == 0){
				// % gc == gc_count/(gc_count + at_count)
				glVertex2f(plot_buffer, gc_count*dy/gc_window_size + m_lower_y);
			}

			// Adjust count and iterator positions
			switch(iter_head.base()){
				case DNA3Seq::G:
				case DNA3Seq::C:
					gc_count--;
					break;
			};

			iter_head++;

			if( iter_head == mol.sequence().end() ){
				iter_head = mol.sequence().begin();
			}

			switch(iter_tail.base()){
				case DNA3Seq::G:
				case DNA3Seq::C:
					gc_count++;
					break;
			};

			iter_tail++;

			if( iter_tail == mol.sequence().end() ){
				iter_tail = mol.sequence().begin();
			}

			plot_buffer++;
		}

	glEnd();
}

void Genome::render_AT_graph(const unsigned int &m_width, const GLfloat &m_lower_y, 
							 const GLfloat &m_upper_y)
{
	const GLfloat dy = m_upper_y - m_lower_y;
	unsigned int plot_start;
	unsigned int plot_stop ;
	unsigned int plot_buffer;
	unsigned int plot_buffer_max;
#ifdef _WINDOWS
	const unsigned int plot_window = min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = max(1, base_window/GC_PLOT_POINTS);
#else
	const unsigned int plot_window = std::min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = std::max(1u, base_window/GC_PLOT_POINTS);
#endif // _WINDOWS
	const unsigned int at_window_size = 2*plot_window + 1;

	unsigned int at_count = 0;

	DNA3Seq::iterator iter_head, iter_tail;

	plot_buffer_max = at_window_size;

	base_range(plot_start, plot_stop);

	// The plot_start and plot_stop are computed like base_range()
	plot_start = (plot_start >= plot_window) ? 
			(plot_start - plot_window) % seq_size() :
			seq_size() - (plot_window - plot_start);

	plot_stop = (plot_stop + plot_window) % seq_size();

	iter_head = iter_tail = mol.sequence().begin() + plot_start;

	for(plot_buffer = 0;plot_buffer < plot_buffer_max;plot_buffer++){
		switch(iter_tail.base()){
			case DNA3Seq::A:
			case DNA3Seq::T:
				at_count++;
				break;
		};

		iter_tail++;

		if( iter_tail == mol.sequence().end() ){
			iter_tail = mol.sequence().begin();
		}
	}
	
	// Draw the graph in the inverse background color
	glColor4f(1.0f - background_color[0],
			  1.0f - background_color[1],
			  1.0f - background_color[2],
			  1.0f);

	glBegin(GL_LINE_STRIP);

		plot_buffer = 0;
		plot_buffer_max = 2*base_window;

		while(plot_buffer != plot_buffer_max){
			
			if(plot_buffer % plot_stride == 0){
				// % at == at_count/(at_window_size)
				glVertex2f(plot_buffer, at_count*dy/at_window_size + m_lower_y);
			}

			// Adjust count and iterator positions
			switch(iter_head.base()){
				case DNA3Seq::A:
				case DNA3Seq::T:
					at_count--;
					break;
			};

			iter_head++;

			if( iter_head == mol.sequence().end() ){
				iter_head = mol.sequence().begin();
			}

			switch(iter_tail.base()){
				case DNA3Seq::A:
				case DNA3Seq::T:
					at_count++;
					break;
			};

			iter_tail++;

			if( iter_tail == mol.sequence().end() ){
				iter_tail = mol.sequence().begin();
			}

			plot_buffer++;
		}

	glEnd();
}

void Genome::render_GC_skew_graph(const unsigned int &m_width, const GLfloat &m_lower_y, 
							 const GLfloat &m_upper_y)
{
	const GLfloat dy = m_upper_y - m_lower_y;
	unsigned int plot_start;
	unsigned int plot_stop ;
	unsigned int plot_buffer;
	unsigned int plot_buffer_max;
#ifdef _WINDOWS
	const unsigned int plot_window = min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = max(1, base_window/GC_PLOT_POINTS);
#else
	const unsigned int plot_window = std::min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = std::max(1u, base_window/GC_PLOT_POINTS);
#endif // _WINDOWS
	const unsigned int skew_window_size = 2*plot_window + 1;

	int gPc_count = 0;
	int gMc_count = 0;

	DNA3Seq::iterator iter_head, iter_tail;

	plot_buffer_max = skew_window_size;

	base_range(plot_start, plot_stop);

	// The plot_start and plot_stop are computed like base_range()
	plot_start = (plot_start >= plot_window) ? 
			(plot_start - plot_window) % seq_size() :
			seq_size() - (plot_window - plot_start);

	plot_stop = (plot_stop + plot_window) % seq_size();

	iter_head = iter_tail = mol.sequence().begin() + plot_start;

	for(plot_buffer = 0;plot_buffer < plot_buffer_max;plot_buffer++){
		switch(iter_tail.base()){
			case DNA3Seq::G:
				gPc_count ++;
				gMc_count ++;
				break;
			case DNA3Seq::C:
				gPc_count ++;
				gMc_count --;
				break;
		};

		iter_tail++;

		if( iter_tail == mol.sequence().end() ){
			iter_tail = mol.sequence().begin();
		}
	}
	
	// Draw the graph in the inverse background color
	glColor4f(1.0f - background_color[0],
			  1.0f - background_color[1],
			  1.0f - background_color[2],
			  1.0f);

	glBegin(GL_LINE_STRIP);

		plot_buffer = 0;
		plot_buffer_max = 2*base_window;

		while(plot_buffer != plot_buffer_max){
			
			if(plot_buffer % plot_stride == 0){
				// -1 < GC skew < 1; rescale to [0,1]*dy
				glVertex2f(plot_buffer, dy*0.5f*((float)(gMc_count)/gPc_count + 1.0f) + m_lower_y );
			}

			// Adjust count and iterator positions
			switch(iter_head.base()){
				case DNA3Seq::G:
					gPc_count --;
					gMc_count --;
					break;
				case DNA3Seq::C:
					gPc_count --;
					gMc_count ++;
					break;
			};

			iter_head++;

			if( iter_head == mol.sequence().end() ){
				iter_head = mol.sequence().begin();
			}

			switch(iter_tail.base()){
				case DNA3Seq::G:
					gPc_count ++;
					gMc_count ++;
					break;
				case DNA3Seq::C:
					gPc_count ++;
					gMc_count --;
					break;
			};

			iter_tail++;

			if( iter_tail == mol.sequence().end() ){
				iter_tail = mol.sequence().begin();
			}

			plot_buffer++;
		}

	glEnd();
}

void Genome::render_AT_skew_graph(const unsigned int &m_width, const GLfloat &m_lower_y, 
							 const GLfloat &m_upper_y)
{
	const GLfloat dy = m_upper_y - m_lower_y;
	unsigned int plot_start;
	unsigned int plot_stop ;
	unsigned int plot_buffer;
	unsigned int plot_buffer_max;
#ifdef _WINDOWS
	const unsigned int plot_window = min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = max(1, base_window/GC_PLOT_POINTS);
#else
	const unsigned int plot_window = std::min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = std::max(1u, base_window/GC_PLOT_POINTS);
#endif // _WINDOWS
	const unsigned int skew_window_size = 2*plot_window + 1;

	int aPt_count = 0;
	int aMt_count = 0;

	DNA3Seq::iterator iter_head, iter_tail;

	plot_buffer_max = skew_window_size;

	base_range(plot_start, plot_stop);

	// The plot_start and plot_stop are computed like base_range()
	plot_start = (plot_start >= plot_window) ? 
			(plot_start - plot_window) % seq_size() :
			seq_size() - (plot_window - plot_start);

	plot_stop = (plot_stop + plot_window) % seq_size();

	iter_head = iter_tail = mol.sequence().begin() + plot_start;

	for(plot_buffer = 0;plot_buffer < plot_buffer_max;plot_buffer++){
		switch(iter_tail.base()){
			case DNA3Seq::A:
				aPt_count ++;
				aMt_count ++;
				break;
			case DNA3Seq::T:
				aPt_count ++;
				aMt_count --;
				break;
		};

		iter_tail++;

		if( iter_tail == mol.sequence().end() ){
			iter_tail = mol.sequence().begin();
		}
	}
	
	// Draw the graph in the inverse background color
	glColor4f(1.0f - background_color[0],
			  1.0f - background_color[1],
			  1.0f - background_color[2],
			  1.0f);

	glBegin(GL_LINE_STRIP);

		plot_buffer = 0;
		plot_buffer_max = 2*base_window;

		while(plot_buffer != plot_buffer_max){
			
			if(plot_buffer % plot_stride == 0){
				// -1 < AT skew < 1; rescale to [0,1]*dy
				glVertex2f(plot_buffer, dy*0.5f*((float)(aMt_count)/aPt_count + 1.0f) + m_lower_y );
			}

			// Adjust count and iterator positions
			switch(iter_head.base()){
				case DNA3Seq::A:
					aPt_count --;
					aMt_count --;
					break;
				case DNA3Seq::T:
					aPt_count --;
					aMt_count ++;
					break;
			};

			iter_head++;

			if( iter_head == mol.sequence().end() ){
				iter_head = mol.sequence().begin();
			}

			switch(iter_tail.base()){
				case DNA3Seq::A:
					aPt_count ++;
					aMt_count ++;
					break;
				case DNA3Seq::T:
					aPt_count ++;
					aMt_count --;
					break;
			};

			iter_tail++;

			if( iter_tail == mol.sequence().end() ){
				iter_tail = mol.sequence().begin();
			}

			plot_buffer++;
		}

	glEnd();
}

void Genome::render_keto_skew_graph(const unsigned int &m_width, const GLfloat &m_lower_y, 
							 const GLfloat &m_upper_y)
{
	const GLfloat dy = m_upper_y - m_lower_y;
	unsigned int plot_start;
	unsigned int plot_stop ;
	unsigned int plot_buffer;
	unsigned int plot_buffer_max;
#ifdef _WINDOWS
	const unsigned int plot_window = min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = max(1, base_window/GC_PLOT_POINTS);
#else
	const unsigned int plot_window = std::min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = std::max(1u, base_window/GC_PLOT_POINTS);
#endif // _WINDOWS
	const unsigned int skew_window_size = 2*plot_window + 1;

	int gPt_count = 0;
	int aPc_count = 0;

	DNA3Seq::iterator iter_head, iter_tail;

	plot_buffer_max = skew_window_size;

	base_range(plot_start, plot_stop);

	// The plot_start and plot_stop are computed like base_range()
	plot_start = (plot_start >= plot_window) ? 
			(plot_start - plot_window) % seq_size() :
			seq_size() - (plot_window - plot_start);

	plot_stop = (plot_stop + plot_window) % seq_size();

	iter_head = iter_tail = mol.sequence().begin() + plot_start;

	for(plot_buffer = 0;plot_buffer < plot_buffer_max;plot_buffer++){
		switch(iter_tail.base()){
			case DNA3Seq::A:
				aPc_count ++;
				break;
			case DNA3Seq::T:
				gPt_count ++;
				break;
			case DNA3Seq::G:
				gPt_count ++;
				break;
			case DNA3Seq::C:
				aPc_count ++;
				break;
		};

		iter_tail++;

		if( iter_tail == mol.sequence().end() ){
			iter_tail = mol.sequence().begin();
		}
	}
	
	// Draw the graph in the inverse background color
	glColor4f(1.0f - background_color[0],
			  1.0f - background_color[1],
			  1.0f - background_color[2],
			  1.0f);

	glBegin(GL_LINE_STRIP);

		plot_buffer = 0;
		plot_buffer_max = 2*base_window;

		while(plot_buffer != plot_buffer_max){
			
			if(plot_buffer % plot_stride == 0){
				// -1 < keto skew < 1; rescale to [0,1]*dy
				glVertex2f(plot_buffer, 
					dy*0.5f*(((float)(gPt_count - aPc_count))/skew_window_size + 1.0f) + m_lower_y );
			}

			// Adjust count and iterator positions
			switch(iter_head.base()){
				case DNA3Seq::A:
					aPc_count --;
					break;
				case DNA3Seq::T:
					gPt_count --;
					break;
				case DNA3Seq::G:
					gPt_count --;
					break;
				case DNA3Seq::C:
					aPc_count --;
					break;
			};

			iter_head++;

			if( iter_head == mol.sequence().end() ){
				iter_head = mol.sequence().begin();
			}

			switch(iter_tail.base()){
				case DNA3Seq::A:
					aPc_count ++;
					break;
				case DNA3Seq::T:
					gPt_count ++;
					break;
				case DNA3Seq::G:
					gPt_count ++;
					break;
				case DNA3Seq::C:
					aPc_count ++;
					break;
			};

			iter_tail++;

			if( iter_tail == mol.sequence().end() ){
				iter_tail = mol.sequence().begin();
			}

			plot_buffer++;
		}

	glEnd();
}

void Genome::render_purine_skew_graph(const unsigned int &m_width, const GLfloat &m_lower_y, 
							 const GLfloat &m_upper_y)
{
	const GLfloat dy = m_upper_y - m_lower_y;
	unsigned int plot_start;
	unsigned int plot_stop ;
	unsigned int plot_buffer;
	unsigned int plot_buffer_max;
#ifdef _WINDOWS
	const unsigned int plot_window = min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = max(1, base_window/GC_PLOT_POINTS);
#else
	const unsigned int plot_window = std::min(content_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = std::max(1u, base_window/GC_PLOT_POINTS);
#endif // _WINDOWS
	const unsigned int skew_window_size = 2*plot_window + 1;

	int gPa_count = 0;
	int tPc_count = 0;

	DNA3Seq::iterator iter_head, iter_tail;

	plot_buffer_max = skew_window_size;

	base_range(plot_start, plot_stop);

	// The plot_start and plot_stop are computed like base_range()
	plot_start = (plot_start >= plot_window) ? 
			(plot_start - plot_window) % seq_size() :
			seq_size() - (plot_window - plot_start);

	plot_stop = (plot_stop + plot_window) % seq_size();

	iter_head = iter_tail = mol.sequence().begin() + plot_start;

	for(plot_buffer = 0;plot_buffer < plot_buffer_max;plot_buffer++){
		switch(iter_tail.base()){
			case DNA3Seq::A:
				gPa_count ++;
				break;
			case DNA3Seq::T:
				tPc_count ++;
				break;
			case DNA3Seq::G:
				gPa_count ++;
				break;
			case DNA3Seq::C:
				tPc_count ++;
				break;
		};

		iter_tail++;

		if( iter_tail == mol.sequence().end() ){
			iter_tail = mol.sequence().begin();
		}
	}
	
	// Draw the graph in the inverse background color
	glColor4f(1.0f - background_color[0],
			  1.0f - background_color[1],
			  1.0f - background_color[2],
			  1.0f);

	glBegin(GL_LINE_STRIP);

		plot_buffer = 0;
		plot_buffer_max = 2*base_window;

		while(plot_buffer != plot_buffer_max){
			
			if(plot_buffer % plot_stride == 0){
				// -1 < keto skew < 1; rescale to [0,1]*dy
				glVertex2f(plot_buffer, 
					dy*0.5f*(((float)(gPa_count - tPc_count))/skew_window_size + 1.0f) + m_lower_y );
			}

			// Adjust count and iterator positions
			switch(iter_head.base()){
				case DNA3Seq::A:
					gPa_count --;
					break;
				case DNA3Seq::T:
					tPc_count --;
					break;
				case DNA3Seq::G:
					gPa_count --;
					break;
				case DNA3Seq::C:
					tPc_count --;
					break;
			};

			iter_head++;

			if( iter_head == mol.sequence().end() ){
				iter_head = mol.sequence().begin();
			}

			switch(iter_tail.base()){
				case DNA3Seq::A:
					gPa_count ++;
					break;
				case DNA3Seq::T:
					tPc_count ++;
					break;
				case DNA3Seq::G:
					gPa_count ++;
					break;
				case DNA3Seq::C:
					tPc_count ++;
					break;
			};

			iter_tail++;

			if( iter_tail == mol.sequence().end() ){
				iter_tail = mol.sequence().begin();
			}

			plot_buffer++;
		}

	glEnd();
}

void Genome::render_Tm_graph(const unsigned int &m_width, const GLfloat &m_lower_y, 
							 const GLfloat &m_upper_y)
{
	unsigned int plot_start;
	unsigned int plot_stop ;
	unsigned int plot_buffer;
	unsigned int plot_buffer_max;
#ifdef _WINDOWS
	const unsigned int plot_window = min(tm_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = max(1, base_window/TM_PLOT_POINTS);
#else
	const unsigned int plot_window = std::min(tm_window, (seq_size() - 1)/2);
	const unsigned int plot_stride = std::max(1u, base_window/TM_PLOT_POINTS);
#endif // _WINDOWS

	DNA3Seq::iterator iter_head, iter_tail;

	plot_buffer_max = 2*plot_window + 1;

	base_range(plot_start, plot_stop);

	// The plot_start and plot_stop are computed like base_range()
	plot_start = (plot_start >= plot_window) ? 
			(plot_start - plot_window) % seq_size() :
			seq_size() - (plot_window - plot_start);

	plot_stop = (plot_stop + plot_window) % seq_size();

	iter_head = iter_tail = mol.sequence().begin() + plot_start;

	NucCruc melt;

	melt.Salt(salt);
	melt.Strand(strand);

	for(plot_buffer = 0;plot_buffer < plot_buffer_max;plot_buffer++){
		switch(iter_tail.base()){
			case DNA3Seq::A:
				melt.push_back_native(NucCruc::A);
				break;
			case DNA3Seq::T:
				melt.push_back_native(NucCruc::T);
				break;
			case DNA3Seq::G:
				melt.push_back_native(NucCruc::G);
				break;
			case DNA3Seq::C:
				melt.push_back_native(NucCruc::C);
				break;
			default:
				melt.push_back_native(NucCruc::UNKNOWN);
				break;
		};

		iter_tail++;

		if( iter_tail == mol.sequence().end() ){
			iter_tail = mol.sequence().begin();
		}
	}
	
	// Draw the graph in the inverse background color
	glColor4f(1.0f - background_color[0],
			  1.0f - background_color[1],
			  1.0f - background_color[2],
			  1.0f);

	glBegin(GL_LINE_STRIP);

		plot_buffer = 0;
		plot_buffer_max = 2*base_window;

		while(plot_buffer != plot_buffer_max){
			
			if(plot_buffer % plot_stride == 0){
				glVertex2f(plot_buffer, (melt.Tm() - TM_MIN)/TM_DELTA + m_lower_y);
			}

			// Adjust count and iterator positions
			melt.pop_front();

			iter_head++;

			if( iter_head == mol.sequence().end() ){
				iter_head = mol.sequence().begin();
			}

			switch(iter_tail.base()){
				case DNA3Seq::A:
					melt.push_back_native(NucCruc::A);
					break;
				case DNA3Seq::T:
					melt.push_back_native(NucCruc::T);
					break;
				case DNA3Seq::G:
					melt.push_back_native(NucCruc::G);
					break;
				case DNA3Seq::C:
					melt.push_back_native(NucCruc::C);
					break;
				default:
					melt.push_back_native(NucCruc::UNKNOWN);
					break;
			};

			iter_tail++;

			if( iter_tail == mol.sequence().end() ){
				iter_tail = mol.sequence().begin();
			}

			plot_buffer++;
		}

	glEnd();
}

void Genome::render_genome_plot(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y)
{

	const GLfloat dy = m_upper_y - m_lower_y;
	unsigned int plot_start;
	unsigned int plot_stop;
	unsigned int i;

	// The number of data points to plot
	const unsigned int plot_len = genome_plot.size();
	
	const GLfloat data_dy = genome_plot_bounds_y.second - 
		genome_plot_bounds_y.first;

	const GLfloat norm_y = (data_dy == 0.0) ? 
		1.0f : 
		dy*0.9f/data_dy;

	if( (seq_size() == 0) || (plot_len == 0) ){
		return;
	}
	
	base_range(plot_start, plot_stop);

	// Draw the graph in the inverse background color
	glColor4f(1.0f - background_color[0],
			  1.0f - background_color[1],
			  1.0f - background_color[2],
			  1.0f);

	glBegin(GL_LINES);

	vector< pair<unsigned int, float> >::const_iterator last_iter = genome_plot.begin();
	vector< pair<unsigned int, float> >::const_iterator curr_iter = last_iter;
	curr_iter ++;

	if(plot_start > plot_stop){
	
		for(i = 1;i < plot_len;i++){
			
			// Pre-clip the line segment
			if(((last_iter->first > plot_stop) && (last_iter->first < plot_start)) &&
			   ((curr_iter->first > plot_stop) && (curr_iter->first < plot_start))){
				
				last_iter = curr_iter;
				curr_iter ++;

				continue;
			}
			
			// Map the plot point onto the display
			// while being respectful of the periodic boundary conditions.
			//long int tmp_base_last = (long int)(last_iter->first) - (long int)(plot_start);
			unsigned int tmp_base_last;

			if(last_iter->first < plot_start){
				tmp_base_last = last_iter->first + (seq_size() - plot_start);
			}
			else{
				tmp_base_last = last_iter->first - plot_start;
			}
			
			//long int tmp_base_curr = (long int)(curr_iter->first) - (long int)(plot_start);
			unsigned int tmp_base_curr;

			if(curr_iter->first < plot_start){
				tmp_base_curr = curr_iter->first + (seq_size() - plot_start);
			}
			else{
				tmp_base_curr = curr_iter->first - plot_start;
			}
			
			// Don't interpolate
			//if(curr_iter->first < (last_iter->first + 100) ){
			// Do interpolate
			if(true){ 
				glVertex2f(tmp_base_last, 
						(last_iter->second - genome_plot_bounds_y.first)*norm_y + m_lower_y);

				if(tmp_base_curr < tmp_base_last){

					// Do we need to wrap around?
					glVertex2f(plot_stop + (seq_size() - plot_start), 
						(last_iter->second - genome_plot_bounds_y.first)*norm_y + m_lower_y);

					glVertex2f(0, 
						(curr_iter->second - genome_plot_bounds_y.first)*norm_y + m_lower_y);
				}

				glVertex2f(tmp_base_curr, 
						(curr_iter->second - genome_plot_bounds_y.first)*norm_y + m_lower_y);
			}
			
			last_iter = curr_iter;
			curr_iter ++;
		}
	}
	else{ // plot_stop > plot_start
		for(i = 1;i < plot_len;i++){

			// Pre-clip the line segment
			if((last_iter->first < plot_start) && (curr_iter->first < plot_start)){
				
				last_iter = curr_iter;
				curr_iter ++;

				continue;
			}
			
			if((last_iter->first > plot_stop) && (curr_iter->first > plot_stop)){
				// All done
				break;
			}
			
			// Map the plot point onto the display
			// while being respectful of the periodic boundary conditions.
			
			// Don't interpolate
			//if(curr_iter->first < (last_iter->first + 100) ){
			// Do interpolate
			if(true){ 
			
				//long int tmp_base = (long int)(last_iter->first) - (long int)(plot_start);
				int tmp_base = last_iter->first - plot_start;

				glVertex2f(tmp_base, 
						(last_iter->second - genome_plot_bounds_y.first)*norm_y + m_lower_y);

				//tmp_base = (long int)(curr_iter->first) - (long int)(plot_start);
				tmp_base = curr_iter->first - plot_start;

				glVertex2f(tmp_base, 
						(curr_iter->second - genome_plot_bounds_y.first)*norm_y + m_lower_y);
			}
			
			last_iter = curr_iter;
			curr_iter ++;
		}
	}	

	glEnd();
}

void Genome::render_annot_info(const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		 const GeneAnnotation &m_annot)
{

	stringstream s;

	switch( m_annot.type() ){
		case GeneAnnotation::CDS:
			s << "CDS: ";
			break;
		case GeneAnnotation::GENE:
			s << "GENE: ";
			break;
		case GeneAnnotation::PSEUDO_GENE:
			s << "PSEUDO-GENE: ";
			break;
		case GeneAnnotation::RNA:
			s << "RNA: ";
			break;
		case GeneAnnotation::tRNA:
			s << "tRNA: ";
			break;
		case GeneAnnotation::PRIMER:
			s << "PCR Primer: ";
			break;
		case GeneAnnotation::USER:
			s << "User defined: ";
			break;
		case GeneAnnotation::NONE:
			s << "Intergenic space: ";
			break;
		default:
			s << "Unknown: ";
			break;
	};
	
	if(m_annot.info(GeneAnnotation::LOCUS) != ""){
		s << m_annot.info(GeneAnnotation::LOCUS) << " ";
	}
	else{
		if(m_annot.info(GeneAnnotation::LOCUS_TAG) != ""){
			s << m_annot.info(GeneAnnotation::LOCUS_TAG) << " ";
		}
	}
		
	s << m_annot.seq_id_str() << " ";

	// Handle annotations that wrap around the origin
	const unsigned int gene_stop = ( seq_size() > m_annot.stop() ) ?
		m_annot.stop() : ( m_annot.stop() - seq_size() );
		
	// Indicate gene complement (i.e. 3'->5') using "()"
	if(m_annot.is_complement()){
		s << "(" << m_annot.start() << ".." 
			<< gene_stop << ")" ;
	}
	else{
		// Indicate gene 5'->3' using "[]"
		s << "[" << m_annot.start() << ".." 
			<< gene_stop << "]" ;
	}

	if(m_annot.info(GeneAnnotation::PRODUCT) != ""){
		s << " " << m_annot.info(GeneAnnotation::PRODUCT);
	}
	
	if(m_annot.info(GeneAnnotation::EC) != ""){
		s << " EC(" << m_annot.info(GeneAnnotation::EC) << ")";
	}
	
	if(m_annot.type() == GeneAnnotation::PRIMER){
		if(m_annot.info(GeneAnnotation::PRIMER_TM) != ""){
			s << m_annot.info(GeneAnnotation::PRIMER_TM) << ", ";
		}

		if(m_annot.info(GeneAnnotation::PRIMER_GC) != ""){
			s << m_annot.info(GeneAnnotation::PRIMER_GC) << ", ";
		}
	}

	glColor4fv(text_color);

	TranslateFontBegin(0.0, m_lower_y*0.95f + m_upper_y*0.05f, text_font_scale_factor);
	glCallLists(s.str().size(), GL_UNSIGNED_BYTE, s.str().c_str());
	TranslateFontEnd();

	// Print the NOTE field on a separate line
	//if(m_annot.info(GeneAnnotation::NOTE) != ""){
	//	stringstream note;
	//
	//	note << m_annot.info(GeneAnnotation::NOTE);
	//
	//	glRasterPos2f(0.0, m_lower_y);
	//	glCallLists(note.str().size(), GL_UNSIGNED_BYTE, note.str().c_str());
	//}
}

void Genome::render_info(const GLfloat &m_lower_y, const GLfloat &m_upper_y)
{

	unsigned int base_start;
	unsigned int base_stop ;

	base_range(base_start, base_stop);

	// Genenome name [start_base, stop_base]
	stringstream string_buffer;

	// Indicate the overall reading direction (which will change if the user
	// inverts the molecule) with "[...]" for Seq_strand_plus and
	// "(...)" for Seq_strand_plus
	if(mol.strand() == Seq_strand_plus){
		string_buffer << mol.info(DNAMol::TAXA_NAME) << "(" << mol.seq_id_str() << ") ["
					  << base_start << ".." 
					  << base_stop << "]";
	}
	else{ //mol.strand() ==  Seq_strand_minus
		string_buffer << mol.info(DNAMol::TAXA_NAME) << "(" << mol.seq_id_str() << ") ("
					  << base_start << ".." 
					  << base_stop << ")";
	}

	glColor4fv(text_color);

	const GLfloat info_offset = 0.90f;

	TranslateFontBegin(0.0, info_offset*m_upper_y + (1.0f - info_offset)*m_lower_y, text_font_scale_factor);
	glCallLists(string_buffer.str().size(), GL_UNSIGNED_BYTE, string_buffer.str().c_str());
	TranslateFontEnd();
}

void Genome::render_select(const GLfloat &m_lower_y, const GLfloat &m_upper_y)

{
	// To indicate that this genome has been selected, draw a border around the genome
	// in the inverse background color.
	glColor4f(1.0f - background_color[0],
			1.0f - background_color[1],
			1.0f - background_color[2],
			0.25f);

	//GLfloat right_edge = 0.999*(2.0f*base_window + ((mol.sequence().size() % 2)));

	GLfloat right_edge = 2*base_window;

	if( (seq_size() % 2 == 1) && (right_edge >= seq_size() - 1) ){
		right_edge ++;
	}
	
	#define LINE_SELECTION
	
	#ifdef LINE_SELECTION
		// Fudge factor
		right_edge *= 0.999f;

		// Style the line to make selection visable
		glLineWidth(2.0);
		//glEnable(GL_LINE_STIPPLE);

		glBegin(GL_LINE_LOOP);
			glVertex2f(0.0f, m_lower_y);
			glVertex2f(0.0f, m_upper_y);
			glVertex2f(right_edge, m_upper_y);
			glVertex2f(right_edge, m_lower_y);
		glEnd();

		// Restore line style defaults
		//glDisable(GL_LINE_STIPPLE);
		glLineWidth(1.0);
	#endif //LINE_SELECTION
	
	
	#ifdef FILLED_POLYGON
	glEnable(GL_POLYGON_STIPPLE);
	
	if(is_blending == false){
		glEnable(GL_BLEND);
	}
	
	glBegin(GL_POLYGON);
		glVertex2f(0.0f, m_lower_y);
		glVertex2f(0.0f, m_upper_y);
		glVertex2f(right_edge, m_upper_y);
		glVertex2f(right_edge, m_lower_y);
	glEnd();
	
	glDisable(GL_POLYGON_STIPPLE);
	
	if(is_blending == false){
		glDisable(GL_BLEND);
	}
	#endif // FILLED_POLYGON
}

void Genome::render_translation_guide(const unsigned int &m_x1, const GLfloat &m_y1,
									  const unsigned int &m_x2, const GLfloat &m_y2)
{
	
	unsigned int base_start, base_stop;
	unsigned int start, stop;
	
	base_range(base_start, base_stop);
	
	if( wrap( m_x2 - m_x1, seq_size() ) > wrap( m_x1 - m_x2, seq_size() ) ){
		stop = m_x1;
		start = m_x2;
	}
	else{
		start = m_x1;
		stop = m_x2;
	}

	unsigned int seq_len = 1;
	
	// Compute the subsequence length (i.e. the distance between start and stop
	// in periodic coordinates).
	for(long i = (long)start;i != (long)stop;i++){
		seq_len ++;

		// Wrap around
		if(i == (long)seq_size() - 1){
			i = -1;
		}
	}


	GLfloat start_x;
	GLfloat stop_x;

	if(start >= base_start){
		start_x = start - base_start;
		stop_x = start_x + seq_len;
	}
	else{ // start < base_start
		if(base_start > base_stop){
			if(start < base_stop){
				start_x = start + (seq_size() - base_start);
				stop_x = start_x + seq_len;
			}
			else{
				start_x = 0;
				stop_x = start_x + seq_len - (base_start - start);
			}
		}
		else{
			start_x = 0;
			stop_x = start_x + seq_len - (base_start - start);
		}
	}
	
	// Clap the display of the translation guide to fit within the genome
	const GLfloat window_size = 2.0f*base_window;

	start_x = min(start_x, window_size);
	stop_x = min(stop_x, window_size);

	// Make the translation guide transparent
	glColor4f(1.0f - background_color[0],
			  1.0f - background_color[1],
			  1.0f - background_color[2],
			  0.5f);

	if(is_blending == false){
		glEnable(GL_BLEND);
	}
	
	glBegin(GL_POLYGON);
	
	// Render a rectange to cover up the periodic wrapping of annotations
	glVertex2f(start_x, m_y1);
	glVertex2f(stop_x, m_y1);
	glVertex2f(stop_x, m_y2);
	glVertex2f(start_x, m_y2);
	
	glEnd();

	if(is_blending == false){
		glDisable(GL_BLEND);
	}
}

void Genome::set_color(const list<GeneAnnotation>::iterator &m_iter)
{
	// Note that since PostScript can not handle alpha blending, we use
	// the following kludge:
	if((postscript_output == true) && is_blending){
		float ps_alpha;
		float ps_alpha_m1;

		// Set the color
		if(m_iter->select()){
			ps_alpha = select_color[3];
			ps_alpha_m1 = 1.0f - ps_alpha;

			glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*select_color[0],
				ps_alpha_m1*background_color[1]  + ps_alpha*select_color[1],
				ps_alpha_m1*background_color[2]  + ps_alpha*select_color[2],
				1.0f);
			return;
		}

		if(m_iter->is_custom_color()){
			
			ps_alpha = m_iter->color()[3];
			ps_alpha_m1 = 1.0f - ps_alpha;

			glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*m_iter->color()[0],
				ps_alpha_m1*background_color[1]  + ps_alpha*m_iter->color()[1],
				ps_alpha_m1*background_color[2]  + ps_alpha*m_iter->color()[2],
				1.0f);

			return;
		}

		switch(m_iter->type()){
			case GeneAnnotation::CDS:
				ps_alpha = cds_color[3];
				ps_alpha_m1 = 1.0f - ps_alpha;

				glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*cds_color[0],
					ps_alpha_m1*background_color[1]  + ps_alpha*cds_color[1],
					ps_alpha_m1*background_color[2]  + ps_alpha*cds_color[2],
					1.0f);

				break;
			case GeneAnnotation::GENE:
				ps_alpha = gene_color[3];
				ps_alpha_m1 = 1.0f - ps_alpha;

				glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*gene_color[0],
					ps_alpha_m1*background_color[1]  + ps_alpha*gene_color[1],
					ps_alpha_m1*background_color[2]  + ps_alpha*gene_color[2],
					1.0f);

				break;
			case GeneAnnotation::PSEUDO_GENE:
				ps_alpha = pseudo_gene_color[3];
				ps_alpha_m1 = 1.0f - ps_alpha;

				glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*pseudo_gene_color[0],
					ps_alpha_m1*background_color[1]  + ps_alpha*pseudo_gene_color[1],
					ps_alpha_m1*background_color[2]  + ps_alpha*pseudo_gene_color[2],
					1.0f);

				break;
			case GeneAnnotation::RNA:
			case GeneAnnotation::tRNA:
				ps_alpha = rna_color[3];
				ps_alpha_m1 = 1.0f - ps_alpha;

				glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*rna_color[0],
					ps_alpha_m1*background_color[1]  + ps_alpha*rna_color[1],
					ps_alpha_m1*background_color[2]  + ps_alpha*rna_color[2],
					1.0f);

				break;
			case GeneAnnotation::IMP:
				ps_alpha = annot_color[3];
				ps_alpha_m1 = 1.0f - ps_alpha;

				glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*annot_color[0],
					ps_alpha_m1*background_color[1]  + ps_alpha*annot_color[1],
					ps_alpha_m1*background_color[2]  + ps_alpha*annot_color[2],
					1.0f);

				break;
			case GeneAnnotation::PRIMER:
				ps_alpha = primer_color[3];
				ps_alpha_m1 = 1.0f - ps_alpha;

				glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*primer_color[0],
					ps_alpha_m1*background_color[1]  + ps_alpha*primer_color[1],
					ps_alpha_m1*background_color[2]  + ps_alpha*primer_color[2],
					1.0f);

				break;
			case GeneAnnotation::USER:
				ps_alpha = user_color[3];
				ps_alpha_m1 = 1.0f - ps_alpha;

				glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*user_color[0],
					ps_alpha_m1*background_color[1]  + ps_alpha*user_color[1],
					ps_alpha_m1*background_color[2]  + ps_alpha*user_color[2],
					1.0f);

				break;
			case GeneAnnotation::NONE:
				ps_alpha = intergenic_color[3];
				ps_alpha_m1 = 1.0f - ps_alpha;

				glColor4f(ps_alpha_m1*background_color[0]  + ps_alpha*intergenic_color[0],
					ps_alpha_m1*background_color[1]  + ps_alpha*intergenic_color[1],
					ps_alpha_m1*background_color[2]  + ps_alpha*intergenic_color[2],
					1.0f);

				break;
		};
	}
	else{

		// Set the color
		if(m_iter->select()){
			glColor4fv(select_color);
			return;
		}

		if(m_iter->is_custom_color()){
			glColor4fv(m_iter->color());
			return;
		}

		switch(m_iter->type()){
			case GeneAnnotation::CDS:
				glColor4fv(cds_color);
				break;
			case GeneAnnotation::GENE:
				glColor4fv(gene_color);
				break;
			case GeneAnnotation::PSEUDO_GENE:
				glColor4fv(pseudo_gene_color);
				break;
			case GeneAnnotation::RNA:
			case GeneAnnotation::tRNA:
				glColor4fv(rna_color);
				break;
			case GeneAnnotation::IMP:
				glColor4fv(annot_color);
				break;
			case GeneAnnotation::PRIMER:
				glColor4fv(primer_color);
				break;
			case GeneAnnotation::USER:
				glColor4fv(user_color);
				break;
			case GeneAnnotation::NONE:
				glColor4fv(intergenic_color);
				break;
		};
	}
}

void Genome::set_shadow_color(const list<GeneAnnotation>::iterator &m_iter)
{

	GLfloat shadow_color[4];
	
	// Set the color
	if(m_iter->select()){
		memcpy( shadow_color, select_color, 4*sizeof(GLfloat) );
	}
	else{
		if(m_iter->is_custom_color()){
			memcpy( shadow_color, m_iter->color(), 4*sizeof(GLfloat) );
		}
		else{
			switch(m_iter->type()){
				case GeneAnnotation::CDS:
					memcpy( shadow_color, cds_color, 4*sizeof(GLfloat) );
					break;
				case GeneAnnotation::GENE:
					memcpy( shadow_color, gene_color, 4*sizeof(GLfloat) );
					break;
				case GeneAnnotation::PSEUDO_GENE:
					memcpy( shadow_color, pseudo_gene_color, 4*sizeof(GLfloat) );
					break;
				case GeneAnnotation::RNA:
				case GeneAnnotation::tRNA:
					memcpy( shadow_color, rna_color, 4*sizeof(GLfloat) );
					break;
				case GeneAnnotation::IMP:
					memcpy( shadow_color, annot_color, 4*sizeof(GLfloat) );
					break;
				case GeneAnnotation::PRIMER:
					memcpy( shadow_color, primer_color, 4*sizeof(GLfloat) );
					break;
				case GeneAnnotation::USER:
					memcpy( shadow_color, user_color, 4*sizeof(GLfloat) );
					break;
				case GeneAnnotation::NONE:
					memcpy( shadow_color, intergenic_color, 4*sizeof(GLfloat) );
					break;
			};
		}
	}
	
	shadow_color[0] *= 0.1f;
	shadow_color[1] *= 0.1f;
	shadow_color[2] *= 0.1f;
	shadow_color[3] = 0.5f;
	
	glColor4fv(shadow_color);
}

// This function is supposed to return the inverse color scheme.
// However, using 1 - color as the inverse color leads to some
// hard to read text (especially when inverting red). Find a better
// scheme. For now, the inverse of the gene color is defined to be
// white -- the current background color. This needs to be fixed!
void Genome::set_inv_color(const list<GeneAnnotation>::iterator &m_iter)
{
	
	// Note that since PostScript can not handle alpha blending, we use
	// the following kludge:
	if((postscript_output == true) && (text_color[3] < 1.0f)){
		
		const float blend_channel = (1.0f - text_color[3]);

		if(m_iter->select()){
			glColor4f(blend_channel + text_color[3]*(1.0f - select_color[0]),
					  blend_channel + text_color[3]*(1.0f - select_color[1]),
					  blend_channel + text_color[3]*(1.0f - select_color[2]),
					  text_color[3]);
			return;
		}

		// Set the color
		switch(m_iter->type()){
			case GeneAnnotation::CDS:
				glColor4f(blend_channel + text_color[3]*cds_color[0],
						  blend_channel + text_color[3]*cds_color[1],
						  blend_channel + text_color[3]*cds_color[2],
						  text_color[3]);
				break;
			case GeneAnnotation::GENE:
				glColor4f(blend_channel + text_color[3]*gene_color[0],
					      blend_channel + text_color[3]*gene_color[1],
						  blend_channel + text_color[3]*gene_color[2],
						  text_color[3]);
				break;
			case GeneAnnotation::PSEUDO_GENE:
				glColor4f(blend_channel + text_color[3]*pseudo_gene_color[0],
					      blend_channel + text_color[3]*pseudo_gene_color[1],
						  blend_channel + text_color[3]*pseudo_gene_color[2],
						  text_color[3]);
				break;
			case GeneAnnotation::RNA:
			case GeneAnnotation::tRNA:
				glColor4f(blend_channel + text_color[3]*rna_color[0],
						  blend_channel + text_color[3]*rna_color[1],
						  blend_channel + text_color[3]*rna_color[2],
						  text_color[3]);
				break;
			case GeneAnnotation::IMP:
				glColor4f(blend_channel + text_color[3]*annot_color[0],
						  blend_channel + text_color[3]*annot_color[1],
						  blend_channel + text_color[3]*annot_color[2],
						  text_color[3]);
				break;
			case GeneAnnotation::PRIMER:
				glColor4f(blend_channel + text_color[3]*primer_color[0],
						  blend_channel + text_color[3]*primer_color[1],
						  blend_channel + text_color[3]*primer_color[2],
						  text_color[3]);
			case GeneAnnotation::USER:
				glColor4f(blend_channel + text_color[3]*user_color[0],
						  blend_channel + text_color[3]*user_color[1],
						  blend_channel + text_color[3]*user_color[2],
						  text_color[3]);
				break;
			case GeneAnnotation::NONE:
				glColor4f(blend_channel + text_color[3]*intergenic_color[0],
						  blend_channel + text_color[3]*intergenic_color[1],
						  blend_channel + text_color[3]*intergenic_color[2],
						  text_color[3]);
				break;
		};
	}
	else{

		if(m_iter->select()){
			glColor4f(1.0f - select_color[0],
					  1.0f - select_color[1],
					  1.0f - select_color[2],
					  text_color[3]);
			return;
		}

		glColor4f(1.0f,
				  1.0f,
				  1.0f,
				  text_color[3]);
	}
}

void Genome::enable_blend(const GLfloat &m_alpha)
{
	glEnable(GL_BLEND);

	is_blending = true;

	select_color[3] = m_alpha;
	background_color[3] = m_alpha;
	//text_color[3] = m_alpha;  <-- Don't blend text
	gene_color[3] = m_alpha;
	pseudo_gene_color[3] = m_alpha;
	cds_color[3] = m_alpha;
	annot_color[3] = m_alpha;
	primer_color[3] = m_alpha;
	rna_color[3] = m_alpha;
	intergenic_color[3] = m_alpha;
	select_neg_color[3] = m_alpha;
	select_plus_color[3] = m_alpha;

	list<GeneAnnotation>::iterator iter;

	for(iter = mol.begin();iter != mol.end();iter++){
		if(iter->is_custom_color()){
			iter->color_alpha(m_alpha);
		}
	}
}

void Genome::enable_blend_text(const GLfloat &m_alpha)
{
	glEnable(GL_BLEND);

	text_color[3] = m_alpha;
}

void Genome::disable_blend()
{
	glDisable(GL_BLEND);
	
	is_blending = false;

	select_color[3] = 1.0f;
	background_color[3] = 1.0f;
	//text_color[3] = 1.0f; <-- Don't blend text
	gene_color[3] = 1.0f;
	pseudo_gene_color[3] = 1.0f;
	cds_color[3] = 1.0f;
	annot_color[3] = 1.0f;
	primer_color[3] = 1.0f;
	rna_color[3] = 1.0f;
	intergenic_color[3] = 1.0f;
	select_neg_color[3] = 1.0f;
	select_plus_color[3] = 1.0f;

	list<GeneAnnotation>::iterator iter;

	for(iter = mol.begin();iter != mol.end();iter++){
		if(iter->is_custom_color()){
			iter->color_alpha(1.0f);
		}
	}
}

void Genome::disable_blend_text()
{
	// Only disable blending if we're currently not blending
	if(is_blending == false){
		glDisable(GL_BLEND);
	}

	text_color[3] = 1.0f;
}

void Genome::render_box(const GLfloat &x1, const GLfloat &y1,
						const GLfloat &x2, const GLfloat &y2)
{
	glBegin(GL_QUADS);

		// Render a rectangle
		// (follow the right hand rule -- i.e. 
		// counter-clockwise winding order).

		glVertex2f(x1, y1);
		glVertex2f(x2, y1);
		glVertex2f(x2, y2);
		glVertex2f(x1, y2);

	glEnd();
}

void Genome::render_right_arrow(const GLfloat &x1, const GLfloat &y1,
							    const GLfloat &x2, const GLfloat &y2,
							    const list<GeneAnnotation>::iterator &m_iter,
								const FontMetricGL &m_font_metric,
								const GLfloat &m_text_alpha)
{
	const GLfloat shaft = SHAFT_DX*x2 + (1.0f - SHAFT_DX)*x1;


	glBegin(GL_POLYGON);
	
		// Render a barbless arrow pointing to the right
		//
		//	||||||||||||||\	.
		//	||||||||||||||/	.
		//
		//	<-----80%---->
		//		 shaft
		
		// (follow the right hand rule -- i.e. 
		// counter-clockwise winding order).

		glVertex2f(x1, y1);
		glVertex2f(shaft, y1);
		glVertex2f(x2, 0.5f*(y2 + y1));
		glVertex2f(shaft, y2);
		glVertex2f(x1, y2);
		
	glEnd();

	if(!m_iter->has_id_str() || (m_text_alpha < MIN_TEXT_ALPHA)){
		return;
	}

	stringstream s;

	if(m_iter->info(GeneAnnotation::LOCUS) != ""){
	
		s << m_iter->info(GeneAnnotation::LOCUS);
	}
	else{
		if(m_iter->info(GeneAnnotation::LOCUS_TAG) != ""){
		
			s << m_iter->info(GeneAnnotation::LOCUS_TAG);
		}
	}

	unsigned int str_len = s.str().size();

	if(str_len == 0){
		return;
	}

	GLfloat x_font_scale = m_font_metric.size( s.str() );
	const GLfloat y_font_offset = (1.0f - SEGMENT_DY)*y1 + SEGMENT_DY*y2;
	
	if(x_font_scale <= 0.0f){
		return;
	}
	else{
		x_font_scale = 1.0f/x_font_scale;
	}

	enable_blend_text(m_text_alpha);

	set_inv_color(m_iter);

	glPushMatrix();

	glTranslatef(x1, y_font_offset, 0.0f);

	glScalef(x_font_scale*(shaft - x1), y2 - y1, 1.0f);

	glCallLists(s.str().size(), GL_UNSIGNED_BYTE, s.str().c_str());

	glPopMatrix();

	disable_blend_text();
}

void Genome::render_left_arrow(const GLfloat &x1, const GLfloat &y1,
							   const GLfloat &x2, const GLfloat &y2,
							   const list<GeneAnnotation>::iterator &m_iter,
							   const FontMetricGL &m_font_metric,
							   const GLfloat &m_text_alpha)
{
	const GLfloat head = (1.0f - SHAFT_DX)*x2 + SHAFT_DX*x1;

	glBegin(GL_POLYGON);
	
		// Render a barbless arrow pointing to the left
		//
		//	/||||||||||||||
		//	\||||||||||||||
		//  <> head
		
		// (follow the right hand rule -- i.e. 
		// counter-clockwise winding order).
		
		glVertex2f(head, y1);
		glVertex2f(x2, y1);
		glVertex2f(x2, y2);
		glVertex2f(head, y2);
		glVertex2f(x1, 0.5f*(y2 + y1));
		
	glEnd();

	if(!m_iter->has_id_str() || (m_text_alpha < MIN_TEXT_ALPHA)){
		return;
	}

	stringstream s;

	if(m_iter->info(GeneAnnotation::LOCUS) != ""){
		s << m_iter->info(GeneAnnotation::LOCUS);
	}
	else{
		if(m_iter->info(GeneAnnotation::LOCUS_TAG) != ""){
			s << m_iter->info(GeneAnnotation::LOCUS_TAG);
		}
	}

	unsigned int str_len = s.str().size();

	if(str_len == 0){
		return;
	}

	GLfloat x_font_scale = m_font_metric.size(s.str());
	const GLfloat y_font_offset = (1.0f - SEGMENT_DY)*y1 + SEGMENT_DY*y2;

	if(x_font_scale <= 0.0f){
		return;
	}
	else{
		x_font_scale = 1.0f/x_font_scale;
	}

	enable_blend_text(m_text_alpha);

	set_inv_color(m_iter);

	glPushMatrix();

	glTranslatef(head, y_font_offset, 0.0f);

	glScalef(x_font_scale*(x2 - head), y2 - y1, 1.0f);

	glCallLists(s.str().size(), GL_UNSIGNED_BYTE, s.str().c_str());

	glPopMatrix();

	disable_blend_text();
}

void Genome::render_right_segmented_arrow(const GLfloat &x1, const GLfloat &y1,
							    const GLfloat &x2, const GLfloat &y2,
							    const list<GeneAnnotation>::iterator &m_iter,
								const FontMetricGL &m_font_metric,
								const GLfloat &m_text_alpha)
{
	// Render a barbless, segmented arrow pointing to the right
	//
	//
	//	||   |||  ||||\		} seg_dy = 15%
	//	|||||||||||||| \ .	}
	//	|||||||||||||| /.	} shaft_dy = 100% - 2*seg_dy
	//	||   |||  ||||/		} seg_dy = 15%
	//
	//
	//	<----shaft--->
	//
	// (follow the right hand rule -- i.e. 
	// counter-clockwise winding order).

#ifdef _DEBUG
	if(m_iter->is_segmented() == false){
		throw "Genome::render_right_segmented_arrow: Non-segmented annotation!";
	}
#endif // _DEBUG

	const list< pair<unsigned int, unsigned int> > &seg_list = m_iter->segments();

	const unsigned int b_h = seg_list.front().first; // Start of first segment (in bases)
	const unsigned int b_t = seg_list.back().first; // Start of last segment (in bases)

	const unsigned int L = m_iter->stop() - m_iter->start(); // Annotation length (in bases)

	const GLfloat ya = (1.0f - SEGMENT_DY)*y1 + SEGMENT_DY*y2;
	const GLfloat yb = (1.0f - SEGMENT_DY)*y2 + SEGMENT_DY*y1;


	GLfloat shaft = SHAFT_DX*x2 + (1.0f - SHAFT_DX)*x1;
	const GLfloat last_seg = (GLfloat)(b_t - b_h)*(x2 - x1)/L + x1;

	// Where do we terminate the shaft?
	if(shaft < last_seg){
		shaft = last_seg;
	}

	glBegin(GL_POLYGON);
	
		// Render a rectange
		//
		//	||||||||||||||
		//	||||||||||||||
		//
		//	<-----shaft---->

		// (follow the right hand rule -- i.e. 
		// counter-clockwise winding order).

		glVertex2f(x1, ya);
		glVertex2f(shaft, ya);
		glVertex2f(shaft, yb);
		glVertex2f(x1, yb);
		
	glEnd();

	glBegin(GL_POLYGON);
	
		// Render a right pointing triangle
		//
		//	|\
		//	|/
		//

		// (follow the right hand rule -- i.e. 
		// counter-clockwise winding order).

		glVertex2f(shaft, y1);
		glVertex2f(x2, 0.5f*(y2 + y1));
		glVertex2f(shaft, y2);
		
	glEnd();

	// Render the "segment" flags symmetrically about the arrow
	list< pair<unsigned int, unsigned int> >::const_iterator iter;
	const GLfloat frac_len = (x2 - x1)/L;
	
	for(iter = seg_list.begin();iter != seg_list.end();iter++){

		GLfloat x_start = x1 + frac_len*(GLfloat)(iter->first - m_iter->start());
		GLfloat x_stop = min(shaft, x1 + frac_len*(GLfloat)(iter->second - m_iter->start()));

		glBegin(GL_POLYGON);
			// (follow the right hand rule -- i.e. 
			// counter-clockwise winding order).

			glVertex2f(x_start, y1);
			glVertex2f(x_stop, y1);
			glVertex2f(x_stop, ya);
			glVertex2f(x_start, ya);

		glEnd();

		glBegin(GL_POLYGON);
			// (follow the right hand rule -- i.e. 
			// counter-clockwise winding order).

			glVertex2f(x_start, yb);
			glVertex2f(x_stop, yb);
			glVertex2f(x_stop, y2);
			glVertex2f(x_start, y2);

		glEnd();
	}

	if(!m_iter->has_id_str() || (m_text_alpha < MIN_TEXT_ALPHA)){
		return;
	}

	stringstream s;

	if(m_iter->info(GeneAnnotation::LOCUS) != ""){
		s << m_iter->info(GeneAnnotation::LOCUS);
	}
	else{
		if(m_iter->info(GeneAnnotation::LOCUS_TAG) != ""){
			s << m_iter->info(GeneAnnotation::LOCUS_TAG);
		}
	}

	unsigned int str_len = s.str().size();

	if(str_len == 0){
		return;
	}

	GLfloat x_font_scale = m_font_metric.size( s.str() );
	const GLfloat y_font_offset = (1.0f - SEGMENT_DY)*ya + SEGMENT_DY*yb;

	if(x_font_scale <= 0.0f){
		return;
	}
	else{
		x_font_scale = 1.0f/x_font_scale;
	}

	enable_blend_text(m_text_alpha);

	set_inv_color(m_iter);

	glPushMatrix();

	glTranslatef(x1, y_font_offset, 0.0f);
	glScalef(x_font_scale*(shaft - x1), yb - ya, 1.0f);

	glCallLists(s.str().size(), GL_UNSIGNED_BYTE, s.str().c_str());

	glPopMatrix();

	disable_blend_text();
}

void Genome::render_left_segmented_arrow(const GLfloat &x1, const GLfloat &y1,
							   const GLfloat &x2, const GLfloat &y2,
							   const list<GeneAnnotation>::iterator &m_iter,
							   const FontMetricGL &m_font_metric,
							   const GLfloat &m_text_alpha)
{

	// Render a barbless, segmented arrow pointing to the left
	//
	//
	//
	//	 /||   |||  ||||	} seg_dy = 15%
	//	/ ||||||||||||||	}
	//	\ ||||||||||||||	} shaft_dy = 100% - 2*seg_dy
	//	 \||   |||  ||||	} seg_dy = 15%
	//
	//
	// <-> head
	// 
	// (follow the right hand rule -- i.e. 
	// counter-clockwise winding order).

#ifdef _DEBUG
	if(m_iter->is_segmented() == false){
		throw "Genome::render_right_segmented_arrow: Non-segmented annotation!";
	}
#endif // _DEBUG

	const list< pair<unsigned int, unsigned int> > &seg_list = m_iter->segments();

	const unsigned int b_h = seg_list.front().first;  // Start of first segment (in bases)
	const unsigned int b_t = seg_list.front().second; // End of first segment (in bases)

	const unsigned int L = m_iter->stop() - m_iter->start(); // Annotation length (in bases)
	
	const GLfloat ya = (1.0f - SEGMENT_DY)*y1 + SEGMENT_DY*y2;
	const GLfloat yb = (1.0f - SEGMENT_DY)*y2 + SEGMENT_DY*y1;

	GLfloat head = (1.0f - SHAFT_DX)*x2 + SHAFT_DX*x1;
	
	const GLfloat last_seg = (GLfloat)(b_t - b_h)*(x2 - x1)/L + x1;

	// Where do we terminate the shaft?
	if(head > last_seg){
		head = last_seg;
	}

	glBegin(GL_POLYGON);
	
		// Render a rectange
		//
		//	||||||||||||||
		//	||||||||||||||
		//
		// (follow the right hand rule -- i.e. 
		// counter-clockwise winding order).

		glVertex2f(head, ya);
		glVertex2f(x2, ya);
		glVertex2f(x2, yb);
		glVertex2f(head, yb);
		
	glEnd();

	glBegin(GL_POLYGON);
	
		// Render a left pointing triangle
		//
		//	/|
		//	\|
		//

		// (follow the right hand rule -- i.e. 
		// counter-clockwise winding order).

		glVertex2f(head, y1);
		glVertex2f(head, y2);
		glVertex2f(x1, 0.5f*(y1 + y2));
		
	glEnd();

	// Render the "segment" flags symmetrically about the arrow
	list< pair<unsigned int, unsigned int> >::const_iterator iter;
	const GLfloat frac_len = (x2 - x1)/L;
	
	for(iter = seg_list.begin();iter != seg_list.end();iter++){

		GLfloat x_start = max(head, x1 + frac_len*(GLfloat)(iter->first - m_iter->start()));
		GLfloat x_stop = x1 + frac_len*(GLfloat)(iter->second - m_iter->start());

		glBegin(GL_POLYGON);
			// (follow the right hand rule -- i.e. 
			// counter-clockwise winding order).

			glVertex2f(x_start, y1);
			glVertex2f(x_stop, y1);
			glVertex2f(x_stop, ya);
			glVertex2f(x_start, ya);

		glEnd();

		glBegin(GL_POLYGON);
			// (follow the right hand rule -- i.e. 
			// counter-clockwise winding order).

			glVertex2f(x_start, yb);
			glVertex2f(x_stop, yb);
			glVertex2f(x_stop, y2);
			glVertex2f(x_start, y2);

		glEnd();
	}

	if(!m_iter->has_id_str() || (m_text_alpha < MIN_TEXT_ALPHA)){
		return;
	}

	stringstream s;

	if(m_iter->info(GeneAnnotation::LOCUS) != ""){
		s << m_iter->info(GeneAnnotation::LOCUS);
	}
	else{
		if(m_iter->info(GeneAnnotation::LOCUS_TAG) != ""){
			s << m_iter->info(GeneAnnotation::LOCUS_TAG);
		}
	}

	unsigned int str_len = s.str().size();

	if(str_len == 0){
		return;
	}

	GLfloat x_font_scale = m_font_metric.size(s.str());
	const GLfloat y_font_offset = (1.0f - SEGMENT_DY)*ya + SEGMENT_DY*yb;

	if(x_font_scale <= 0.0f){
		return;
	}
	else{
		x_font_scale = 1.0f/x_font_scale;
	}

	enable_blend_text(m_text_alpha);

	set_inv_color(m_iter);

	glPushMatrix();

	glTranslatef(head, y_font_offset, 0.0f);
	glScalef(x_font_scale*(x2 - head), yb - ya, 1.0f);

	glCallLists(s.str().size(), GL_UNSIGNED_BYTE, s.str().c_str());

	glPopMatrix();

	disable_blend_text();
}

void Genome::render_auto_pick(const list<GeneAnnotation>::iterator &m_iter, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		const FontMetricGL &m_font_metric)
{
	
	if(!is_annotation_shown){
		return;
	}
	
	// The size of the window in bases
	const int window_size_x = 2*(int)(base_window);
	
	const GLfloat dy = m_upper_y - m_lower_y;
	
	const GLfloat delta = 0.3f;
	
	const GLfloat upper_y = (1.0f - delta)*dy + m_lower_y;
	const GLfloat lower_y = delta*dy + m_lower_y;
	
	const GLfloat left_x = delta*window_size_x;
	const GLfloat right_x = (1.0f - delta)*window_size_x;
	
	const GLfloat upper_y_shadow = upper_y - 0.05f;
	const GLfloat lower_y_shadow = lower_y - 0.05f;
	
	const GLfloat left_x_shadow = left_x - 0.05f*window_size_x;
	const GLfloat right_x_shadow = right_x - 0.05f*window_size_x;
	
	// Draw shadows first (using the painters algorithm)
	set_shadow_color(m_iter);

	if(m_iter->type() == GeneAnnotation::NONE){
	
		// intergenic space gets a box, not an arrow
		//render_box(left_x, lower_y, right_x, upper_y);
	}
	else{
	
		const GLfloat none[] = {0.0f, 0.0f, 0.0f, 0.0f};
		const GLfloat specular[] = {0.3f, 0.3f, 0.3f, 0.3f};
		const GLfloat shininess = 15.0f;
		const GLfloat un_shiny = 0.0f;
		
		glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
  		glMaterialfv(GL_FRONT, GL_SHININESS, &shininess);
	
		if( m_iter->is_complement() ){

			if( m_iter->is_segmented() ){
				
				glEnable(GL_BLEND);
				
				render_left_segmented_arrow(left_x_shadow, lower_y_shadow, right_x_shadow, upper_y_shadow, 
					m_iter, m_font_metric, 0.0f);
				
				glDisable(GL_BLEND);
				
				set_color(m_iter);
				
				glEnable(GL_LIGHT1);
				glNormal3f(0.0f, 0.0f, 1.0f);
				render_left_segmented_arrow(left_x, lower_y, right_x, upper_y, m_iter, m_font_metric, 1.0f);
				
				glDisable(GL_LIGHT1);
			}
			else{
				
				glEnable(GL_BLEND);
				
				render_left_arrow(left_x_shadow, lower_y_shadow, right_x_shadow, upper_y_shadow, 
					m_iter, m_font_metric, 0.0f);
				
				glDisable(GL_BLEND);
				
				set_color(m_iter);
				
				glEnable(GL_LIGHT1);
				glNormal3f(0.0f, 0.0f, 1.0f);
				render_left_arrow(left_x, lower_y, right_x, upper_y, m_iter, m_font_metric, 1.0f);
				
				glDisable(GL_LIGHT1);
			}
		}
		else{

			if( m_iter->is_segmented() ){
				
				glEnable(GL_BLEND);
				
				render_right_segmented_arrow(left_x_shadow, lower_y_shadow, right_x_shadow, upper_y_shadow, 
					m_iter, m_font_metric, 0.0f);	
				
				glDisable(GL_BLEND);
				
				set_color(m_iter);
				
				glEnable(GL_LIGHT1);
				glNormal3f(0.0f, 0.0f, 1.0f);
				render_right_segmented_arrow(left_x, lower_y, right_x, upper_y, m_iter, m_font_metric, 1.0f);
				
				glDisable(GL_LIGHT1);
			}
			else{
				
				glEnable(GL_BLEND);
				
				render_right_arrow(left_x_shadow, lower_y_shadow, right_x_shadow, upper_y_shadow, 
					m_iter, m_font_metric, 0.0f);
				
				glDisable(GL_BLEND);
				
				set_color(m_iter);
				
				glEnable(GL_LIGHT1);
				
				glNormal3f(0.0f, 0.0f, 1.0f);
				render_right_arrow(left_x, lower_y, right_x, upper_y, m_iter, m_font_metric, 1.0f);
				
				glDisable(GL_LIGHT1);
			}
		}
		
		glMaterialfv(GL_FRONT, GL_SPECULAR, none);
	  	glMaterialfv(GL_FRONT, GL_SHININESS, &un_shiny);
	}
}		
		
// Spin the selected genomes by the specified amount
void GenomePage::spin(const float &m_dx)
{
	list<Genome>::iterator iter;

	// If no genomes are selected, move everything
	if(num_select() == 0){
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			iter->shift(m_dx);
		}
	}
	else{

		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if(iter->select()){
				iter->shift(m_dx);
			}
		}
	}
}

// Return true if we're animating and need to redisplay
// Return false if not redisplay is required
bool GenomePage::render(const bool &m_is_printing /*= false*/)
{	
	// Are we animating?
	if(momentum && spinning){
		// Get current time
		//elapsed_time = clock();

#ifdef WIN32
		// The accuracy depends upon the OS that you're running
		// Win95 has accuracy to 1 ms, while WinNT will average about
		// 5 ms or more (fine for our purposes). If you're running NT you can
		// set the accuracy to a finer value. See the documentation.
		// (However, this runs fine under NT as it is...)
		elapsed_time = timeGetTime();
#else		
		// Unfortunately, use of gettimeofday() is most likely
		// platform specific!
		timeval timer;
		
		gettimeofday(&timer, NULL);
		elapsed_time = timer.tv_usec;		
#endif // WIN32

		// Is it too soon to re-render the scene?
		if(abs(elapsed_time - last_elapsed_time) < 15 /* msec */){
			return true;
		}

		last_elapsed_time = elapsed_time;

		// It's time to spin!
		spin(spin_dx);
	}

	glClear(GL_COLOR_BUFFER_BIT);

	// Set the global font dy variable
	font_dy = y_top - y_bottom;
	font_dx = base_window;
	
	if(genome_list.empty() == true){
		render_splash_screen(font_metric);

		return false; // We don't need to redisplay the splash screen
	}
	
	const unsigned int clip_edge = 2*base_window;

	const GLfloat dy = 1.0f;
	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	// Do we need to render a translation guide?
	const bool render_translation_guide = (translation_buffer_base.empty() == false);

	// Render the genomes by order in the list
	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if( !iter->show() ){
			continue;
		}
		
		if( !( (yh < y_bottom) || (yl > y_top) ) ){
		
			iter->render(width, yl, yh, clip_edge, font_metric, m_is_printing);

			// Give the user a visual cue as to which sequence is currrently selected for
			// translation
			if( render_translation_guide && ( iter == translation_buffer_genome.back() ) ){

				iter->render_translation_guide(translation_buffer_base.back(), yl, 
					// Render the translation guide using the base location in the
					// frame of the genome being selected!
					iter->cur_base_loc( (2.0f*base_window*mouse_loc[X])/width ), yh);
			}

		}

		yl = yh;
		yh += dy;
	}

	// Show a multiple sequence plot if requested by the user
	if(conservation_plot){
		// Plot multiple sequence conservation
		render_multiple_sequence_conservation();
	}

	// If the user is looking at sequence, track the mouse
	if(!m_is_printing && (zoom_level == ZOOM_LEVEL_SEQ)){
		render_mouse_guide();
	}

	#ifdef _DEBUG
	try{
		get_error_OGL(__FILE__, __LINE__);
	}
	catch(const char *error){

	#ifdef WIN32
			AfxMessageBox(error);
	#endif // WIN32

	}
	#endif // _DEBUG

	// Do we need to redisplay?
	return (momentum && spinning);
}

void GenomePage::render_ps(const std::string &m_name)
{
	GLtoPS ps_engine(m_name);

	// Expand the clipping region to prevent text 
	// near the top or bottom from
	// being clipped away. This number may need to be
	// adjustable by the user!
	const GLfloat fudge_factor = 0.02f;

	glLoadIdentity();
	gluOrtho2D(0.0, 2*base_window, 
			   y_bottom, 
			   y_top + fudge_factor*(y_top - y_bottom));

	// Note that since PostScript can not handle alpha blending, we use
	// the following kludge:
	postscript_output = true;

	do {
		ps_engine.Start();
		
		render(true /* Set the printing flag */);

	}
	while(ps_engine.Stop() == false);

	// Restore use of the alpha channel
	postscript_output = false;

	// Restore the clipping
	glLoadIdentity();
	gluOrtho2D(0.0, 2*base_window, 
			   y_bottom, y_top);

}

// "Why render gifs in this day and age of png, jpeg, etc." you ask?
// Because we can use the gif creation code included in the NCBI library!
void GenomePage::render_gif(const std::string &m_name)
{
	
	// Render the image with the printing flag enabled
	// (to turn off all non-printing decorations).
	render(true);

	// Prepare color map
	unsigned char *pixel_buffer = NULL;

	pixel_buffer = new unsigned char[width*height*3*sizeof(unsigned char)];
	
	if(pixel_buffer == NULL){
		throw "Unable to allocate pixel buffer";
	}

	glPixelStorei(GL_PACK_ALIGNMENT,1);

	glReadBuffer(GL_BACK);

	glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, pixel_buffer);

	// Make the color map. Note that we shouldn't use an unsigned char
	// for the color_index variable since we will be unable to test for
	// index values greater than 255
	unsigned int color_index = 0;
	vector<unsigned int> color_map(256);
	
	// Since the vector will be initialized to 0, start the color
	// map off with black (at color_index == 0)
	color_map[color_index] = 0;			
	color_index ++;

	unsigned int i, j;

	for(i = 0;i < width;i++){
		for(j = 0;j < height;j++){
			unsigned int pos = (j*width + i)*3*sizeof(unsigned char);

			unsigned int color = pixel_buffer[pos] +			// R
							    (pixel_buffer[pos + 1] << 8) +	// G
								(pixel_buffer[pos + 2] << 16);	// B
			if(color_map.end() == find(color_map.begin(), color_map.end(), color)){
				color_map[color_index] = color;
				
				color_index ++;
			}
		}
	}
	
	gdImagePtr gif_ptr = gdImageCreate(width, height);

	if(gif_ptr == NULL){
		throw "Unable to allocate memory for gif image";
	}
	
	// Allocate the color map
	map<unsigned int, int> gif_color_map;

	for(i = 0;i < color_index;i++){

		if( gif_color_map.find(color_map[i]) != gif_color_map.end() ){
			continue;
		}

		int r = color_map[i] & 0x000000ff;
		int g = (color_map[i] >> 8) & 0x000000ff;
		int b = (color_map[i] >> 16) & 0x000000ff;

		// If there are colors left in the color table, allocate a new color,
		// otherwise, map to the closest available color.
		int gif_index = (i < gdMaxColors) ?
			gdImageColorAllocate(gif_ptr, r, g, b) :
			gdImageColorClosest(gif_ptr, r, g, b);

		gif_color_map[color_map[i]] = gif_index;
	}

	for(i = 0;i < width;i++){
		for(j = 0;j < height;j++){
			unsigned int pos = (j*width + i)*3*sizeof(unsigned char);

			unsigned int color = pixel_buffer[pos] +			// R
							    (pixel_buffer[pos + 1] << 8) +	// G
								(pixel_buffer[pos + 2] << 16);	// B

			gdImageSetPixel(gif_ptr, i, height - 1 - j, gif_color_map[color]);
		}
	}

	FILE *fout = fopen(m_name.c_str(), "w+b");

	if(fout == NULL){
		throw "Unable to open gif file for writing!";
	}

	gdImageGif(gif_ptr, fout);

	fclose(fout);
	
	delete [] pixel_buffer;

	gdImageDestroy(gif_ptr);

	// Deallocate the color map
	//map<unsigned int, int>::iterator map_iter;

	//for(map_iter = gif_color_map.begin();map_iter != gif_color_map.end();map_iter++){
	//	gdImageColorDeallocate(gif_ptr, map_iter->second);
	//}
}


void GenomePage::render_mouse_guide()
{
	const GLfloat left = (int)((2.0f*base_window*mouse_loc[X])/width);
	const GLfloat dx = font_metric.size("A");	// Use a representative base to determine the
												// guide width

	// The guide color is inverse background
	glColor4f(1.0f - background[0],
			  1.0f - background[1],
			  1.0f - background[2],
			  1.0f);

	// Note that the mouse guide transforms like a font in the X-dimension. 
	// The following transformations are based on the "TranslateFontBegin" macro.
	// The font macros are not used because the Y-dimension does not need to be
	// rescaled.

	glPushMatrix();
	glTranslatef(left, 0.0f, 0.0f);
	glScalef(font_dx*text_font_scale_factor*GP_POLYGON_FONT_SCALE, 1.0f, 1.0f);

	glBegin(GL_LINES);
		glVertex2f(0.0f, y_bottom);
		glVertex2f(0.0f, y_top);
		glVertex2f(dx, y_bottom);
		glVertex2f(dx, y_top);
	glEnd();

	glPopMatrix();
}

void GenomePage::render_splash_screen(const FontMetricGL &m_font_metric)
{
	// The splash screen makes sure that the user is never confronted with a blank window

	const GLfloat title_color[] = {1.0f, 0.5f, 0.0f, 1.0f};
	const GLfloat title_scale_factor = 4.0f;
	const string title = "Genomorama v" GV_VERSION_MAJOR "." GV_VERSION_MINOR;

	const GLfloat sub_title_color[] = {0.0f, 1.0f, 0.0f, 1.0f};
	const GLfloat sub_title_scale_factor = 1.75f;
	const string sub_title = "Open a genome annontation file or a FASTA file to get started";
	
	const GLfloat sub_sub_title_color[] = {0.0f, 0.0f, 0.0f, 1.0f};
	const GLfloat sub_sub_title_scale_factor = 1.5f;
	const string sub_sub_title = "Documentation and source code available from " AVAILIABILITY;

	const GLfloat sub_sub_sub_title_color[] = {0.0f, 0.0f, 0.0f, 1.0f};
	const GLfloat sub_sub_sub_title_scale_factor = 1.5f;
	const string sub_sub_sub_title = "Please send bug reports to " CONTACT_ADDRESS;

	glColor4fv(title_color);

	GLfloat left_edge = base_window*(1.0f - 
			0.5f*m_font_metric.size(title)*GP_POLYGON_FONT_SCALE*title_scale_factor);

	TranslateFontBegin(left_edge, 0.75f, title_scale_factor);
	glCallLists( title.size(), GL_UNSIGNED_BYTE, title.c_str() );
	TranslateFontEnd();
	
	glColor4fv(sub_title_color);

	left_edge = base_window*(1.0f - 
			0.5f*m_font_metric.size(sub_title)*GP_POLYGON_FONT_SCALE*sub_title_scale_factor);

	TranslateFontBegin(left_edge, 0.5f, sub_title_scale_factor);
	glCallLists( sub_title.size(), GL_UNSIGNED_BYTE, sub_title.c_str() );
	TranslateFontEnd();

	glColor4fv(sub_sub_title_color);
	
	left_edge = base_window*(1.0f - 
			0.5f*m_font_metric.size(sub_sub_title)*GP_POLYGON_FONT_SCALE*sub_sub_title_scale_factor);

	TranslateFontBegin(left_edge, 0.1f, sub_sub_title_scale_factor);
	glCallLists( sub_sub_title.size(), GL_UNSIGNED_BYTE, sub_sub_title.c_str() );
	TranslateFontEnd();

	glColor4fv(sub_sub_sub_title_color);

	left_edge = base_window*(1.0f - 
			0.5f*m_font_metric.size(sub_sub_sub_title)*GP_POLYGON_FONT_SCALE*sub_sub_sub_title_scale_factor);

	TranslateFontBegin(left_edge, 0.05f, sub_sub_sub_title_scale_factor);
	glCallLists( sub_sub_sub_title.size(), GL_UNSIGNED_BYTE, sub_sub_sub_title.c_str() );
	TranslateFontEnd();
}
